// ***************************************************************************************
// File: command.h
// Description: declarations of the Nirva command classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: A command contains a single command context.
// ***************************************************************************************

#ifndef _COMMAND_H
#define _COMMAND_H

// Classes declared in this file
class NvCommand;
class MyNvServiceCommand;

#include "task.h"
#include "container.h"
#include "nvsext.h"		// external interface to services
#include "nvperl.h"		// external interface to perl
#include "nvjava.h"		// external interface to java
#include "nvdotnet.h"	// external interface to dotnet
#include <string>


#define	XML_IN_NONE		0
#define	XML_IN_XML		1
#define	XML_IN_SOAP		2

// Ouput types
#define	OUTPUT_BINARY		0
#define	OUTPUT_FILE			1
#define	OUTPUT_XML			2

// Encoding
#define	ENCODING_NONE		0
#define	ENCODING_UTF8		1
#define	ENCODING_LATIN1		2

void EncodeCommandResult(int iEncoding, std::string& StringToEncode);
void EncodeCommandInputString(int iEncoding, std::string& StringToEncode);

/////////////////////////////////////////////////////////////////////////////
// NvCommand class
// The command class contains the context of a command
class NvCommand
{
private:
public:
	void *m_pSession;							// Session of the command
	void *m_pParentSession;						// Parent Session of the command. Used when another session is called from a procedure or service
	bool m_bInitNamedSessionCommand;			// Set to true when this is the command just got the ownership of a named session
	bool m_bExitNamedSessionCommand;			// Set to true when this is the command just released the ownership of a named session
	bool m_bHasLockedSession;					// Set to true when the command has locked the session

	int m_iCommandSource;
	std::string m_sService;
	std::string m_sClass;
	std::string m_sCommand;
	std::string m_sInContainer;
	std::string m_sOutContainer;
	std::string m_sLanguage;
	std::string m_sHost;						// Used in commands from browser
	std::string m_sSourceService;				// Source service when command is sent from a service
	std::string m_sClientAddress;				// TCP/IP client address when command comes from http
	std::string m_sLastOutputBuffer;			// last output buffer
	std::string m_sLastOutputBufferEncoded;		// last output buffer encoded


	std::string m_sMachineName;
	std::string m_sMachineUser;

	NvError Error;							// Error object
	NvObjIndStringList Parameters;			// Command Parameters
	unsigned long DataSize;					// associated data size
	char *Data;								// associated data
	std::string DataFile;					// associated data file
	bool bDeleteDataFile;					// flag for automatic removing of data file
		
	NvObjIndStringList HttpInHeaders;		// Http input headers (used for commands from browser)
	NvObjIndStringList HttpOutHeaders;		// Http output headers (used for commands from browser)

	int iOutputType;						// Generally OUTPUT_XML but may be changed by some commands
	int iXmlInputType;						// Type of XML input XML_IN_NONE, XML_IN_XML, XML_IN_SOAP
	int iXmlCoding;							// Type of XML coding required
	bool bWebs;								// Set to true when web service connector is used
	std::string m_sInputCharset;			// Eventual input charset
	std::string m_sXmlNameSpace;			// Eventual XML namespace for the output
	bool bLockSession;						// Locks the session for this command (used for recursive call to a session from an EXEC command)
	bool bClientUnicode;					// Set to true when a nvc client is unicode
	bool bFastMode;							// Fast mode
	bool bDebugCommand;						// Set to true when in a debug command


	NvCommand();
	virtual ~NvCommand();

	void Clear();
	void FreeData();
	
	void ParseParameters(const char *ParametersToParse = NULL, NvObjIndStringList* Variables = NULL, char sep = 0);
	void PostParseParameters(NvObjIndStringList* Variables = NULL);
	std::string GetVariable(std::string Value, NvObjIndStringList* Variables = NULL);
	void GetStringValue(const char *String, char *Buffer = NULL, int BufferSize = 0, int *DataSizeHere = NULL, int iEncoding = ENCODING_NONE);
	void GetLastOutputBuffer(char *Buffer = NULL, int BufferSize = 0);
	
	// Send a command to Nirva (this is used from service or perl script or java method or dotnet method)
	bool Command(const char *CommandString, char *Buffer = NULL, int BufferSize = 0, int *DataSizeHere = NULL, int CommandSource = NV_SOURCE_SERVICE, NvError *ErrorToKeep = NULL, int iEncoding = ENCODING_NONE);
};



/////////////////////////////////////////////////////////////////////////////
// MyNvServiceCommand class
// Interface to services
class MyNvServiceCommand : public NvServiceCommand
{
public:
	void *m_pCommand;					// Pointer to a Nirva command object. Should not be used directly by the service
	std::string m_sServiceSessionId;		// String used by the service to store its internal session ID
	int i_Encoding;						// Encoding

	MyNvServiceCommand();
	virtual ~MyNvServiceCommand();


	// Set the error
	void SetError(const char *ErrorClass = "", int ErrorCode = 0, const char *ErrorInfo = "");
	
	// Send a command to Nirva
	bool Command(const char *CommandString, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL);

	void GetSessionId(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetClass(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetCommand(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetInContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetOutContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetLanguage(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineName(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineUser(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetError(const char *InfoType, char *Buffer, int BufferSize, int *DataSize = NULL);
	int GetNumParameters();
	int GetSource();
	bool ParameterExist(const char *ParameterName);
	void GetParameter(const char *ParameterName, char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetParameterName(int index, char *Buffer, int BufferSize, int *DataSize = NULL);
	void AddHttpHeader(const char *HeaderName, const char *HeaderValue);
	void SetServiceSession(void *ServiceSession);
	void* GetServiceSession();
	bool IsCommand(const char *Class = NULL, const char *Command = NULL, int Source = -1);
};



/////////////////////////////////////////////////////////////////////////////
// MyNvPerlCommand class
// Interface to perl
class MyNvPerlCommand : public NvPerlCommand
{
public:
	void *m_pCommand;					// Pointer to a Nirva command object. Should not be used directly by the perl interface
	NvError CopyError;					// Copy Error object used to keep the last error when the error control is given to the script
	int i_Encoding;						// Encoding

	MyNvPerlCommand();
	virtual ~MyNvPerlCommand();

	// System perl log
	void SysLog(const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1, const char *Source = NULL);
	int GetLogLevel();

	// Set the error
	void SetError(const char *ErrorInfo = "");
	void SetErrorEx(const char *ErrorService, const char *ErrorClass, int ErrorCode, const char *ErrorInfo = "");
	
	// Send a command to Nirva
	bool Command(const char *CommandString, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL, bool ControlErrorToScript = false);
	void GetSessionId(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetService(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetClass(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetCommand(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetInContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetOutContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetLanguage(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineName(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineUser(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetError(const char *InfoType, char *Buffer, int BufferSize, int *DataSize = NULL);
	int GetNumParameters();
	int GetSource();
	bool ParameterExist(const char *ParameterName);
	void GetParameter(const char *ParameterName, char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetParameterName(int index, char *Buffer, int BufferSize, int *DataSize = NULL);
	void SetEncoding(const char *Encoding);
};


/////////////////////////////////////////////////////////////////////////////
// MyNvJavaCommand class
// Interface to java
class MyNvJavaCommand : public NvJavaCommand
{
public:
	void *m_pCommand;					// Pointer to a Nirva command object. Should not be used directly by the java interface
	NvError CopyError;					// Copy Error object used to keep the last error when the error control is given to the script
	int i_Encoding;						// Encoding

	MyNvJavaCommand();
	virtual ~MyNvJavaCommand();

	// System java log
	void SysLog(const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1, const char *Source = NULL);
	int GetLogLevel();

	// Set the error
	void SetError(const char *ErrorInfo = "");
	void SetErrorEx(const char *ErrorService, const char *ErrorClass, int ErrorCode, const char *ErrorInfo = "");
	
	// Send a command to Nirva
	bool Command(const char *CommandString, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL);
	void GetSessionId(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetService(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetClass(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetCommand(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetInContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetOutContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetLanguage(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineName(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineUser(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetError(const char *InfoType, char *Buffer, int BufferSize, int *DataSize = NULL);
	int GetNumParameters();
	int GetSource();
	bool ParameterExist(const char *ParameterName);
	void GetParameter(const char *ParameterName, char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetParameterName(int index, char *Buffer, int BufferSize, int *DataSize = NULL);
};



/////////////////////////////////////////////////////////////////////////////
// MyNvDotnetCommand class
// Interface to dotnet
class MyNvDotnetCommand : public NvDotnetCommand
{
public:
	void *m_pCommand;					// Pointer to a Nirva command object. Should not be used directly by the dotnet interface
	NvError CopyError;					// Copy Error object used to keep the last error when the error control is given to the script
	int i_Encoding;						// Encoding

	MyNvDotnetCommand();
	virtual ~MyNvDotnetCommand();

	// System dotnet log
	void SysLog(const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1, const char *Source = NULL);
	int GetLogLevel();

	// Set the error
	void SetError(const char *ErrorInfo = "");
	void SetErrorEx(const char *ErrorService, const char *ErrorClass, int ErrorCode, const char *ErrorInfo = "");
	
	// Send a command to Nirva
	bool Command(const char *CommandString, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL);
	void GetSessionId(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetService(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetClass(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetCommand(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetInContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetOutContainer(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetLanguage(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineName(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetMachineUser(char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetError(const char *InfoType, char *Buffer, int BufferSize, int *DataSize = NULL);
	int GetNumParameters();
	int GetSource();
	bool ParameterExist(const char *ParameterName);
	void GetParameter(const char *ParameterName, char *Buffer, int BufferSize, int *DataSize = NULL);
	void GetParameterName(int index, char *Buffer, int BufferSize, int *DataSize = NULL);
};


#endif //_COMMAND_H
