// ***************************************************************************************
// File: dotnet.cpp
// Description: interface from nvs to dotnet
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: the real dotnet interface is implemented in a dedicated library
//           named nvdotnet. This file is the layer that interfaces to this
//           library.
// ***************************************************************************************


#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvDotnet

NvDotnet::NvDotnet()
{
	hLibDotnet = NULL;
	DotnetLibInit = NULL;
	DotnetLibFree = NULL;
	DotnetLibRunClass = NULL;
	DotnetLibUnloadDomain = NULL;
	NumInstances = 0;
	m_bAvailable = false;
}

NvDotnet::~NvDotnet()
{
	// Unload the dotnet interface library if loaded
	FreeDotnetLib();
}


bool NvDotnet::LoadDotnetLib(const char *NirvaDir, bool UseLocking)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectDotnet);
	if(hLibDotnet != NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectDotnet);
		return true;
	}

	NumInstances = 0;


	// Load the library here
	string DotnetLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(DotnetLibPath);
	DotnetLibPath += "Bin";
	AddDirSuffix(DotnetLibPath);
	DotnetLibPath += "nvdotnet";
#ifdef WIN32
#ifdef NV_DEBUG
	DotnetLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(DotnetLibPath);
	DotnetLibPath += "Bin";
	AddDirSuffix(DotnetLibPath);
	DotnetLibPath += "nvdotnetd";
#endif
	DotnetLibPath += ".dll";
#endif
#ifdef LINUX
	DotnetLibPath += ".so";
#endif
#ifdef HPUX
	DotnetLibPath += ".sl";
#endif
#ifdef HPUXI
	DotnetLibPath += ".so";
#endif
#ifdef SOLARIS
	DotnetLibPath += ".so";
#endif
#ifdef AIX
	DotnetLibPath += ".a";
#endif

	hLibDotnet = MyLoadLibrary(DotnetLibPath.c_str());
	if(hLibDotnet == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectDotnet);
		return false;
	}

	DotnetLibInit = (DOTNETINIT)MyGetLibFunction(hLibDotnet, "NvInitDotnet");
	DotnetLibFree = (DOTNETFREE)MyGetLibFunction(hLibDotnet, "NvFreeDotnet");
	DotnetLibRunClass = (DOTNETRUNCLASS)MyGetLibFunction(hLibDotnet, "NvRunDotnetClass");
	DotnetLibUnloadDomain = (DOTNETUNLOADDOMAIN)MyGetLibFunction(hLibDotnet, "NvDotnetUnloadDomain");
	if((DotnetLibInit == NULL) || (DotnetLibFree == NULL) || (DotnetLibRunClass == NULL) || (DotnetLibUnloadDomain == NULL))
	{
		MyFreeLibrary(hLibDotnet);
		DotnetLibInit = NULL;
		DotnetLibFree = NULL;
		DotnetLibRunClass = NULL;
		DotnetLibUnloadDomain = NULL;
		hLibDotnet = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectDotnet);
		return false;
	}

	if((DotnetLibInit)(NirvaDir) == 0)
	{
		MyFreeLibrary(hLibDotnet);
		DotnetLibInit = NULL;
		DotnetLibFree = NULL;
		DotnetLibRunClass = NULL;
		DotnetLibUnloadDomain = NULL;
		hLibDotnet = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectDotnet);
		return false;
	}


	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectDotnet);

	return true;
}



void NvDotnet::FreeDotnetLib(bool UseLocking)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectDotnet);
	NumInstances = 0;
	if(hLibDotnet == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectDotnet);
		return;
	}

	(DotnetLibFree)();
	MyFreeLibrary(hLibDotnet);
	DotnetLibInit = NULL;
	DotnetLibFree = NULL;
	DotnetLibRunClass = NULL;
	DotnetLibUnloadDomain = NULL;
	hLibDotnet = NULL;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectDotnet);
}



bool NvDotnet::IsDotnetLibLoaded(bool UseLocking)
{
	bool RetVal = true;
	if(UseLocking)
		NV_LOCKLOCK(&ProtectDotnet);
	if(hLibDotnet == NULL)
		RetVal = false;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectDotnet);
	return RetVal;
}


// Run a Dotnet method from .
int NvDotnet::RunMethod(const char* Method, void *pCommand)
{
	NV_LOCKLOCK(&ProtectDotnet);
	if(!IsDotnetLibLoaded(false))
	{
		NV_UNLOCKLOCK(&ProtectDotnet);
		return 0;
	}
	NumInstances++;
	NV_UNLOCKLOCK(&ProtectDotnet);
	int RetCode = 1;
	try
	{
		RetCode = (DotnetLibRunClass)(Method, pCommand);
	}
	catch(...)
	{
		RetCode = 0;
	}
	NV_LOCKLOCK(&ProtectDotnet);
	if(NumInstances > 0)
		NumInstances--;
	NV_UNLOCKLOCK(&ProtectDotnet);
	return RetCode;

}

// Unload a dotnet domain
void NvDotnet::UnloadDomain(const char* Type, const char* Name)
{
	NV_LOCKLOCK(&ProtectDotnet);
	if(!IsDotnetLibLoaded(false))
	{
		NV_UNLOCKLOCK(&ProtectDotnet);
		return;
	}
	NV_UNLOCKLOCK(&ProtectDotnet);
	try
	{
		(DotnetLibUnloadDomain)(Type, Name);
	}
	catch(...)
	{
	}
}


int NvDotnet::ThreadRunMethod(const char* Method, void *pCommand, bool SameThread)
{
	NvDotnetMethod *DotnetMethod = NVNEW NvDotnetMethod;
	if(DotnetMethod == NULL)
		return 0;

	int RetCode = 1;
	try
	{
		RetCode = DotnetMethod->RunMethod(Method, pCommand, SameThread);
	}
	catch(...)
	{
		RetCode = 0;
	}

	delete DotnetMethod;

	return RetCode;
}


NvDotnetMethod::NvDotnetMethod()
{
	b_Initialized = false;
}


int NvDotnetMethod::Main()
{
	return GetMain()->DotnetInterface.RunMethod(MethodPath, pCommandHere);
}


// called before the Main function (see class NvTask)
bool NvDotnetMethod::OnInit()
{
	NvTask::OnInit();	// call the base class

	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_DOTNET);
	
	char sThreadId[20];
	sprintf(sThreadId, "%d", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type DOTNET", NV_LOG_INFO, 1, sThreadId, "system");

	b_Initialized = true;

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type DOTNET", NV_LOG_INFO, 1, sThreadId, "system");
	return true;
}



// called after the Main function (see class NvTask)
void NvDotnetMethod::OnCleanup()
{
	char sThreadId[20];
	sprintf(sThreadId, "%d", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type DOTNET", NV_LOG_INFO, 1, sThreadId, "system");


	// to call before leaving
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type DOTNET", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}



bool NvDotnetMethod::IsDotnetAvailable()
{
	return GetMain()->DotnetInterface.IsDotnetLibLoaded();
}


// Run a dotnet method
// The dotnet method has the following format: methodassemblypath:methodname
// Typically, the methodassemblypath will the an assembly file in the procedure directory
// The method name is the name of a static method embedded in the given assembly file
// If not provided, the function will use main or Main
int NvDotnetMethod::RunMethod(const char* Method, void *pCommand, bool SameThread)
{
	if(!IsDotnetAvailable())
		return 0;
	MethodPath = Method;
	pCommandHere = pCommand;

	if(!Run(SameThread))
		return 0;

	if(!SameThread)
	{
		// Wait for end of thread but let other threads do their work
		int CheckEvery = 1;
		while(IsThreadRunning())
		{
			SLEEP(CheckEvery);

			// We first do the check very often but less and less while the time is growing
			CheckEvery += (CheckEvery/10)+1;
			if(CheckEvery > 10000)		// When the check time is 10 seconds, we check every 10 seconds
				CheckEvery = 10000;
		}
	}

	if(!m_iExitCode)
		return 0;
	return 1;
}



