// ***************************************************************************************
// File: intdb.cpp
// Description: internal proprietary record collection
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#ifdef NV_SERVER
#include "nvs.h"
#else
#ifdef WIN32
#include "..\nvc\nvcc.h"
#else
#include "nvcc.h"
#endif
#endif
#include "intdb.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

#ifdef NV_SERVER
extern bool bOldObjectsUnicode;				// Old objects unicode
#endif	//NV_SERVER

using namespace std;



// Global function for loading and saving file

	
bool WriteTag(char TagType, unsigned long TagLength, const char* TagData, FILE *stream, const char* FileName, bool RemoveFileIfError)
{
	unsigned long  nbBytes;
	if(TagType != FILE_NOTAG)
	{
		unsigned char buffer[10];
		memset(buffer, 0, sizeof(buffer));
		buffer[0] = TagType;
		int nbBytesToWrite = 3;
		if(TagLength < 32768)
		{
			buffer[1] = (char)(TagLength >> 8);
			buffer[2] = (char)(TagLength);
		}
		else
		{
			buffer[1] = 0xFF;
			buffer[2] = 0xFF;
			buffer[3] = (unsigned char)(TagLength >> 24);
			buffer[4] = (unsigned char)(TagLength >> 16);
			buffer[5] = (unsigned char)(TagLength >> 8);
			buffer[6] = (unsigned char)(TagLength);
			nbBytesToWrite += 4;
		}

		nbBytes = (unsigned long)(fwrite(buffer, 1, nbBytesToWrite, stream));
		if(nbBytes != nbBytesToWrite)
		{
			fclose(stream);
			if(RemoveFileIfError && (FileName != NULL))
				RemoveFile(FileName);
			return false;
		}
	}
	if(TagLength > 0)
	{
		if(TagType != FILE_NOTAG)
		{
			char *CopyTagData = (char *)malloc(TagLength);
			int i;
			for(i=0; i<(int)TagLength; i++)	// Very simple code
				CopyTagData[i] = ~TagData[i] + 2;
			nbBytes = (unsigned long)fwrite(CopyTagData, 1, TagLength, stream);
			free(CopyTagData);
		}
		else
			nbBytes = (unsigned long)fwrite(TagData, 1, TagLength, stream);
		if(nbBytes != TagLength)
		{
			fclose(stream);
			if(RemoveFileIfError && (FileName != NULL))
				RemoveFile(FileName);
			return false;
		}
	}

	return true;
}


bool WriteBufferTag(char TagType, unsigned long TagLength, const char* TagData, char *WriteBuffer)
{
	unsigned long  nbBytes = 0;
	if(TagType != FILE_NOTAG)
	{
		unsigned char buffer[10];
		memset(buffer, 0, sizeof(buffer));
		buffer[0] = TagType;
		int nbBytesToWrite = 3;
		if(TagLength < 32768)
		{
			buffer[1] = (char)(TagLength >> 8);
			buffer[2] = (char)(TagLength);
		}
		else
		{
			buffer[1] = 0xFF;
			buffer[2] = 0xFF;
			buffer[3] = (unsigned char)(TagLength >> 24);
			buffer[4] = (unsigned char)(TagLength >> 16);
			buffer[5] = (unsigned char)(TagLength >> 8);
			buffer[6] = (unsigned char)(TagLength);
			nbBytesToWrite += 4;
		}

		memcpy(WriteBuffer+nbBytes, buffer, nbBytesToWrite);
		nbBytes += nbBytesToWrite;
	}
	if(TagLength > 0)
	{
		if(TagType != FILE_NOTAG)
		{
			char *CopyTagData = (char *)malloc(TagLength);
			int i;
			for(i=0; i<(int)TagLength; i++)	// Very simple code
				CopyTagData[i] = ~TagData[i] + 2;
			memcpy(WriteBuffer+nbBytes, CopyTagData, TagLength);
			nbBytes += TagLength;
			free(CopyTagData);
		}
		else
			memcpy(WriteBuffer+nbBytes, TagData, TagLength);
	}

	return true;
}


// ReadNextTag
// Reads the next Tag from memory stream
// Input:
//			Stream		Pointer to memory stream (incremented by the function)
//			TagType		Pointer to a char in which the function writes the Tag type
//			TagLength	Pointer to a char in which the function writes the Tag length
//			TagData		Memory allocated by the function containing the eventual Tag data
//
// Output:
//			The function returns the number of bytes read or -1 if error (memory error)
//
int ReadNextTag(char *Stream, unsigned char *TagType, unsigned long *TagLength, char** TagData)
{
	int BytesRead = 1;
	*TagData = NULL;

	*TagType = *Stream++;
	if(*TagType == FILE_DB_END)
		return BytesRead;

	*TagLength = (unsigned long)((unsigned char)(*Stream++)) << 8;
	*TagLength += (unsigned long)((unsigned char)(*Stream++));

	BytesRead += 2;

	if(*TagLength == 0xFFFF)
	{
		// This is a big tag so we must get the real size from next 4 bytes
		*TagLength = (unsigned long)((unsigned char)(*Stream++)) << 24;
		*TagLength = (unsigned long)((unsigned char)(*Stream++)) << 16;
		*TagLength = (unsigned long)((unsigned char)(*Stream++)) << 8;
		*TagLength += (unsigned long)((unsigned char)(*Stream++));
		BytesRead += 4;
	}


	int i;

	if(*TagLength != 0)
	{
		*TagData = (char *)malloc((*TagLength) + 1);
		if((*TagData) == NULL)
			return -1;
		memcpy(*TagData, Stream, *TagLength);
		BytesRead += *TagLength;
		Stream += *TagLength;
		*(*TagData+ *TagLength) = '\0';
		for(i=0; i<(int)(*TagLength); i++)		// Very small codage
			(*TagData)[i] = ~((*TagData)[i]-2);
	}

	return BytesRead;
}


void FreeData(char* Data)
{
	if(Data != NULL)
		free(Data);
}



/////////////////////////////////////////////////////////////////////////////
// Intdbfield

Intdbfield::Intdbfield()
{
}

Intdbfield::~Intdbfield()
{
	Clear();
}

void Intdbfield::Clear()
{
	Entries.clear();
}

void Intdbfield::CopyFrom(Intdbfield* Source)
{
	Clear();
	Entries = Source->Entries;
}

string Intdbfield::GetEntry(int index, const char *Default)
{
	string RetVal = "";
	if(Default != NULL)
		RetVal = Default;
	if(index >= (int)(Entries.size()))
		return RetVal;
	if(index < 0)
		return RetVal;
	return Entries[index];
}

bool Intdbfield::SetEntry(int index, const char *Value)
{
	if(Value == NULL)
		return false;
	if(index >= (int)(Entries.size()))
		return false;
	if(index < 0)
		return false;
	Entries[index] = Value;
	
	return true;
}

bool Intdbfield::InsertEntry(const char *Value, int index)
{
	if(Value == NULL)
		return false;
	if((index != -1) && (index < 0))
		index = 0;
	if(index >= (int)(Entries.size()))
		index = -1;

	if(index == -1)
		Entries.push_back(Value);
	else
		Entries.insert(Entries.begin()+index, Value);
	
	return true;
}

bool Intdbfield::RemoveEntry(int index)
{
	if(index < 0)
		return false;
	if(index >= (int)(Entries.size()))
		return false;

	Entries.erase(Entries.begin()+index);
	
	return true;
}

bool Intdbfield::SwapEntries(int index1, int index2)
{
	if(index1 < 0)
		return false;
	if(index1 >= (int)(Entries.size()))
		return false;
	if(index2 < 0)
		return false;
	if(index2 >= (int)(Entries.size()))
		return false;

	string sProvi = Entries[index1];
	Entries[index1] = Entries[index2];
	Entries[index2] = sProvi;
	
	return true;
}

int Intdbfield::GetSize()
{
	return (int)(Entries.size());
}


bool Intdbfield::Verify(const char *Value, int Operator, bool Numeric)
{
	if(!Entries.size())
	{
		if(Operator == SEARCH_OPERATOR_NE)
		{
			if(strcmpnocase(Value, "NV_CELL_EMPTY_VALUE") != 0)
				return true;
			else
				return false;
		}
		else
		{
			if(strcmpnocase(Value, "NV_CELL_EMPTY_VALUE") == 0)
				return true;
			else
				return false;
		}
	}

	int i;
	if(Numeric)
	{
		double ValueToSearch = atof(Value);
//		int ValueToSearch = atoi(Value);
		for(i=0; i<(int)(Entries.size()); i++)
		{
//			int ValueInField = atoi(Entries[i].c_str());
			double ValueInField = atof(Entries[i].c_str());
			switch(Operator)
			{
				case SEARCH_OPERATOR_E:
					if(ValueInField == ValueToSearch)
						return true;
					break;
				case SEARCH_OPERATOR_NE:
					if(ValueInField == ValueToSearch)
						return false;
					break;
				case SEARCH_OPERATOR_ST:
					if(ValueInField < ValueToSearch)
						return true;
					break;
				case SEARCH_OPERATOR_GT:
					if(ValueInField > ValueToSearch)
						return true;
					break;
				case SEARCH_OPERATOR_STE:
					if(ValueInField <= ValueToSearch)
						return true;
					break;
				case SEARCH_OPERATOR_GTE:
					if(ValueInField >= ValueToSearch)
						return true;
					break;
			}
		}
		if(Operator == SEARCH_OPERATOR_NE)
			return true;
	}
	else
	{
		string ValueToSearch = Value;
		bool WithEndWildcard = false;
		bool WithStartWildcard = false;
		string::size_type index;
		if((ValueToSearch.size()) && (index = ValueToSearch.find('*')) == 0)
		{
			ValueToSearch = ValueToSearch.substr(1);
			WithStartWildcard = true;
		}
		if((ValueToSearch.size()) && (index = ValueToSearch.find('*')) == (ValueToSearch.size()-1))
		{
			ValueToSearch = ValueToSearch.substr(0, index);
			WithEndWildcard = true;
		}

		for(i=0; i<(int)Entries.size(); i++)
		{
			string ValueInField = Entries[i];
			bool ComparaisonResult = false;
			if(WithEndWildcard)
			{
				char *buf = strstrnocase(ValueInField.c_str(), ValueToSearch.c_str());
				if(WithStartWildcard)
				{
					if(buf != NULL)
						ComparaisonResult = true;
				}
				else
				{
					if(buf == ValueInField.c_str())
						ComparaisonResult = true;
				}
			}
			if(!ComparaisonResult)
			{
				if(WithStartWildcard)
				{
					char *buf = strstrnocase(ValueInField.c_str(), ValueToSearch.c_str());
					if(buf != NULL)
					{
						// something has been found
						// we must verify that it's at the end
						if((buf+ValueToSearch.length()) == (ValueInField.c_str()+ValueInField.length()))
							ComparaisonResult = true;
					}
				}
				else
				{
					if(strcmpnocase(ValueInField.c_str(), ValueToSearch.c_str()) == 0)
						ComparaisonResult = true;
				}
			}

			switch(Operator)
			{
				case SEARCH_OPERATOR_E:
					if(ComparaisonResult)
						return true;
					break;
				case SEARCH_OPERATOR_NE:
					if(ComparaisonResult)
						return false;
					break;
			}
		}
		if(Operator == SEARCH_OPERATOR_NE)
			return true;
	}
	return false;
}



void Intdbfield::Sort(bool Numeric, bool Ascending)
{
	int num;
	

	num = (int)Entries.size();
	if(num <= 1)
		return;

	// Do a Quick sort
	unsigned int stack_pointer = 0;
	int first_stack[32];
	int last_stack[32];
	int Insertion_Sort_Bound = 30;
	int last = num-1;
	int first = 0;

	for (;;)
	{
		if (last - first <= Insertion_Sort_Bound)
		{
			/* for small sort, use insertion sort */
			unsigned int indx;
			string prev_val = Entries[first];
			string cur_val;

			for (indx = (unsigned int)first + 1; indx <= (unsigned int)last; ++indx)
			{
				cur_val = Entries[indx];
				if (SortFunc(prev_val, cur_val, Numeric, Ascending) > 0)
				{
					/* out of order: array[indx-1] > array[indx] */
					unsigned int indx2;
					Entries[indx] = prev_val; /* move up the larger item first */

					/* find the insertion point for the smaller item */
					for (indx2 = indx - 1; indx2 > (unsigned int)first; )
					{
						string temp_val = Entries[indx2 - 1];
						if (SortFunc(temp_val, cur_val, Numeric, Ascending) > 0)
						{
							Entries[indx2--] = temp_val;
							/* still out of order, move up 1 slot to make room */
						}
						else
							break;
					}
					Entries[indx2] = cur_val; /* insert the smaller item right here */
				}
				else
				{
					/* in order, advance to next element */
					prev_val = cur_val;
				}
			}
		}
		else
		{
			string pivot;
		
			/* try quick sort */
			{
				string temp;
				unsigned int med = (first + last) >> 1;
				/* Choose pivot from first, last, and median position. */
				/* Sort the three elements. */
				temp = Entries[first];
				if (SortFunc(temp, Entries[last], Numeric, Ascending) > 0)
				{
					Entries[first] = Entries[last]; Entries[last] = temp;
				}
				temp = Entries[med];
				if (SortFunc(Entries[first], temp, Numeric, Ascending) > 0)
				{
					Entries[med] = Entries[first]; Entries[first] = temp;
				}
				temp = Entries[last];
				if (SortFunc(Entries[med], temp, Numeric, Ascending) > 0)
				{
					Entries[last] = Entries[med]; Entries[med] = temp;
				}
				pivot = Entries[med];
			}
			{
				unsigned int up;
				{
					unsigned int down;
					/* First and last element will be loop stopper. */
					/* Split array into two partitions. */
					down = first;
					up = last;
					for (;;)
					{
						do
						{
							++down;
						} while (SortFunc(pivot, Entries[down], Numeric, Ascending) > 0);
					
						do
						{
							--up;
						} while (SortFunc(Entries[up], pivot, Numeric, Ascending) > 0);
					
						if (up > down)
						{
							string temp;
							/* interchange L[down] and L[up] */
							temp = Entries[down]; Entries[down]= Entries[up]; Entries[up] = temp;
						}
						else
							break;
					}
				}
				{
					unsigned int len1; /* length of first segment */
					unsigned int len2; /* length of second segment */
					len1 = up - first + 1;
					len2 = last - up;
					/* stack the partition that is larger */
					if (len1 >= len2)
					{
						first_stack[stack_pointer] = first;
						last_stack[stack_pointer++] = up;
					
						first = up + 1;
						/*  tail recursion elimination of
						*  Qsort(This,fun_ptr,up + 1,last)
						*/
					}
					else
					{
						first_stack[stack_pointer] = up + 1;
						last_stack[stack_pointer++] = last;

						last = up;
						/* tail recursion elimination of
						* Qsort(This,fun_ptr,first,up)
						*/
					}
				}
				continue;
			}
		/* end of quick sort */
		}
		if (stack_pointer > 0)
		{
			/* Sort segment from stack. */
			first = first_stack[--stack_pointer];
			last = last_stack[stack_pointer];
		}
		else
			break;
	} /* end for */
}

int Intdbfield::SortFunc(string& a, string& b, bool Numeric, bool Ascending)
{
	if(Numeric)
	{
		double Value1 = atof(a.c_str());
		double Value2 = atof(b.c_str());
		if(Ascending)
		{
			if(Value1 > Value2)
				return 1;
			else if(Value1 < Value2)
				return -1;
			else
				return 0;
		}
		else
		{
			if(Value1 < Value2)
				return 1;
			else if(Value1 > Value2)
				return -1;
			else
				return 0;
		}
	}
	else
	{
		int result = strcmpnocase(a.c_str(), b.c_str());
		if(Ascending)
		{
			if(result > 0)
				return 1;
			else if(result <0)
				return -1;
			else
				return 0;
		}
		else
		{
			if(result < 0)
				return 1;
			else if(result > 0)
				return -1;
			else
				return 0;
		}
	}
}


bool Intdbfield::IsFromField(Intdbfield *Source, bool Remove)
{

	if(Source == NULL)
		return false;

	bool Found = false;
	bool RetCode = true;

	Intdbfield NewSecuField;

	for(int i=0; i<(int)Entries.size(); i++)
	{
		Found = false;
		string Entry = GetEntry(i);
		for(int j=0; j<(int)(Source->Entries.size()); j++)
		{
			if(strcmpnocase(Entry.c_str(), Source->GetEntry(j).c_str()) == 0)
			{
				Found = true;
				if(Remove)
					NewSecuField.Entries.push_back(Entry);
				break;
			}
		}
		if(!Found)
		{
			if(!Remove)
				return false;
			RetCode = false;
		}
	}

	if(!RetCode && Remove)
		CopyFrom(&NewSecuField);

	return RetCode;
}


/////////////////////////////////////////////////////////////////////////////
// Intdbfielddesc

Intdbfielddesc::Intdbfielddesc()
{
	Clear();
}

Intdbfielddesc::~Intdbfielddesc()
{
	Clear();
}

void Intdbfielddesc::Clear()
{
	Name = "";
	Description = "";
	Type = FIELD_TYPE_ALPHA;
}

void Intdbfielddesc::CopyFrom(Intdbfielddesc* Source)
{
	Name = Source->Name;
	Description =  Source->Description;
	Type =  Source->Type;
}



/////////////////////////////////////////////////////////////////////////////
// Intdbrecord

Intdbrecord::Intdbrecord()
{
}

Intdbrecord::~Intdbrecord()
{
	Clear();
}

void Intdbrecord::Clear()
{
	int i;
	Intdbfield *SecuField;
	for(i=0; i<(int)Fields.size(); i++)
	{
		SecuField = (Intdbfield*)(Fields[i]);
		delete SecuField;
	}
	Fields.clear();
}

void Intdbrecord::ClearFieldData()
{
	int i;
	Intdbfield *SecuField;
	for(i=0; i<(int)Fields.size(); i++)
	{
		SecuField = (Intdbfield*)(Fields[i]);
		SecuField->Clear();
	}
}

void Intdbrecord::RemoveField(int index)
{
	if(index >= (int)Fields.size())
		return;
	if(index < 0)
		return;

	Intdbfield *SecuField = (Intdbfield*)(Fields[index]);
	if(SecuField != NULL)
		delete SecuField;
	Fields.erase(Fields.begin() + index);
}





void Intdbrecord::InsertField(int index)
{

	if(index >= (int)Fields.size())
		index = -1;

	if((index != -1) && (index < 0))
		index = 0;

	Intdbfield *SecuField = NVNEW Intdbfield;
	if(SecuField == NULL)
		return;

	if(index == -1)
		Fields.push_back((void *)SecuField);
	else
		Fields.insert(Fields.begin()+index, (void *)SecuField);
}





void Intdbrecord::SwapFields(int index1, int index2)
{
	if(index1 < 0)
		return;
	if(index2 < 0)
		return;

	if(index1 >= (int)Fields.size())
		return;
	if(index2 >= (int)Fields.size())
		return;

	void *Field = Fields[index1];
	Fields[index1] = Fields[index2];
	Fields[index2] = Field;
}


void Intdbrecord::CopyFrom(Intdbrecord* Source)
{
	Clear();
	int i;
	Intdbfield *SecuField;
	Intdbfield *NewSecuField;
	for(i=0; i<(int)(Source->Fields.size()); i++)
	{
		NewSecuField = NVNEW Intdbfield;
		if(NewSecuField == NULL)
			break;
		SecuField = (Intdbfield*)(Source->Fields[i]);
		NewSecuField->CopyFrom(SecuField);
		Fields.push_back((void *)NewSecuField);
	}
}

Intdbfield* Intdbrecord::GetField(int index)
{
	if(index < 0)
		return NULL;
	if(index >= (int)Fields.size())
		return NULL;
	return (Intdbfield*)(Fields[index]);
}


int Intdbrecord::GetSize()
{
	return (int)Fields.size();
}



/////////////////////////////////////////////////////////////////////////////
// Intdbtable

Intdbtable::Intdbtable()
{
	Name = "";
	Description = "";
	PrimaryName = "";
	PrimaryCaseSensitive = false;
}

Intdbtable::~Intdbtable()
{
	Clear();
}

void Intdbtable::Clear()
{
	Name = "";
	Description = "";
	PrimaryName = "";
	PrimaryCaseSensitive = false;
	Primary.clear();

	int i;
	Intdbfielddesc *SecuFieldDesc;
	for(i=0; i<(int)FieldDescriptions.size(); i++)
	{
		SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[i]);
		delete SecuFieldDesc;
	}
	FieldDescriptions.clear();

	Intdbrecord *Record;
	for(i=0; i<(int)Records.size(); i++)
	{
		Record = (Intdbrecord*)(Records[i]);
		delete Record;
	}
	Records.clear();
}

void Intdbtable::ClearData(bool OnlyFieldContent)
{
	Intdbrecord *Record;
	int i;
	for(i=0; i<(int)Records.size(); i++)
	{
		Record = (Intdbrecord*)(Records[i]);
		if(OnlyFieldContent)
			Record->ClearFieldData();
		else
			delete Record;
	}
	if(!OnlyFieldContent)
		Records.clear();
	Primary.clear();
}

void Intdbtable::CopyFrom(Intdbtable* Source, bool WithData, int MaxDocs)
{
	Clear();
	Name = Source->Name;
	Description = Source->Description;
	PrimaryName = Source->PrimaryName;
	Primary = Source->Primary;
	PrimaryCaseSensitive = Source->PrimaryCaseSensitive;

	int i;
	Intdbfielddesc *SecuFieldDesc;
	Intdbfielddesc *NewSecuFieldDesc;
	for(i=0; i<(int)(Source->FieldDescriptions.size()); i++)
	{
		NewSecuFieldDesc = NVNEW Intdbfielddesc;
		if(NewSecuFieldDesc == NULL)
			return;
		SecuFieldDesc = (Intdbfielddesc*)(Source->FieldDescriptions[i]);
		NewSecuFieldDesc->CopyFrom(SecuFieldDesc);
		FieldDescriptions.push_back((void *)NewSecuFieldDesc);
	}
	
	
	if(!WithData)
		return;

	Intdbrecord *Record;
	Intdbrecord *NewRecord;
	int NumRecords = (int)(Source->Records.size());
	if((MaxDocs >= 0) && (MaxDocs < NumRecords))
		NumRecords = MaxDocs;
	for(i=0; i<(int)(NumRecords); i++)
	{
		NewRecord = NVNEW Intdbrecord;
		if(NewRecord == NULL)
			return;
		Record = (Intdbrecord*)(Source->Records[i]);
		NewRecord->CopyFrom(Record);
		Records.push_back((void *)NewRecord);
	}

	// Set the hash table if necessary
	SetPrimaryHashTable();
}

bool Intdbtable::SelectFrom(Intdbtable* Source, int source_field_index, int dest_field_index, int selected_field_index, const char* SelectedYes, const char* SelectedNo, bool CaseSensitive)
{
	if(Source == NULL)
		return false;

	if(source_field_index >= Source->GetNumFields())
		return false;

	if(selected_field_index >= GetNumFields())
		return false;

	if(dest_field_index >= GetNumFields())
		return false;

	if(source_field_index < 0)
		return false;

	if(selected_field_index < 0)
		return false;

	if(dest_field_index < 0)
		return false;

	int i,j,k,l;
	Intdbrecord *Record;
	Intdbrecord *SourceRecord;
	Intdbfield *SelectedField;

	// PM PRIMARY
	if(IsPrimaryColumn(selected_field_index))
		return false;	// The selected field cannot be the primary key

	Intdbfield *Field;
	Intdbfield *SourceField;
	string SourceValue;
	string Value;
	bool Found = false;
	for(i=0; i<(int)Records.size(); i++)
	{
		Record = (Intdbrecord*)(Records[i]);
		Field = Record->GetField(dest_field_index);
		SelectedField = Record->GetField(selected_field_index);
		Found = false;
		for(j=0; j<Field->GetSize(); j++)
		{
			string Value = Field->GetEntry(j);
			if(!CaseSensitive)
				MakeUpper(Value);
			for(k=0; k<(int)(Source->Records.size()); k++)
			{
				SourceRecord = (Intdbrecord*)(Source->Records[k]);
				SourceField = SourceRecord->GetField(source_field_index);
				for(l=0; l<SourceField->GetSize(); l++)
				{
					string SourceValue = SourceField->GetEntry(l);
					if(!CaseSensitive)
						MakeUpper(SourceValue);
					if(Value == SourceValue)
					{
						Found = true;
						break;
					}
				}
				if(Found)
					break;
			}
			if(Found)
				break;
		}
		if(SelectedField != NULL)
		{
			SelectedField->Clear();
			if(Found)
			{
				if((SelectedYes != NULL) && (strlen(SelectedYes)))
					SelectedField->InsertEntry(SelectedYes);
			}
			else
			{
				if((SelectedNo != NULL) && (strlen(SelectedNo)))
					SelectedField->InsertEntry(SelectedNo);
			}
		}
	}

	return true;
}

bool Intdbtable::Join(Intdbtable* Source, const char* Columns, const char* Foreign, bool Replace)
{
	if(Source == NULL)
		return false;
	if(Foreign == NULL)
		return false;

	// First split Foreign column
	string SrcForeign = Foreign;
	string DstForeign = Foreign;
	string::size_type index = SrcForeign.find(':');
	if(index != string::npos)
	{
		SrcForeign = SrcForeign.substr(0, index);
		DstForeign = DstForeign.substr(index+1);
	}
	CleanString(SrcForeign);
	CleanString(DstForeign);

	int index_src_foreign = Source->GetFieldIndex(SrcForeign.c_str());
	int index_dst_foreign = GetFieldIndex(DstForeign.c_str());
	if(index_src_foreign == -1)
		return false;
	if(index_dst_foreign == -1)
		return false;

	vector<int> src_columns;
	vector<int> dst_columns;

	string ColumnsHere = "";
	if(Columns != NULL)
		ColumnsHere = Columns;
	CleanString(ColumnsHere, true);

	int i,j,k;
	int src_index,dst_index;
	int NumColumnDst = GetNumFields();
	if(!ColumnsHere.length())
	{
		// all columns of the source
		int NumColumnSrc = Source->GetNumFields();
		for(i=0; i<NumColumnSrc; i++)
		{
			string FieldName = Source->GetFieldName(i);
			dst_index = GetFieldIndex(FieldName.c_str());
			if((dst_index != -1) && !Replace)
				continue;
			src_columns.push_back(i);
			if(dst_index == -1)
			{
				dst_index = NumColumnDst;
				string FieldDesc = Source->GetFieldDesc(i);
				bool bNumeric = Source->IsNumericField(i);
				InsertField(FieldName.c_str(), FieldDesc.c_str(), bNumeric); 
				NumColumnDst++;
			}
			dst_columns.push_back(dst_index);
		}
	}
	else
	{
		// Only selected columns
		while(ColumnsHere.length())
		{
			string srcColumn = ColumnsHere;
			index = srcColumn.find(';');
			if(index != string::npos)
			{
				srcColumn = ColumnsHere.substr(0, index);
				ColumnsHere = ColumnsHere.substr(index+1);
			}
			else
				ColumnsHere = "";
			CleanString(srcColumn);
			string dstColumn = srcColumn;
			index = srcColumn.find(':');
			if(index != string::npos)
			{
				srcColumn = srcColumn.substr(0, index);
				dstColumn = dstColumn.substr(index+1);
			}
			CleanString(srcColumn);
			CleanString(dstColumn);
			src_index = Source->GetFieldIndex(srcColumn.c_str());
			if(src_index == -1)
				continue;	// the source column doesn't exist
			dst_index = GetFieldIndex(dstColumn.c_str());
			if((dst_index != -1) && !Replace)
				continue;
			src_columns.push_back(src_index);
			if(dst_index == -1)
			{
				dst_index = NumColumnDst;
				string FieldDesc = Source->GetFieldDesc(src_index);
				bool bNumeric = Source->IsNumericField(src_index);
				InsertField(dstColumn.c_str(), FieldDesc.c_str(), bNumeric); 
				NumColumnDst++;
			}
			dst_columns.push_back(dst_index);
		}
	}

	if(!src_columns.size())
		return true;	// nothing to do
	int NumRows = GetSize();
	int srcNumRows = Source->GetSize();
	Intdbrecord *Record;
	Intdbrecord *SourceRecord;
	Intdbfield *Field;
	Intdbfield *SourceField;
	bool bIsPrimary = Source->IsPrimaryColumn(index_src_foreign);
	for(i=0;i<NumRows;i++)
	{
		Record = (Intdbrecord*)(Records[i]);
		Field = Record->GetField(index_dst_foreign);
		if(Field == NULL)
			continue;
		string SearchValue = Field->GetEntry(0);
		if(!SearchValue.length())
			continue;
		src_index = -1;
		SourceRecord = NULL;
		if(bIsPrimary)
		{
			src_index = Source->GetPrimary(SearchValue);
			if(src_index != -1)
				SourceRecord = (Intdbrecord*)(Source->Records[src_index]);
		}
		else
		{
			for(j=0; j<srcNumRows; j++)
			{
				SourceRecord = (Intdbrecord*)(Source->Records[j]);
				SourceField = SourceRecord->GetField(index_src_foreign);
				if(SourceField == NULL)
					continue;
				for(k=0; k<SourceField->GetSize(); k++)
				{
					string SourceValue = SourceField->GetEntry(k);
					if(SearchValue == SourceValue)
					{
						src_index = j;
						break;
					}
				}
				if(src_index != -1)
					break;
			}
		}
		if(SourceRecord == NULL)
			src_index = -1;	// should not occur

		// next step is to copy record data from source to destination
		int NumFields = (int)src_columns.size();
		for(j=0; j<NumFields; j++)
		{
			Field = Record->GetField(dst_columns[j]);
			if(Field == NULL)
				continue;
			if(src_index == -1)
			{
				Field->Clear();
			}
			else
			{
				SourceField = SourceRecord->GetField(src_columns[j]);
				if(SourceField == NULL)
					continue;
				Field->CopyFrom(SourceField);
			}
		}
	}

	return true;
}


bool Intdbtable::Add(Intdbtable* Source, int Mode)
{
	if(Source == NULL)
		return false;
	
	int NumColumns = GetNumFields();
	// Both source and destination must have the same number of columns
	if(GetNumFields() != Source->GetNumFields())
		return false;

	Intdbrecord *Record;
	Intdbrecord *NewRecord;
	Intdbfield *Field;
	string HashValue;
	int i;
	map<string, int> NewPrimary;				// New Primary key hash table

	int PrimaryFieldIndex = -1;
	if(PrimaryName.length())
		PrimaryFieldIndex = GetFieldIndex(PrimaryName.c_str());
	if(PrimaryFieldIndex == -1)
	{
		// No primary field
		for(i=0; i<(int)(Source->Records.size()); i++)
		{
			NewRecord = NVNEW Intdbrecord;
			if(NewRecord == NULL)
				return false;
			Record = (Intdbrecord*)(Source->Records[i]);
			NewRecord->CopyFrom(Record);
			Records.push_back((void *)NewRecord);
		}
		return true;
	}

	// The table has primary field so we must check for mode
	if(Mode == ADD_MODE_NORMAL)
	{
		// We must check doublons
		for(i=0; i<(int)(Source->Records.size()); i++)
		{
			Record = (Intdbrecord*)(Source->Records[i]);
			Field = (Intdbfield*)(Record->Fields[PrimaryFieldIndex]);
			if(Field == NULL)
				continue;	// Should not occur
			if(Field->Entries.size() > 1)
			{
				// This is an error
				// the primary key can only have one line
				return false;
			}
			HashValue = GetHashValue(Field->GetEntry(0));
			if(!HashValue.length())
				continue; // should not occur
			if(Primary.find(HashValue) != Primary.end())
			{
				// The entry ever exists
				// This is an error
				return false;
			}
			if(NewPrimary.find(HashValue) != NewPrimary.end())
			{
				// The entry ever exists
				// This is an error
				return false;
			}
			NewPrimary[HashValue] = i;
		}
		// All the records have been checked and there is no doublons
		// SO we can add the new records
		for(i=0; i<(int)(Source->Records.size()); i++)
		{
			NewRecord = NVNEW Intdbrecord;
			if(NewRecord == NULL)
				return false;
			Record = (Intdbrecord*)(Source->Records[i]);
			NewRecord->CopyFrom(Record);
			Records.push_back((void *)NewRecord);
		}

		// Set the hash table
		SetPrimaryHashTable();
	}
	else
	{
		// We process one record at a time
		for(i=0; i<(int)(Source->Records.size()); i++)
		{
			Record = (Intdbrecord*)(Source->Records[i]);
			Field = (Intdbfield*)(Record->Fields[PrimaryFieldIndex]);
			if(Field == NULL)
				continue;	// Should not occur
			if(Field->Entries.size() > 1)
			{
				// This is an error
				// the primary key can only have one line (the result maybe partially modified in this case)
				return false;
			}
			string PrimaryValue = Field->GetEntry(0);
			int RecordIndex = GetPrimary(PrimaryValue);
			if(RecordIndex < 0)
			{
				// No primary entry found
				// so we can add the record
				NewRecord = NVNEW Intdbrecord;
				if(NewRecord == NULL)
					return false;
				Record = (Intdbrecord*)(Source->Records[i]);
				NewRecord->CopyFrom(Record);
				Records.push_back((void *)NewRecord);
				// also update primary map
				SetPrimary(PrimaryValue, (int)Records.size()-1);
			}
			else
			{
				// The entry ever exists
				if(Mode == ADD_MODE_REPLACE)
				{
					// We must replace the record
					NewRecord = (Intdbrecord*)(Records[RecordIndex]);
					Record = (Intdbrecord*)(Source->Records[i]);
					NewRecord->CopyFrom(Record);
				}
				// For the ADD_MODE_NOREPLACE we just skip the record
			}
		}
	}

	return true;
}


// Used internally to build the primary key hash table
bool Intdbtable::SetPrimaryHashTable(bool bRemoveDoubleEntries)
{
	Primary.clear();
	if(!PrimaryName.length())
		return true;
	int FieldIndex = GetFieldIndex(PrimaryName.c_str());
	if(FieldIndex == -1)
		return true;	// Primary field is not valid

	int i;
	Intdbrecord *Record;
	Intdbfield *Field;
	string HashValue;
	for(i=0; i<(int)Records.size(); i++)
	{
		Record = (Intdbrecord*)(Records[i]);
		Field = (Intdbfield*)(Record->Fields[FieldIndex]);
		if(Field == NULL)
			continue;	// Should not occur
		if(Field->Entries.size() > 1)
		{
			// This is an error
			// the primary key can only have one line
			Primary.clear();
			return false;
		}
		if(Field->Entries.size() == 0)
			continue;	// do nothing if the field is empty
		HashValue = GetHashValue(Field->GetEntry(0));
		if(!HashValue.length())
			continue;
		if(Primary.find(HashValue) != Primary.end())
		{
			// The entry ever exists
			// This is an error
			if(bRemoveDoubleEntries)
			{
				delete Record;
				Records.erase(Records.begin() + i);
				i--;
				continue;
			}
			else
			{
				Primary.clear();
				return false;
			}
		}
		Primary[HashValue] = i;
	}

	return true;
}

// Used internally to compute primary key hash value
string Intdbtable::GetHashValue(string Value)
{
	string RetVal = "";
	CleanString(Value);
	if(!PrimaryCaseSensitive)
		MakeUpper(Value);
	//if(Value.length())
	//{
		RetVal = MD5_string(Value);
		MakeUpper(RetVal);
	//}
	return RetVal;
}

int Intdbtable::GetPrimary(string Value)
{
	int RetVal = -1;
	if(!PrimaryName.length())
		return RetVal;
	string HashValue = GetHashValue(Value);
	if(!HashValue.length())
		return RetVal;
	if(Primary.find(HashValue) != Primary.end())
	{
		int RecordIndex = Primary[HashValue];
		if(RecordIndex >= 0)
			if(RecordIndex < (int)Records.size())
				RetVal = RecordIndex;
	}

	return RetVal;
}

bool Intdbtable::IsPrimaryColumn(int col_index)
{
	if(!PrimaryName.length())
		return false;
	
	string FieldName = GetFieldName(col_index);

	if(strcmpnocase(FieldName.c_str(), PrimaryName.c_str()) == 0)
		return true;

	return false;
}


void Intdbtable::SetPrimary(string Value, int index)
{
	if(!PrimaryName.length())
		return;
	string HashValue = GetHashValue(Value);
	if(!HashValue.length())
		return;
	Primary[HashValue] = index;
}


void Intdbtable::RemovePrimary(string Value)
{
	if(!PrimaryName.length())
		return;
//	if(!Value.length())
//		return;
	string HashValue = GetHashValue(Value);
	if(!HashValue.length())
		return;
	if(Primary.find(HashValue) != Primary.end())
		Primary.erase(HashValue);
}



bool Intdbtable::SetPrimaryField(string FieldName, bool CaseSensitive, bool bRemoveDoubleEntries)
{
	CleanString(FieldName);
	if(!FieldName.length())
	{
		// The request is to remove the primary key
		Primary.clear();
		CaseSensitive = false;
		PrimaryName = "";
		return true;
	}
	int FieldIndex = GetFieldIndex(FieldName.c_str());
	if(FieldIndex == -1)
		return false;	// New primary field is not valid

	// We save the old primary in case of problem
	string OldPrimaryName = PrimaryName;
	bool OldPrimaryCaseSensitive = PrimaryCaseSensitive;
	map<std::string, int> OldPrimary;
	OldPrimary = Primary;	
	
	PrimaryName = FieldName;
	PrimaryCaseSensitive = CaseSensitive;

	if(!SetPrimaryHashTable(bRemoveDoubleEntries))
	{
		PrimaryName = OldPrimaryName;
		PrimaryCaseSensitive = OldPrimaryCaseSensitive;
		Primary = OldPrimary;
		return false;
	}

	return true;
}


string Intdbtable::GetPrimaryField()
{
	return PrimaryName;
}



string Intdbtable::GetRecordPrimaryValue(int index)
{
	string RetVal = "";
	if(!PrimaryName.length())
		return RetVal;
	Intdbrecord *Record = GetRecord(index);
	if(Record == NULL)
		return RetVal;

	int FieldIndex = GetFieldIndex(PrimaryName.c_str());
	if(FieldIndex == -1)
		return RetVal;	// The primary field is not valid

	RetVal = GetCellLine(index, FieldIndex);

	return RetVal;
}



Intdbrecord* Intdbtable::GetRecord(int index)
{
	if(index >= (int)Records.size())
		return NULL;
	if(index < 0)
		return NULL;
	return (Intdbrecord*)(Records[index]);
}



Intdbfield* Intdbtable::GetField(int record_index, int field_index)
{
	Intdbrecord *Record = GetRecord(record_index);
	if(Record == NULL)
		return NULL;


	Intdbfield* Field = Record->GetField(field_index);
	return Field;
}


int Intdbtable::GetFieldIndex(const char* FieldName, int StartAt)
{
	if(FieldName == NULL)
		return -1;
	if(!strlen(FieldName))
		return -1;

	int i;
	for(i=StartAt; i<(int)FieldDescriptions.size(); i++)
	{
		Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[i]);;
		if(strcmpnocase(FieldName, SecuFieldDesc->Name.c_str()) == 0)
			return i;
	}
	return -1;
}


string Intdbtable::GetFieldName(int FieldIndex)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return "";

	if(FieldIndex < 0)
		return "";

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if(SecuFieldDesc != NULL)
		return SecuFieldDesc->Name;

	return "";
}


bool Intdbtable::SetFieldName(int FieldIndex, string FieldName)
{
	CleanString(FieldName, true);
	if(!FieldName.length())
		return false;
	if(FieldIndex >= (int)FieldDescriptions.size())
		return false;

	if(FieldIndex < 0)
		return false;

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if(SecuFieldDesc != NULL)
	{
		int FieldIndexExist = GetFieldIndex(FieldName.c_str());
		if(FieldIndexExist != -1)
		{
			// The field ever exists
			if(FieldIndexExist == FieldIndex)
				return true; // ever has the given name
			else
				return false;
		}
		bool bIsPrimary = IsPrimaryColumn(FieldIndex);
		SecuFieldDesc->Name = FieldName;
		if(bIsPrimary)
			PrimaryName = FieldName;
	}
	return true;
}


void Intdbtable::SetFieldDesc(int FieldIndex, string FieldDesc)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return;

	if(FieldIndex < 0)
		return;

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if(SecuFieldDesc != NULL)
		SecuFieldDesc->Description = FieldDesc;
}


string Intdbtable::GetFieldDesc(int FieldIndex)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return "";

	if(FieldIndex < 0)
		return "";

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if(SecuFieldDesc != NULL)
		return SecuFieldDesc->Description;

	return "";
}



void Intdbtable::SetFieldNumeric(int FieldIndex, bool Numeric)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return;

	if(FieldIndex < 0)
		return;

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if(SecuFieldDesc != NULL)
	{
		if(Numeric)
			SecuFieldDesc->Type = FIELD_TYPE_NUMERIC;
		else
			SecuFieldDesc->Type = FIELD_TYPE_ALPHA;
	}
}

bool Intdbtable::IsNumericField(int FieldIndex)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return false;

	if(FieldIndex < 0)
		return false;

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);

	if((SecuFieldDesc != NULL) && (SecuFieldDesc->Type == FIELD_TYPE_NUMERIC))
		return true;

	return false;
}


int Intdbtable::GetSize()
{
	return (int)Records.size();
}


int Intdbtable::GetNumFields()
{
	return (int)FieldDescriptions.size();
}


int Intdbtable::Info(Intdbtable **Result)
{
	*Result = NVNEW Intdbtable;
	if(*Result == NULL)
		return INTDB_MEMORYERROR;

	(*Result)->CopyFrom(this, false);
	return 0;
}



void Intdbtable::RemoveRecord(int index)
{
	int numRecords = GetSize();
	if(!numRecords)
		return;
	if(index >= numRecords)
		return;

	Intdbrecord *Record = GetRecord(index);
	if(Record != NULL)
		delete Record;

	Records.erase(Records.begin() + index);

	SetPrimaryHashTable();
}


void Intdbtable::ClearRecord(int index)
{
	if(index == -1)
		index = GetSize() - 1;

	Intdbrecord* Record = GetRecord(index);
	if(Record != NULL)
	{
		if(PrimaryName.length())	// Eventually remove entry from hash table
			RemovePrimary(GetRecordPrimaryValue(index));
		Record->ClearFieldData();
	}
}



void Intdbtable::RemoveField(int FieldIndex)
{
	if(FieldIndex >= (int)FieldDescriptions.size())
		return;

	if(FieldIndex < 0)
		return;

	Intdbfielddesc *SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[FieldIndex]);
	if(SecuFieldDesc == NULL)
		return;

	int i;
	int numRecords = GetSize();
	for (i=0; i<numRecords; i++)
	{
		Intdbrecord *Record = GetRecord(i);
		if(Record == NULL)
			continue;
		if(IsPrimaryColumn(FieldIndex))	// Eventually remove the hash table
		{
			PrimaryName = "";
			SetPrimaryHashTable();
		}
		Record->RemoveField(FieldIndex);
	}

	delete SecuFieldDesc;
	FieldDescriptions.erase(FieldDescriptions.begin() + FieldIndex);
}




int Intdbtable::RemoveRecords(const char* Query)
{
	int numRecords = GetSize();
	if(!numRecords)
		return 0;	// nothing to do

	if(Query == NULL)
		return INTDB_INVALIDQUERY;

	int Result = Search(Query, NULL, NULL, SEARCH_MODE_REMOVE);

	// Rebuild the hash table if successful
	if(Result == 0)
		SetPrimaryHashTable();

	return Result;
}




int Intdbtable::ClearDataRecords(const char* Query)
{
	int numRecords = GetSize();
	if(!numRecords)
		return 0;	// nothing to do

	if(Query == NULL)
		return INTDB_INVALIDQUERY;

	int Result = Search(Query, NULL, NULL, SEARCH_MODE_CLEAR);

	// Rebuild the hash table if successful
	if(Result == 0)
		SetPrimaryHashTable();

	return Result;
}


Intdbrecord* Intdbtable::CreateEmptyRecord()
{
	Intdbrecord *Record = NVNEW Intdbrecord;
	if(Record == NULL)
		return NULL;

	int i;
	for(i=0; i<(int)FieldDescriptions.size(); i++)
	{
		Intdbfield *Field = NVNEW Intdbfield;
		if(Field == NULL)
		{
			delete Record;
			return NULL;
		}
		Record->Fields.push_back((void *)Field);
	}

	Records.push_back((void *)Record);

	return Record;
}


void Intdbtable::InsertRecords(int num_records, int record_index)
{
	if(num_records <= 0)
		return;

	if((record_index != -1) && (record_index >= (int)Records.size()))
		record_index = -1;

	if((record_index != -1) && (record_index < 0))
		record_index = 0;


	int i;
	for(i=0; i<num_records; i++)
	{

		Intdbrecord *Record = NVNEW Intdbrecord;
		if(Record == NULL)
			return;

		int j;
		for(j=0; j<(int)FieldDescriptions.size(); j++)
		{
			Intdbfield *Field = NVNEW Intdbfield;
			if(Field == NULL)
			{
				delete Record;
				return;
			}
			Record->Fields.push_back((void *)Field);
		}

		if(record_index == -1)
			Records.push_back((void *)Record);
		else
			Records.insert(Records.begin()+record_index, (void *)Record);
	}
	if(record_index != -1)
		SetPrimaryHashTable();	// Rebuild the primary index because the record indexes have changed if records are not added at the end 
}





bool Intdbtable::SetRecordData(int index, const char* RecordData, const char* FieldSeparator, const char* EntrySeparator)
{
	if(index == -1)
		index = GetSize() - 1;

	Intdbrecord *Record = GetRecord(index);
	if(Record == NULL)
		return false;

	if(FieldSeparator == NULL)
		return false;
	if(strlen(FieldSeparator) == 0)
		return false;

	if(EntrySeparator == NULL)
		return false;
	if(strlen(EntrySeparator) == 0)
		return false;

	string line = "";
	if(RecordData != NULL)
		line = RecordData;

	int i;
	Intdbfield *SecuField = NULL;
	int FieldSeparatorLength = (int)strlen(FieldSeparator);
	int EntrySeparatorLength = (int)strlen(EntrySeparator);
	string PrimaryValue = "";

	Intdbrecord *SecuRecord = NVNEW Intdbrecord;
	if(SecuRecord == NULL)
		return false;
	vector<string> FieldsInLine;
	string ActualField;
	string::size_type index1;
	while(FieldsInLine.size() < FieldDescriptions.size())
	{
		ActualField = line;
		if((index1 = line.find(FieldSeparator)) != string::npos)
		{
			ActualField = line.substr(0, index1);
			line = line.substr(index1+FieldSeparatorLength);
		}
		else
			line = "";
//		CleanString(ActualField);
		FieldsInLine.push_back(ActualField);
	}

	// Now all the field contents are in FieldsInLine
	for(i=0; i<(int)FieldDescriptions.size(); i++)
	{
		SecuField = NVNEW Intdbfield;
		if(SecuField == NULL)
		{
			delete SecuRecord;
			return false; 
		}
		SecuRecord->Fields.push_back((void *)SecuField);

		// we must now write the field entries
		ActualField = FieldsInLine[i];
		string ActualFieldEntry;
		bool bPrimaryColumn = false;
		if(IsPrimaryColumn(i))
			bPrimaryColumn = true;
		while(1)
		{
			if((index1 = ActualField.find(EntrySeparator)) == string::npos)
			{
//				CleanString(ActualField);
				if(ActualField.length())
				{
					if(ActualField == "NV_EMPTY")
						ActualField = "";
					SecuField->Entries.push_back(ActualField);
				}
				if(bPrimaryColumn)
					PrimaryValue = ActualField;
				break;
			}

			ActualFieldEntry = ActualField.substr(0, index1);
//			CleanString(ActualFieldEntry);
			ActualField = ActualField.substr(index1+EntrySeparatorLength);
			if(bPrimaryColumn)
				continue;
			if(ActualFieldEntry.length())
			{
				if(ActualFieldEntry == "NV_EMPTY")
					ActualFieldEntry = "";
				SecuField->Entries.push_back(ActualFieldEntry);
			}
		}
	}
	if(GetPrimaryField().length())
	{
		int indexprimary = GetPrimary(PrimaryValue);
		if((indexprimary != -1) && (indexprimary != index))
		{
			// The primary key ever exist in another record
			delete SecuRecord;
			return false; 
		}
		// We must eventually remove the old primary index
		int FieldIndex = GetFieldIndex(GetPrimaryField().c_str());
		if(FieldIndex != -1)
		{
			string OldPrimaryValue = GetCellLine(index, FieldIndex, 0, "NV_INTERNAL_NO_VALUE");
			if(OldPrimaryValue != "NV_INTERNAL_NO_VALUE")
				RemovePrimary(OldPrimaryValue);
		}
	}
	ClearRecord(index);
	Record->CopyFrom(SecuRecord);
	SetPrimary(PrimaryValue, index);

	delete SecuRecord;

	return true;
}




bool Intdbtable::AddRecords(const char* RecordData, const char* RecordSeparator, const char* FieldSeparator, const char* EntrySeparator)
{

	Intdbrecord *Record = NULL;

	if(RecordSeparator == NULL)
		return false;
	if(strlen(RecordSeparator) == 0)
		return false;

	if(FieldSeparator == NULL)
		return false;
	if(strlen(FieldSeparator) == 0)
		return false;

	if(EntrySeparator == NULL)
		return false;
	if(strlen(EntrySeparator) == 0)
		return false;

	string data = "";
	if(RecordData != NULL)
		data = RecordData;

	int i;
	Intdbfield *SecuField = NULL;
	int RecordSeparatorLength = (int)strlen(RecordSeparator);
	int FieldSeparatorLength = (int)strlen(FieldSeparator);
	int EntrySeparatorLength = (int)strlen(EntrySeparator);
	string PrimaryValue = "";

	string line = "";
	string::size_type index;
	int indexdb = GetSize();
	while(data.length())
	{
		index = data.find(RecordSeparator);
		if(index == string::npos)
		{
			line = data;
			data = "";
		}
		else
		{
			line = data.substr(0, index);
			data = data.substr(index+RecordSeparatorLength);
		}

		Record = NVNEW Intdbrecord;
		if(Record == NULL)
			return false;
		vector<string> FieldsInLine;
		string ActualField;
		string::size_type index1;
		while(FieldsInLine.size() < FieldDescriptions.size())
		{
			ActualField = line;
			if((index1 = line.find(FieldSeparator)) != string::npos)
			{
				ActualField = line.substr(0, index1);
				line = line.substr(index1+FieldSeparatorLength);
			}
			else
				line = "";
			FieldsInLine.push_back(ActualField);
		}

		// Now all the field contents are in FieldsInLine
		for(i=0; i<(int)FieldDescriptions.size(); i++)
		{
			SecuField = NVNEW Intdbfield;
			if(SecuField == NULL)
			{
				delete Record;
				return false;
			}
			Record->Fields.push_back((void *)SecuField);

			// we must now write the field entries
			ActualField = FieldsInLine[i];
			string ActualFieldEntry;
			bool bPrimaryColumn = false;
			if(IsPrimaryColumn(i))
				bPrimaryColumn = true;
			while(1)
			{
				if((index1 = ActualField.find(EntrySeparator)) == string::npos)
				{
					if(ActualField.length())
					{
						if(ActualField == "NV_EMPTY")
							ActualField = "";
						SecuField->Entries.push_back(ActualField);
					}
					if(bPrimaryColumn)
						PrimaryValue = ActualField;
					break;
				}

				ActualFieldEntry = ActualField.substr(0, index1);
				ActualField = ActualField.substr(index1+EntrySeparatorLength);
				if(bPrimaryColumn)
					continue;
				if(ActualFieldEntry.length())
				{
					if(ActualFieldEntry == "NV_EMPTY")
						ActualFieldEntry = "";
					SecuField->Entries.push_back(ActualFieldEntry);
				}
			}
		}
		if(PrimaryValue.length())
		{
			int indexprimary = GetPrimary(PrimaryValue);
			if(indexprimary != -1)
			{
				// The primary key ever exist in another record
				delete Record;
				return false; 
			}
		}
		Records.push_back(Record);
		SetPrimary(PrimaryValue, indexdb);
		indexdb++;
	}

	return true;
}



void Intdbtable::SwapRecords(int index1, int index2)
{
	if(index1 < 0)
		return;
	if(index2 < 0)
		return;

	if(index1 >= (int)Records.size())
		return;
	if(index2 >= (int)Records.size())
		return;

	void *Record = Records[index1];
	Records[index1] = Records[index2];
	Records[index2] = Record;

	if(PrimaryName.length())
	{
		// We must also swap the records in the hash table
		string Value1 = GetRecordPrimaryValue(index2); //value1 is now on index 2 because we ever swap
		string Value2 = GetRecordPrimaryValue(index1);
		SetPrimary(Value1, index2);
		SetPrimary(Value2, index1);
	}
}


int Intdbtable::GetRecords(int first, int last, Intdbtable **Result, const char* Columns)
{
	if(Result != NULL)
		*Result = NULL;
	if(Result != NULL)
	{
		*Result = NVNEW Intdbtable;
		if(*Result == NULL)
			return INTDB_MEMORYERROR;
		(*Result)->PrimaryName = PrimaryName;
		(*Result)->PrimaryCaseSensitive = PrimaryCaseSensitive;
	}
	// Prepare the structure of the new table
	if(Result != NULL)
		(*Result)->CopyFrom(this, false);	// We don't copy table data
	if(!Records.size())
		return 0;

	if(first == -1)
		return 0;
	if(first < 0)
		first = 0;
	if(last < 0)
		last = (int)Records.size()-1;
	if(last >= (int)Records.size())
		last = (int)Records.size()-1;
	if(first > last)
		return 0;

	Intdbrecord *SecuRecord;
	Intdbrecord *NewSecuRecord;
	int i;
	for(i=first; i<=last; i++)
	{
		SecuRecord = (Intdbrecord *)(Records[i]);
		NewSecuRecord = NVNEW Intdbrecord;
		if(NewSecuRecord == NULL)
			return INTDB_MEMORYERROR;
		NewSecuRecord->CopyFrom(SecuRecord);
		(*Result)->Records.push_back((void *)NewSecuRecord);
	}
	
	// Now filter the columns
	(*Result)->FilterFields(Columns);


	return 0;
}


void Intdbtable::SwapFields(int Fieldindex1, int Fieldindex2)
{
	if(Fieldindex1 < 0)
		return;
	if(Fieldindex2 < 0)
		return;

	if(Fieldindex1 >= (int)FieldDescriptions.size())
		return;
	if(Fieldindex2 >= (int)FieldDescriptions.size())
		return;

	void *FieldDesc = FieldDescriptions[Fieldindex1];
	FieldDescriptions[Fieldindex1] = FieldDescriptions[Fieldindex2];
	FieldDescriptions[Fieldindex2] = FieldDesc;

	int i;
	int numRecords = GetSize();
	for (i=0; i<numRecords; i++)
	{
		Intdbrecord *Record = GetRecord(i);
		if(Record == NULL)
			continue;
		Record->SwapFields(Fieldindex1, Fieldindex2);
	}
}




void Intdbtable::InsertField(const char* FieldName, const char* FieldDesc, bool numeric, int FieldIndex)
{
	if(FieldName == NULL)
		return;

	string FieldNameHere = FieldName;
	CleanString(FieldNameHere);
	if(!FieldNameHere.length())
		return;

	if(GetFieldIndex(FieldName) != -1)
		return;	// The field ever exists

	if(FieldIndex >= (int)FieldDescriptions.size())
		FieldIndex = -1;

	if((FieldIndex != -1) && (FieldIndex < 0))
		FieldIndex = 0;

	Intdbfielddesc *SecuFieldDesc = NVNEW Intdbfielddesc;
	if(SecuFieldDesc == NULL)
		return;

	SecuFieldDesc->Name = FieldNameHere;
	if(FieldDesc != NULL)
		SecuFieldDesc->Description = FieldDesc;

	if(numeric)
		SecuFieldDesc->Type = FIELD_TYPE_NUMERIC;
	else
		SecuFieldDesc->Type = FIELD_TYPE_ALPHA;

	if(FieldIndex == -1)
		FieldDescriptions.push_back((void *)SecuFieldDesc);
	else
		FieldDescriptions.insert(FieldDescriptions.begin()+FieldIndex, (void *)SecuFieldDesc);

	int i;
	int numRecords = GetSize();
	for (i=0; i<numRecords; i++)
	{
		Intdbrecord *Record = GetRecord(i);
		if(Record == NULL)
			continue;
		Record->InsertField(FieldIndex);
	}
}




bool Intdbtable::AddToField(int index, const char* FieldName, const char* Value, bool AppendMode)
{
	Intdbrecord *Record = GetRecord(index);
	if(Record == NULL)
		return false;

	int FieldIndex = GetFieldIndex(FieldName);
	if(FieldIndex == -1)
		return false;

	Intdbfield *Field = Record->GetField(FieldIndex);
	if(Field == NULL)
		return false;

	if(AppendMode)
	{
		if(IsPrimaryColumn(FieldIndex))
			if(Field->Entries.size())	// Cannot add something to a primary key not empty
				return false;
	}
	else
	{
		if(IsPrimaryColumn(FieldIndex))	// Eventually remove the hash table
			RemovePrimary(GetRecordPrimaryValue(index));
		Field->Clear();
	}

	if(strlen(Value))
	{
		if(IsPrimaryColumn(FieldIndex))
		{
			int PrimaryIndex = GetPrimary(Value);
				if((PrimaryIndex != -1) && (PrimaryIndex != index))
					return false;	// cannot set the primary key because it ever exists
		}

		Field->Entries.push_back(Value);
		if(IsPrimaryColumn(FieldIndex))
		{
			// We must set the primary value if it's the first entry
			if(Field->Entries.size() == 1)
				SetPrimary(Value, index);
		}
	}

	return true;
}



string Intdbtable::GetFirstLineField(int index, const char* FieldName, const char *Default)
{

	Intdbrecord *Record = GetRecord(index);
	if(Record == NULL)
		return "";

	string RetVal = "";

	int FieldIndex = GetFieldIndex(FieldName);
	if(FieldIndex == -1)
		return RetVal;

	Intdbfield *Field = Record->GetField(FieldIndex);
	if(Field == NULL)
		return RetVal;
	
	RetVal = Field->GetEntry(0, Default);

	return RetVal;
}

	



int Intdbtable::Search(const char* Query, Intdbtable **Result, const char* SortField, int Mode, int* FirstIndex, int MaxDocs, const char* Columns, bool RowIndexCol)
{
	if(Result != NULL)
		*Result = NULL;

	if(FirstIndex != NULL)
		*FirstIndex = -1;
	
	if(Query == NULL)
		return INTDB_INVALIDQUERY;
	
	if(!strlen(Query))
		return INTDB_INVALIDQUERY;
	
	char *QueryHere = (char *)malloc(strlen(Query) + 1);
	strcpy(QueryHere, Query);
	strupr(QueryHere);
	CleanString(QueryHere);

	string sQueryHere = QueryHere;

	FreeData(QueryHere);

	string::size_type index;

	// Replace "= " with "=" 
	while((index = sQueryHere.find("= ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace "> " with ">" 
	while((index = sQueryHere.find("> ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace "< " with "<" 
	while((index = sQueryHere.find("< ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace " =" with "=" 
	while((index = sQueryHere.find(" =")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " >" with ">" 
	while((index = sQueryHere.find(" >")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " <" with "<" 
	while((index = sQueryHere.find(" <")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " !" with "!" 
	while((index = sQueryHere.find(" !")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " AND " with "@@AN" 
	while((index = sQueryHere.find(" AND ")) !=  string::npos)
	{
		string provistr = sQueryHere.substr(0, index);
		provistr += "@@AN";
		provistr += sQueryHere.substr(index+5);
		sQueryHere = provistr;
	}

	// Replace " OR " with "@@OR" 
	while((index = sQueryHere.find(" OR ")) !=  string::npos)
	{
		string provistr = sQueryHere.substr(0, index);
		provistr += "@@OR";
		provistr += sQueryHere.substr(index+4);
		sQueryHere = provistr;
	}

	vector<string> Operand1s;
	vector<string> Operand2s;
	vector<int> Operators;
	vector<int> Combinators;

	if(Result != NULL)
	{
		*Result = NVNEW Intdbtable;
		if(*Result == NULL)
			return INTDB_MEMORYERROR;
		(*Result)->PrimaryName = PrimaryName;
		(*Result)->PrimaryCaseSensitive = PrimaryCaseSensitive;
	}

	// Check for search all string
	if(sQueryHere == "*")
	{
		if(Result != NULL)
		{
			(*Result)->CopyFrom(this, true, MaxDocs);	// In case of search all,  we return all the table (limited to MaxDocs)
			if(RowIndexCol)
			{
				(*Result)->InsertField("NV_SRC_ROW_INDEX", "Row index", true, 0);
				int NumRows = (*Result)->GetSize();
				int i;
				char Buf[20];
				for(i=0; i<NumRows; i++)
				{
					Intdbrecord *SecuRecord = (*Result)->GetRecord(i);
					Intdbfield *SecuField = SecuRecord->GetField(0);
					sprintf(Buf, "%d", i+1);
					SecuField->Clear();
					SecuField->InsertEntry(Buf);
				}
			}
			if((SortField != NULL) && strlen(SortField))
			{
//				int FieldIndex = GetFieldIndex(SortField);

//				if(FieldIndex == -1)
//				{
//					delete (*Result);
//					return INTDB_INVALIDQUERY;
//				}
				(*Result)->Sort(SortField);
			}
			else
				(*Result)->SetPrimaryHashTable();	// Ever made if there is a sort requested
		}

		if(Mode == SEARCH_MODE_REMOVE)
			ClearData();
		if(Mode == SEARCH_MODE_CLEAR)
			ClearData(true);

		// Now filter the columns
		if((Mode == SEARCH_MODE_SEARCH) && (Result != NULL))
			(*Result)->FilterFields(Columns);
		return 0;
	}

	while(1)
	{
		int Combinator = SEARCH_COMBINATOR_NONE;
		string Expression = "";
		if(((index = sQueryHere.find("@@")) != string::npos) && (sQueryHere.length() > index+4))
		{
			Expression = sQueryHere.substr(0, index);
			if((sQueryHere[index+2] == 'A') && (sQueryHere[index+3] == 'N'))
				Combinator = SEARCH_COMBINATOR_AND;
			if((sQueryHere[index+2] == 'O') && (sQueryHere[index+3] == 'R'))
				Combinator = SEARCH_COMBINATOR_OR;
			sQueryHere  = sQueryHere.substr(index+4);
		}
		else
			Expression = sQueryHere;

		int Operator = SEARCH_OPERATOR_NONE;
		int OperatorLength = 1;
		if((index = Expression.find("!=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_NE;
			OperatorLength = 2;
		}
		else if((index = Expression.find("<=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_STE;
			OperatorLength = 2;
		}
		else if((index = Expression.find(">=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_GTE;
			OperatorLength = 2;
		}
		else if((index = Expression.find("<")) != string::npos)
			Operator = SEARCH_OPERATOR_ST;
		else if((index = Expression.find(">")) != string::npos)
			Operator = SEARCH_OPERATOR_GT;
		else if((index = Expression.find("=")) != string::npos)
			Operator = SEARCH_OPERATOR_E;
		
		if(Operator == SEARCH_OPERATOR_NONE)
		{
			if(Result != NULL)
				delete (*Result);
			return INTDB_INVALIDQUERY;
		}

		string Operand1 = Expression.substr(0, index);
		string Operand2 = Expression.substr(index+OperatorLength);

		CleanString(Operand1);
		CleanString(Operand2);

		if((!Operand1.length()) || (!Operand1.length()))
		{
			if(Result != NULL)
				delete (*Result);
			return INTDB_INVALIDQUERY;
		}

		Operand1s.push_back(Operand1);
		Operand2s.push_back(Operand2);
		Operators.push_back(Operator);
		Combinators.push_back(Combinator);

		if(Combinator == SEARCH_COMBINATOR_NONE)
			break;
	}

	// Prepare the structure of the new table
	if(Result != NULL)
		(*Result)->CopyFrom(this, false);	// We don't copy table data

	// Now, we make the search
	int i, j;
	Intdbrecord *SecuRecord;
	Intdbrecord *NewSecuRecord;
	Intdbfield *SecuField;
	Intdbfielddesc *SecuFieldDesc;

	vector<int> MarkedRecords;

	int NumDocsFound = 0;
	for(i=0; i<(int)Records.size(); i++)
	{
		SecuRecord = (Intdbrecord *)(Records[i]);
		if((MaxDocs != -1) && (NumDocsFound >= MaxDocs))
			break;
		for(j=0; j<(int)Operators.size(); j++)
		{
			int Operator = Operators[j];
			int Combinator = Combinators[j];
			string Operand1 = Operand1s[j];
			string Operand2 = Operand2s[j];
			int FieldIndex = GetFieldIndex(Operand1.c_str());
			if(FieldIndex == -1)	// The field must exist
			{
				if(Result != NULL)
					delete (*Result);
				return INTDB_INVALIDQUERY;
			}
			SecuField = SecuRecord->GetField(FieldIndex);
			SecuFieldDesc = (Intdbfielddesc *)(FieldDescriptions[FieldIndex]);
			bool Numeric = false;
			if(SecuFieldDesc->Type == FIELD_TYPE_NUMERIC)
				Numeric = true;

			bool Verify = SecuField->Verify(Operand2.c_str(), Operator, Numeric);
			if(!Verify)	// If the field doesn't verify the request, we go to the next expression
			{
				if(Combinators[j] == SEARCH_COMBINATOR_OR)
					continue;
				else
					break;
			}
//			if((j>0) && (Combinators[j-1] == SEARCH_COMBINATOR_OR))
			if(Combinators[j] == SEARCH_COMBINATOR_OR)
			{
				// If we have a OR operator, we add the record and go to the next record
				// This avoid to verify other requests
				if(Result != NULL)
				{
					NewSecuRecord = NVNEW Intdbrecord;
					if(NewSecuRecord == NULL)
						return INTDB_MEMORYERROR;
					NewSecuRecord->CopyFrom(SecuRecord);
					(*Result)->Records.push_back((void *)NewSecuRecord);
				}
				NumDocsFound++;
				if((FirstIndex != NULL) && (*FirstIndex == -1))
					*FirstIndex = i;
//				if(Mode != SEARCH_MODE_SEARCH)
					MarkedRecords.push_back(i);
				break;
			}
			
			if(j == Operators.size() - 1)
			{
				// If we are here, the record satisfies the request
				if(Result != NULL)
				{
					NewSecuRecord = NVNEW Intdbrecord;
					if(NewSecuRecord == NULL)
						return INTDB_MEMORYERROR;
					NewSecuRecord->CopyFrom(SecuRecord);
					(*Result)->Records.push_back((void *)NewSecuRecord);
				}
				NumDocsFound++;
				if((FirstIndex != NULL) && (*FirstIndex == -1))
					*FirstIndex = i;
//				if(Mode != SEARCH_MODE_SEARCH)
					MarkedRecords.push_back(i);
				break;
			}

		}
	}

	if(Result != NULL)
	{
		if(RowIndexCol)
		{
			(*Result)->InsertField("NV_SRC_ROW_INDEX", "Row index", true, 0);
			int NumRows = (int)MarkedRecords.size();
			int i;
			char Buf[20];
			for(i=0; i<NumRows; i++)
			{
				Intdbrecord *SecuRecord = (*Result)->GetRecord(i);
				Intdbfield *SecuField = SecuRecord->GetField(0);
				sprintf(Buf, "%d", MarkedRecords[i]+1);
				SecuField->Clear();
				SecuField->InsertEntry(Buf);
			}
		}
		if((SortField != NULL) && strlen(SortField))
		{
//			int FieldIndex = GetFieldIndex(SortField);
//
//			if(FieldIndex == -1)
//			{
//				delete (*Result);
//				return INTDB_INVALIDQUERY;
//			}

			(*Result)->Sort(SortField);
		}
		else
			(*Result)->SetPrimaryHashTable();	// Ever made if there is a sort requested

		// Now filter the columns
		(*Result)->FilterFields(Columns);
	}

	if(MarkedRecords.size())
	{
		for(i=0; i<(int)MarkedRecords.size(); i++)
		{
			int IndexMarkedRecord = MarkedRecords[i];
			SecuRecord = (Intdbrecord *)(Records[IndexMarkedRecord]);
			if(Mode == SEARCH_MODE_REMOVE)
			{
				delete SecuRecord;
				Records[IndexMarkedRecord] = NULL;
			}
			if(Mode == SEARCH_MODE_CLEAR)
				SecuRecord->ClearFieldData();
		}
		if(Mode == SEARCH_MODE_REMOVE)
		{
			vector<void *> CopyRecords;
			for(i=0; i<(int)Records.size(); i++)
				if(Records[i] != NULL)
					CopyRecords.push_back(Records[i]);

			Records.clear();
			for(i=0; i<(int)CopyRecords.size(); i++)
				Records.push_back(CopyRecords[i]);
		}
	}
	
	return 0;
}



bool Intdbtable::FilterFields(const char* Columns)
{
	string ColumnsHere = "";
	if(Columns != NULL)
		ColumnsHere = Columns;
	CleanString(ColumnsHere, true);
	if(!ColumnsHere.length())
		return true;	// nothing to do

	// Check if we have a column named NV_SRC_ROW_INDEX
	bool hasNV_SRC_ROW_INDEX = false;
	if(GetFieldIndex("NV_SRC_ROW_INDEX") != -1)
		hasNV_SRC_ROW_INDEX = true;

	// We must get the column name to keep (so the ones to remove)
	map<string,int> ColumnsToKeep;
	map<string,string> ColumnsToKeepNewNames;
	string::size_type index;
	string Provistr;

	int i,j;
	int colindex;
	int NumColumnDst = GetNumFields();
	i=1;
	while(ColumnsHere.length())
	{
		string Column = ColumnsHere;
		index = Column.find(';');
		if(index != string::npos)
		{
			Column = ColumnsHere.substr(0, index);
			ColumnsHere = ColumnsHere.substr(index+1);
		}
		else
			ColumnsHere = "";
		CleanString(Column);
		if(!Column.length())
			continue;
		string ColomnOldName = Column;
		index = ColomnOldName.find(':');
		if(index != string::npos)
		{
			Column = ColomnOldName.substr(0, index);
			ColomnOldName = ColomnOldName.substr(index+1);
		}
		CleanString(Column);
		CleanString(ColomnOldName);
		if(!ColomnOldName.length())
			continue;
		if(!Column.length())
			Column = ColomnOldName;
		colindex = GetFieldIndex(ColomnOldName.c_str());
		if(colindex == -1)
			continue;	// the source column doesn't exist
		Provistr = ColomnOldName;
		MakeUpper(Provistr);
		ColumnsToKeep[Provistr] = i;
		ColumnsToKeepNewNames[Provistr] = Column;
		i++;
	}

	// if the table has a NV_SRC_ROW_INDEX column and it's not in the required columpns, we add it as first column
	if(hasNV_SRC_ROW_INDEX)
	{
		if(ColumnsToKeep.find("NV_SRC_ROW_INDEX") == ColumnsToKeep.end())
		{
			ColumnsToKeep[Provistr] = 0;
			ColumnsToKeepNewNames[Provistr] = "NV_SRC_ROW_INDEX";
		}
	}

	vector<string> columns_to_remove;
	for(i=0; i<NumColumnDst; i++)
	{
		string ColName = GetFieldName(i);
		Provistr = ColName;
		MakeUpper(Provistr);
		if(Provistr == "NV_SRC_ROW_INDEX")
			continue;	// We can't remove this column when requested
		if(ColumnsToKeep.find(Provistr) == ColumnsToKeep.end())
			columns_to_remove.push_back(ColName);
	}
	for(i=0; i<(int)columns_to_remove.size(); i++)
	{
		colindex = GetFieldIndex(columns_to_remove[i].c_str());
		if(colindex != -1)
			RemoveField(colindex);
	}

	// Now we must sort the required columns

	// make the bubble sort

	vector<int> neworder;
	NumColumnDst = GetNumFields();
	for(i=0; i<NumColumnDst; i++)
	{
		string ColName = GetFieldName(i);
		Provistr = ColName;
		MakeUpper(Provistr);
		int Order = ColumnsToKeep[Provistr];
		neworder.push_back(Order);
	}
	int num = (int)neworder.size();
	bool mutate = false;
	
	for(i=1;i<num;++i)
	{
		for(j=num-1;j>=i;--j)
		{
			mutate = false;
			int Value1 = neworder[j-1];
			int Value2 = neworder[j];
			if(Value1 > Value2)
				mutate = true;
			if(mutate)
			{
				neworder[j-1] = Value2;
				neworder[j] = Value1;
				SwapFields(j-1,j);
			}
		}
	}    

	// finally change the column names
	for(i=0; i<NumColumnDst; i++)
	{
		string ColName = GetFieldName(i);
		Provistr = ColName;
		MakeUpper(Provistr);
		if(ColumnsToKeepNewNames[Provistr] != ColName)
			SetFieldName(i, ColumnsToKeepNewNames[Provistr]);
	}

	return true;
}


void Intdbtable::Sort(const char* SortField)
{
	int num;
	
	if(SortField == NULL)
		return;
	
	if(!strlen(SortField))
		return;

	bool Ascending = true;
	const char* RealSortField = SortField;

	if(SortField[0] == '-')
	{
		Ascending = false;
		RealSortField++;
	}

	if(!strlen(RealSortField))
		return;


	int FieldIndex = GetFieldIndex(RealSortField);
	if(FieldIndex == -1)
		return;


	Intdbfielddesc *FieldDesc = (Intdbfielddesc *)(FieldDescriptions[FieldIndex]);
	bool TypeNumeric = false;
	if(FieldDesc->Type == FIELD_TYPE_NUMERIC)
		TypeNumeric = true;


	num = (int)Records.size();
	if(num <=1)
		return;

	// Do a Quick sort
	unsigned int stack_pointer = 0;
	int first_stack[32];
	int last_stack[32];
	int Insertion_Sort_Bound = 30;
	int last = num-1;
	int first = 0;

	for (;;)
	{
		if (last - first <= Insertion_Sort_Bound)
		{
			/* for small sort, use insertion sort */
			unsigned int indx;
			void* prev_val = Records[first];
			void* cur_val;

			for (indx = (unsigned int)first + 1; indx <= (unsigned int)last; ++indx)
			{
				cur_val = Records[indx];
				if (SortFunc(prev_val, cur_val, FieldIndex, TypeNumeric, Ascending) > 0)
				{
					/* out of order: array[indx-1] > array[indx] */
					unsigned int indx2;
					Records[indx] = prev_val; /* move up the larger item first */

					/* find the insertion point for the smaller item */
					for (indx2 = indx - 1; indx2 > (unsigned int)first; )
					{
						void* temp_val = Records[indx2 - 1];
						if (SortFunc(temp_val, cur_val, FieldIndex, TypeNumeric, Ascending) > 0)
						{
							Records[indx2--] = temp_val;
							/* still out of order, move up 1 slot to make room */
						}
						else
							break;
					}
					Records[indx2] = cur_val; /* insert the smaller item right here */
				}
				else
				{
					/* in order, advance to next element */
					prev_val = cur_val;
				}
			}
		}
		else
		{
			void* pivot;
		
			/* try quick sort */
			{
				void* temp;
				unsigned int med = (first + last) >> 1;
				/* Choose pivot from first, last, and median position. */
				/* Sort the three elements. */
				temp = Records[first];
				if (SortFunc(temp, Records[last], FieldIndex, TypeNumeric, Ascending) > 0)
				{
					Records[first] = Records[last]; Records[last] = temp;
				}
				temp = Records[med];
				if (SortFunc(Records[first], temp, FieldIndex, TypeNumeric, Ascending) > 0)
				{
					Records[med] = Records[first]; Records[first] = temp;
				}
				temp = Records[last];
				if (SortFunc(Records[med], temp, FieldIndex, TypeNumeric, Ascending) > 0)
				{
					Records[last] = Records[med]; Records[med] = temp;
				}
				pivot = Records[med];
			}
			{
				unsigned int up;
				{
					unsigned int down;
					/* First and last element will be loop stopper. */
					/* Split array into two partitions. */
					down = first;
					up = last;
					for (;;)
					{
						do
						{
							++down;
						} while (SortFunc(pivot, Records[down], FieldIndex, TypeNumeric, Ascending) > 0);
					
						do
						{
							--up;
						} while (SortFunc(Records[up], pivot, FieldIndex, TypeNumeric, Ascending) > 0);
					
						if (up > down)
						{
							void* temp;
							/* interchange L[down] and L[up] */
							temp = Records[down]; Records[down]= Records[up]; Records[up] = temp;
						}
						else
							break;
					}
				}
				{
					unsigned int len1; /* length of first segment */
					unsigned int len2; /* length of second segment */
					len1 = up - first + 1;
					len2 = last - up;
					/* stack the partition that is larger */
					if (len1 >= len2)
					{
						first_stack[stack_pointer] = first;
						last_stack[stack_pointer++] = up;
					
						first = up + 1;
						/*  tail recursion elimination of
						*  Qsort(This,fun_ptr,up + 1,last)
						*/
					}
					else
					{
						first_stack[stack_pointer] = up + 1;
						last_stack[stack_pointer++] = last;

						last = up;
						/* tail recursion elimination of
						* Qsort(This,fun_ptr,first,up)
						*/
					}
				}
				continue;
			}
		/* end of quick sort */
		}
		if (stack_pointer > 0)
		{
			/* Sort segment from stack. */
			first = first_stack[--stack_pointer];
			last = last_stack[stack_pointer];
		}
		else
			break;
	} /* end for */

	// Rebuild the hash table
	SetPrimaryHashTable();
}

int Intdbtable::SortFunc(void* a, void* b, int FieldIndex, bool Numeric, bool Ascending)
{
	Intdbrecord *Record1 = (Intdbrecord *)(a);
	Intdbrecord *Record2 = (Intdbrecord *)(b);
	Intdbfield *Field1 = (Intdbfield *)(Record1->GetField(FieldIndex));;
	Intdbfield *Field2 = (Intdbfield *)(Record2->GetField(FieldIndex));
	string FieldEntry1 = Field1->GetEntry(0);
	string FieldEntry2 = Field2->GetEntry(0);

	return Field1->SortFunc(FieldEntry1, FieldEntry2, Numeric, Ascending);
}



int Intdbtable::Load(const char *FileName, int iEncoding, bool bUnicode)
{
	if(FileName == NULL)
		return INTDB_INVALIDFILE;
	if(!strlen(FileName))
		return INTDB_INVALIDFILE;


	FILE *stream = Nvfopen(FileName, "rb");
	if(stream == NULL)
		return INTDB_CANNOTOPENFILE;

	Clear();

	// we get the file size
	fseek(stream, 0, SEEK_END);
	unsigned long FileSize = ftell(stream);
	fseek(stream, 0, SEEK_SET);

	// The file must contain at least the identifier
	if(FileSize <= DBFILE_SIZE_IDENTIFIER)
	{
		fclose(stream);
		return INTDB_INVALIDFILE;
	}

	// We check that the identifier is OK
	char Identifier[DBFILE_SIZE_IDENTIFIER];
	memset(Identifier, 0, sizeof(Identifier));
	fread(Identifier, 1, DBFILE_SIZE_IDENTIFIER, stream);
	if(strcmp(Identifier, DBFILE_IDENTIFIER) != 0)
	{
		fclose(stream);	       
		return INTDB_INVALIDFILE;
	}

	// and we allocate memory for the file buffer
	char *FileBuffer = (char *)malloc(FileSize - DBFILE_SIZE_IDENTIFIER);
	if(FileBuffer == NULL)
	{
		fclose(stream);	       
		return INTDB_MEMORYERROR;
	}

	// Read file
	unsigned long nbBytes = (unsigned long)fread(FileBuffer, 1, FileSize - DBFILE_SIZE_IDENTIFIER, stream);
	if(ferror(stream) || (nbBytes != (FileSize - DBFILE_SIZE_IDENTIFIER)))
	{
		fclose(stream);	       
		FreeData(FileBuffer);
		return INTDB_CANNOTREADFILE;
	}

	// We have finished with the file
	fclose(stream);	       

	char *FileStream = FileBuffer;
	unsigned char TagType;
	unsigned long TagLength;
	char *TagData;

	Intdbrecord *SecuRecord = NULL;
	Intdbfielddesc *SecuFieldDesc = NULL;
	Intdbfield *SecuField = NULL;
	
	string FileVersion = DBFILE_VERSION;
	int TotalBytes = FileSize - DBFILE_SIZE_IDENTIFIER;
	int BytesRead = 0;

	int BytesReadInTag;

#ifdef NV_SERVER
	if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
		bUnicode = bOldObjectsUnicode; // default object encoding if not not found in file
#endif

	while(1)
	{
		BytesReadInTag = ::ReadNextTag(FileStream, &TagType, &TagLength, &TagData);
		if(BytesReadInTag == -1)
		{
			Clear();
			FreeData(FileBuffer);
			return INTDB_MEMORYERROR;
		}
		BytesRead += BytesReadInTag;
		FileStream += BytesReadInTag;
		if(BytesRead > TotalBytes)
		{
			Clear();
			FreeData(FileBuffer);
			return INTDB_INVALIDFILE;
		}
		if(TagType == FILE_DB_END)
			break;

		switch(TagType)
		{
			case FILE_VERSION:
				if(TagData != NULL)
				{
					FileVersion = TagData;
					CleanString(FileVersion);
					// PM Unicode saving
					// The unicode flag is placed here for now for compatibility reasons
					// otherwise the saved filedb (ex registry) written with new Nirva version
					// cannot be read by older nirva versions
					if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
					{
						string::size_type index;
						if((index = FileVersion.rfind(';')) != string::npos)
						{
							string sUnicode = FileVersion.substr(index+1);
							CleanString(sUnicode);
							if(sUnicode == "1")
								bUnicode = true;
							else
								bUnicode = false;
						}
					}
				}
				break;
			case FILE_DB:
				Clear();
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_DB_NAME:
				break;
			case FILE_DB_DESC:
				break;
			case FILE_DB_VERSION:
				break;
			case FILE_TABLE:
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_TABLE_NAME:
				if(TagData != NULL)
				{
					Name = TagData;
					CleanString(Name);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(Name, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(Name, bUnicode, false);
				}
				break;
			case FILE_TABLE_DESC:
				if(TagData != NULL)
				{
					Description = TagData;
					CleanString(Description);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(Description, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(Description, bUnicode, false);
				}
				break;
			case FILE_TABLE_PRIMARY:
				if(TagData != NULL)
				{
					PrimaryName = TagData;
					CleanString(PrimaryName);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(PrimaryName, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(PrimaryName, bUnicode, false);
				}
				break;
			case FILE_TABLE_PRIMARY_CASE:
				if(TagData != NULL)
				{
					string sPrimaryCaseSensitive = TagData;
					CleanString(sPrimaryCaseSensitive);
					if(sPrimaryCaseSensitive == "1")
						PrimaryCaseSensitive = true;
					else
						PrimaryCaseSensitive = false;
				}
				break;
			case FILE_TABLE_UNICODE:
				if(TagData != NULL)
				{
					if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
					{
						string sUnicode = TagData;
						CleanString(sUnicode);
						if(sUnicode == "1")
							bUnicode = true;
						else
							bUnicode = false;
					}
				}
				break;
			case FILE_DFIELD:
				SecuFieldDesc = NULL;
				SecuFieldDesc = NVNEW Intdbfielddesc;
				if(SecuFieldDesc == NULL)
				{
					Clear();
					FreeData(TagData);
					FreeData(FileBuffer);
					return INTDB_MEMORYERROR;
				}
				FieldDescriptions.push_back((void *)SecuFieldDesc);
				break;
			case FILE_DFIELD_NAME:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Name = TagData;
						CleanString(SecuFieldDesc->Name);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(SecuFieldDesc->Name, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(SecuFieldDesc->Name, bUnicode, false);
					}
				}
				break;
			case FILE_DFIELD_DESC:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Description = TagData;
						CleanString(SecuFieldDesc->Description);
						if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
							DecodeUrlString(SecuFieldDesc->Description, bUnicode, false);
						if(iEncoding == ENCODING_MODE_ENCODE)
							EncodeUrlString(SecuFieldDesc->Description, bUnicode, false);
					}
				}
				break;
			case FILE_DFIELD_TYPE:
				if(SecuFieldDesc != NULL)
					if(TagData != NULL)
						SecuFieldDesc->Type = *TagData; // just one byte here
				break;
			case FILE_RECORD:
				SecuRecord = NULL;
				SecuRecord = NVNEW Intdbrecord;
				if(SecuRecord == NULL)
				{
					Clear();
					FreeData(TagData);
					FreeData(FileBuffer);
					return INTDB_MEMORYERROR;
				}
				Records.push_back((void *)SecuRecord);
				SecuField = NULL;
				break;
			case FILE_FIELD:
				SecuField = NULL;
				if(SecuRecord != NULL)
				{
					SecuField = NVNEW Intdbfield;
					if(SecuField == NULL)
					{
						Clear();
						FreeData(TagData);
						FreeData(FileBuffer);
						return INTDB_MEMORYERROR;
					}
					SecuRecord->Fields.push_back((void *)SecuField);
				}
				break;
			case FILE_ENTRY:
				if(SecuField != NULL)
				{
					if(TagData != NULL)
					{
						string Provistr = TagData;
//						CleanString(Provistr);
						if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
							DecodeUrlString(Provistr, bUnicode, false);
						if(iEncoding == ENCODING_MODE_ENCODE)
							EncodeUrlString(Provistr, bUnicode, false);
						if(Provistr.size())
							SecuField->Entries.push_back(Provistr);
					}
				}
				break;
			default:
				// There is no more default otherwise we cannot add new tags type without creating a compatibility problem
				// if a user installs a new version and comes back to an older version
				//Clear();
				//FreeData(FileBuffer);
				//FreeData(TagData);
				//return INTDB_INVALIDFILE;
				break;
		}
		FreeData(TagData);
	}

	FreeData(FileBuffer);

	if(!CheckValid())
	{
		Clear();	       
		return INTDB_INVALIDDB;
	}

	// Rebuild the hash table
	if(!SetPrimaryHashTable())
	{
		Clear();	       
		return INTDB_CANNOTSETHASHTABLE;
	}
	
	return 0;
}





bool Intdbtable::LoadFromBuffer(char *Buffer, unsigned long BufferSize, int iEncoding, bool bUnicode)
{
	if(Buffer == NULL)
		return false;

	Clear();


	// The buffer must contain at least the identifier
	if(BufferSize < DBFILE_SIZE_IDENTIFIER)
		return false;
//	if(strlen(Buffer) != DBFILE_SIZE_IDENTIFIER)
//		return false;

	// We check that the identifier is OK
	char Identifier[DBFILE_SIZE_IDENTIFIER];
	memset(Identifier, 0, sizeof(Identifier));
	strncpy(Identifier, Buffer, DBFILE_SIZE_IDENTIFIER-1);
	if(strcmp(Identifier, DBFILE_IDENTIFIER) != 0)
		return false;


	char *FileStream = Buffer + DBFILE_SIZE_IDENTIFIER;
	unsigned char TagType;
	unsigned long TagLength;
	char *TagData;

	Intdbrecord *SecuRecord = NULL;
	Intdbfielddesc *SecuFieldDesc = NULL;
	Intdbfield *SecuField = NULL;
	
	string FileVersion = DBFILE_VERSION;
	int TotalBytes = (int)BufferSize - DBFILE_SIZE_IDENTIFIER;
	int BytesRead = 0;

	int BytesReadInTag;
	
#ifdef NV_SERVER
	if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
		bUnicode = bOldObjectsUnicode; // default object encoding if not not found in file
#endif
	while(1)
	{
		BytesReadInTag = ::ReadNextTag(FileStream, &TagType, &TagLength, &TagData);
		if(BytesReadInTag == -1)
		{
			Clear();
			return false;
		}
		BytesRead += BytesReadInTag;
		FileStream += BytesReadInTag;
		if(BytesRead > TotalBytes)
		{
			Clear();
			return false;
		}
		if(TagType == FILE_DB_END)
			break;

		switch(TagType)
		{
			case FILE_VERSION:
				if(TagData != NULL)
				{
					FileVersion = TagData;
					CleanString(FileVersion);
					// PM Unicode saving
					// The unicode flag is placed here for now for compatibility reasons
					// otherwise the saved filedb (ex registry) written with new Nirva version
					// cannot be read by older nirva versions
					if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
					{
						string::size_type index;
						if((index = FileVersion.rfind(';')) != string::npos)
						{
							string sUnicode = FileVersion.substr(index+1);
							CleanString(sUnicode);
							if(sUnicode == "1")
								bUnicode = true;
							else
								bUnicode = false;
						}
					}
				}
				break;
			case FILE_DB:
				Clear();
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_DB_NAME:
				break;
			case FILE_DB_DESC:
				break;
			case FILE_DB_VERSION:
				break;
			case FILE_TABLE:
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_TABLE_NAME:
				if(TagData != NULL)
				{
					Name = TagData;
					CleanString(Name);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(Name, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(Name, bUnicode, false);
				}
				break;
			case FILE_TABLE_DESC:
				if(TagData != NULL)
				{
					Description = TagData;
					CleanString(Description);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(Description, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(Description, bUnicode, false);
				}
				break;
			case FILE_TABLE_PRIMARY:
				if(TagData != NULL)
				{
					PrimaryName = TagData;
					CleanString(PrimaryName);
					if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
						DecodeUrlString(PrimaryName, bUnicode, false);
					if(iEncoding == ENCODING_MODE_ENCODE)
						EncodeUrlString(PrimaryName, bUnicode, false);
				}
				break;
			case FILE_TABLE_PRIMARY_CASE:
				if(TagData != NULL)
				{
					string sPrimaryCaseSensitive = TagData;
					CleanString(sPrimaryCaseSensitive);
					if(sPrimaryCaseSensitive == "1")
						PrimaryCaseSensitive = true;
					else
						PrimaryCaseSensitive = false;
				}
				break;
			case FILE_TABLE_UNICODE:
				if(TagData != NULL)
				{
					if(iEncoding == ENCODING_MODE_DECODE_FROM_FILE)
					{
						string sUnicode = TagData;
						CleanString(sUnicode);
						if(sUnicode == "1")
							bUnicode = true;
						else
							bUnicode = false;
					}
				}
				break;
			case FILE_DFIELD:
				SecuFieldDesc = NULL;
				SecuFieldDesc = NVNEW Intdbfielddesc;
				if(SecuFieldDesc == NULL)
				{
					Clear();
					FreeData(TagData);
					return false;
				}
				FieldDescriptions.push_back((void *)SecuFieldDesc);
				break;
			case FILE_DFIELD_NAME:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Name = TagData;
						CleanString(SecuFieldDesc->Name);
						if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
							DecodeUrlString(SecuFieldDesc->Name, bUnicode, false);
						if(iEncoding == ENCODING_MODE_ENCODE)
							EncodeUrlString(SecuFieldDesc->Name, bUnicode, false);
					}
				}
				break;
			case FILE_DFIELD_DESC:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Description = TagData;
						CleanString(SecuFieldDesc->Description);
						if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
							DecodeUrlString(SecuFieldDesc->Description, bUnicode, false);
						if(iEncoding == ENCODING_MODE_ENCODE)
							EncodeUrlString(SecuFieldDesc->Description, bUnicode, false);
					}
				}
				break;
			case FILE_DFIELD_TYPE:
				if(SecuFieldDesc != NULL)
					if(TagData != NULL)
						SecuFieldDesc->Type = *TagData; // just one byte here
				break;
			case FILE_RECORD:
				SecuRecord = NULL;
				SecuRecord = NVNEW Intdbrecord;
				if(SecuRecord == NULL)
				{
					Clear();
					FreeData(TagData);
					return false;
				}
				Records.push_back((void *)SecuRecord);
				SecuField = NULL;
				break;
			case FILE_FIELD:
				SecuField = NULL;
				if(SecuRecord != NULL)
				{
					SecuField = NVNEW Intdbfield;
					if(SecuField == NULL)
					{
						Clear();
						FreeData(TagData);
						return false;
					}
					SecuRecord->Fields.push_back((void *)SecuField);
				}
				break;
			case FILE_ENTRY:
				if(SecuField != NULL)
				{
					if(TagData != NULL)
					{
						string Provistr = TagData;
//						CleanString(Provistr);
						if((iEncoding == ENCODING_MODE_DECODE) || (iEncoding == ENCODING_MODE_DECODE_FROM_FILE))
							DecodeUrlString(Provistr, bUnicode, false);
						if(iEncoding == ENCODING_MODE_ENCODE)
							EncodeUrlString(Provistr, bUnicode, false);
						if(Provistr.size())
							SecuField->Entries.push_back(Provistr);
					}
				}
				break;
			default:
				// There is no more default otherwise we cannot add new tags type without creating a compatibility problem
				// if a user installs a new version and comes back to an older version
				//Clear();
				//FreeData(TagData);
				//return false;
				break;
		}
		FreeData(TagData);
	}


	if(!CheckValid())
	{
		Clear();	       
		return false;
	}

	// Rebuild the hash table
	if(!SetPrimaryHashTable())
	{
		Clear();	       
		return false;
	}

	return true;
}


// Check the consistency of a database
bool Intdbtable::CheckValid()
{
	int j;
	Intdbrecord *SecuRecord = NULL;
	Intdbfielddesc *SecuFieldDesc = NULL;

	// The table must have a name
	if(!Name.size())
		return false;

	int NbFields = (int)FieldDescriptions.size();
	for(j=0; j<(int)FieldDescriptions.size(); j++)
	{
		SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[j]);

		// A field must have a name
		if(!SecuFieldDesc->Name.size())
			return false;

		// Check for doubles
		if(GetFieldIndex(SecuFieldDesc->Name.c_str(), j+1) != -1)
			return false;
	}
	
	// Now check that each record has the correct number of fields
	for(j=0; j<(int)Records.size(); j++)
	{
		SecuRecord = (Intdbrecord*)(Records[j]);
		if(SecuRecord->Fields.size() != NbFields)
			return false;
	}

	return true;
}


int Intdbtable::Save(const char *FileName)
{
	if(FileName == NULL)
		return INTDB_INVALIDFILE;
	if(!strlen(FileName))
		return INTDB_INVALIDFILE;

	FILE *stream = Nvfopen(FileName, "wb");
	if(stream == NULL)
		return INTDB_CANNOTOPENFILE;


	char buffer[256];
	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer, DBFILE_IDENTIFIER);
							  
	Intdbtable *SecuTable = this;
	Intdbrecord *SecuRecord;
	Intdbfielddesc *SecuFieldDesc;
	Intdbfield *SecuField;
	
	// Write file identifier
	if(!::WriteTag(FILE_NOTAG, DBFILE_SIZE_IDENTIFIER, buffer, stream)) 
		return INTDB_CANNOTWRITEFILE;

	// Write file version
	// PM Unicode saving
	// The unicode flag is placed here for now for compatibility reasons
	// otherwise the saved filedb (ex registry) written with new Nirva version
	// cannot be read by older nirva versions
	string sValue = DBFILE_VERSION;
#ifdef NV_SERVER
	sValue += ";";
	if(GetMain()->IsUnicode())
		sValue += "1";
	else
		sValue += "0";
#endif   
	if(!::WriteTag(FILE_VERSION, (unsigned long)sValue.length(), sValue.data(), stream)) 
		return INTDB_CANNOTWRITEFILE;


	int j,k,l;

	// Write table tag
	if(!::WriteTag(FILE_TABLE, 0, NULL, stream)) 
		return INTDB_CANNOTWRITEFILE;

	// Write unicode flag
	// PM Unicode saving
	// Commented for now for descending compatibility reasons
	// The unicode flag is added to the version tag
//#ifdef NV_SERVER
//	sValue = "0";
//	if(GetMain()->IsUnicode())
//		sValue = "1";
//	if(!::WriteTag(FILE_TABLE_UNICODE, (unsigned long)sValue.length(), sValue.data(), stream)) 
//		return INTDB_CANNOTWRITEFILE;
//#endif
	// Write table name tag
	if(!::WriteTag(FILE_TABLE_NAME, (unsigned long)Name.length(), Name.data(), stream)) 
		return INTDB_CANNOTWRITEFILE;

	// Write table description tag
	if(!::WriteTag(FILE_TABLE_DESC, (unsigned long)Description.length(), Description.data(), stream)) 
		return INTDB_CANNOTWRITEFILE;

	// Write table primary key tag
	if(!::WriteTag(FILE_TABLE_PRIMARY, (unsigned long)PrimaryName.length(), PrimaryName.data(), stream)) 
		return INTDB_CANNOTWRITEFILE;

	// Write table primary key case sensitive tag
	sValue = "0";
	if(PrimaryCaseSensitive)
		sValue = "1";
	if(!::WriteTag(FILE_TABLE_PRIMARY_CASE, (unsigned long)sValue.length(), sValue.data(), stream)) 
		return INTDB_CANNOTWRITEFILE;
	
	
	// Now, write field descriptions
	for(j=0; j<(int)FieldDescriptions.size(); j++)
	{
		SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[j]);

		// Write field description tag
		if(!::WriteTag(FILE_DFIELD, 0, NULL, stream)) 
			return INTDB_CANNOTWRITEFILE;

		// Write field description name tag
		if(!::WriteTag(FILE_DFIELD_NAME, (unsigned long)(SecuFieldDesc->Name.length()), SecuFieldDesc->Name.data(), stream)) 
			return INTDB_CANNOTWRITEFILE;

		// Write field description description tag
		if(!::WriteTag(FILE_DFIELD_DESC, (unsigned long)(SecuFieldDesc->Description.length()), SecuFieldDesc->Description.data(), stream)) 
			return INTDB_CANNOTWRITEFILE;

		// Write field description type tag
		if(!::WriteTag(FILE_DFIELD_TYPE, 1, &(SecuFieldDesc->Type), stream)) 
			return INTDB_CANNOTWRITEFILE;
	}
	
	
	// Now, write records
	for(j=0; j<(int)Records.size(); j++)
	{
		SecuRecord = (Intdbrecord*)(Records[j]);

		// Write record tag
		if(!::WriteTag(FILE_RECORD, 0, NULL, stream)) 
			return INTDB_CANNOTWRITEFILE;

		
		// Now, write fields
		for(k=0; k<(int)(SecuRecord->Fields.size()); k++)
		{
			SecuField = (Intdbfield*)(SecuRecord->Fields[k]);

			// Write field tag
			if(!::WriteTag(FILE_FIELD, 0, NULL, stream)) 
				return INTDB_CANNOTWRITEFILE;

		
			// Now, write fields lines (entries)
			for(l=0; l<(int)(SecuField->Entries.size()); l++)
			{
				// Write field entry tag
				if(!::WriteTag(FILE_ENTRY, (unsigned long)(SecuField->Entries[l].length()), SecuField->Entries[l].data(), stream)) 
					return INTDB_CANNOTWRITEFILE;
			}
		}
	}

	// Write end file tag
	if(!::WriteTag((char)FILE_DB_END, 0, NULL, stream)) 
		return INTDB_CANNOTWRITEFILE;

	fclose(stream);	     
	return 0;
}


// This function allocates a buffer containing all table data as written in the table file
char* Intdbtable::SaveToBuffer(unsigned long *BufferSize)
{
	*BufferSize = 0;
	char *ReturnBuffer = NULL;
	char *Buffer = NULL;
	unsigned long BufferSizeHere = 0;

	char buffer[256];
	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer, DBFILE_IDENTIFIER);
							  
	Intdbtable *SecuTable = this;
	Intdbrecord *SecuRecord;
	Intdbfielddesc *SecuFieldDesc;
	Intdbfield *SecuField;
	

	// We must first calculate the buffer size
	
	// File identifier
	BufferSizeHere = DBFILE_SIZE_IDENTIFIER;

	// Version
	BufferSizeHere += (unsigned long)strlen(DBFILE_VERSION)+3;
	// PM Unicode saving
	// table unicode tag in the version
#ifdef NV_SERVER
	BufferSizeHere += 2;
#endif

	int j,k,l;

	// table tag size
	BufferSizeHere += 3;

	// PM Unicode saving
	// table unicode tag
	// Currently saved with file version tag
//#ifdef NV_SERVER
//	BufferSizeHere += 1+3;
//#endif

	// table name tag
	BufferSizeHere += (unsigned long)Name.length()+3;

	// table description tag
	BufferSizeHere += (unsigned long)Description.length()+3;

	// table primary tag
	BufferSizeHere += (unsigned long)PrimaryName.length()+3;

	// table PrimaryCaseSensitive tag
	BufferSizeHere += 1+3;
	
	// Now, field descriptions
	for(j=0; j<(int)FieldDescriptions.size(); j++)
	{
		SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[j]);

		// field description tag
		BufferSizeHere += 3;

		// field description name tag
		BufferSizeHere += (unsigned long)(SecuFieldDesc->Name.length()+3);

		// field description description tag
		BufferSizeHere += (unsigned long)(SecuFieldDesc->Description.length()+3);

		// field description type tag
		BufferSizeHere += 1+3;
	}
	
	
	// Now, write records
	for(j=0; j<(int)Records.size(); j++)
	{
		SecuRecord = (Intdbrecord*)(Records[j]);

		// record tag
		BufferSizeHere += 3;

		
		// Now, fields
		for(k=0; k<(int)(SecuRecord->Fields.size()); k++)
		{
			SecuField = (Intdbfield*)(SecuRecord->Fields[k]);

			// field tag
			BufferSizeHere += 3;

		
			// Now, fields lines (entries)
			for(l=0; l<(int)(SecuField->Entries.size()); l++)
			{
				// field entry tag
				int EntryLength = (int)(SecuField->Entries[l].length());
				BufferSizeHere += (unsigned long)(SecuField->Entries[l].length() + 3);
				if(EntryLength >= 32768)
					BufferSizeHere += 4;
			}
		}
	}

	// end file tag
	BufferSizeHere += 3;


	int TotalBytes = 0;
	int NbBytes = 0;

	Buffer = (char *)malloc(BufferSizeHere);
	if(Buffer == NULL)
		return NULL;

	ReturnBuffer = Buffer;

	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer, DBFILE_IDENTIFIER);
							  
	// Write file identifier
	NbBytes = DBFILE_SIZE_IDENTIFIER;
	if(!::WriteBufferTag(FILE_NOTAG, NbBytes, buffer, Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes;
	Buffer += NbBytes;

	// Write file version
	// PM Unicode saving
	// The unicode flag is placed here for now for compatibility reasons
	// otherwise the saved filedb (ex registry) written with new Nirva version
	// cannot be read by older nirva versions

	strcpy(buffer, DBFILE_VERSION);
#ifdef NV_SERVER
	strcat(buffer, ";");
	if(GetMain()->IsUnicode())
		strcat(buffer, "1");
	else
		strcat(buffer, "0");
#endif   
	NbBytes = (int)strlen(buffer);
	if(!::WriteBufferTag(FILE_VERSION, NbBytes, buffer, Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;


	// Write table tag
	NbBytes = 0;
	if(!::WriteBufferTag(FILE_TABLE, NbBytes, NULL, Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;

	string sValue;
	// Write unicode flag
	// Write unicode flag
	// PM Unicode saving
	// Commented for now for descending compatibility reasons
	// The unicode flag is added to the version tag
//	sValue = "0";
//#ifdef NV_SERVER
//	if(GetMain()->IsUnicode())
//		sValue = "1";
//	NbBytes = sValue.length();
//	if(!::WriteBufferTag(FILE_TABLE_UNICODE, NbBytes, sValue.data(), Buffer)) 
//	{
//		free(ReturnBuffer);
//		return NULL;
//	}
//	TotalBytes += NbBytes+3;
//	Buffer += NbBytes+3;
//#endif
	
	// Write table name tag
	NbBytes = (int)Name.length();
	if(!::WriteBufferTag(FILE_TABLE_NAME, (unsigned long)NbBytes, Name.data(), Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;

	// Write table description tag
	NbBytes = (int)Description.length();
	if(!::WriteBufferTag(FILE_TABLE_DESC, (unsigned long)NbBytes, Description.data(), Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;


	NbBytes = (int)PrimaryName.length();
	// Write table primary key tag
	if(!::WriteBufferTag(FILE_TABLE_PRIMARY, (unsigned long)NbBytes, PrimaryName.data(), Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;

	// Write table primary key case sensitive tag
	sValue = "0";
	if(PrimaryCaseSensitive)
		sValue = "1";
	NbBytes = (int)sValue.length();
	if(!::WriteBufferTag(FILE_TABLE_PRIMARY_CASE, (unsigned long)NbBytes, sValue.data(), Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;

	
	// Now, write field descriptions
	for(j=0; j<(int)FieldDescriptions.size(); j++)
	{
		SecuFieldDesc = (Intdbfielddesc*)(FieldDescriptions[j]);

		// Write field description tag
		NbBytes = 0;
		if(!::WriteBufferTag(FILE_DFIELD, (unsigned long)NbBytes, NULL, Buffer)) 
		{
			free(ReturnBuffer);
			return NULL;
		}
		TotalBytes += NbBytes+3;
		Buffer += NbBytes+3;

		// Write field description name tag
		NbBytes = (int)(SecuFieldDesc->Name.length());
		if(!::WriteBufferTag(FILE_DFIELD_NAME, (unsigned long)NbBytes, SecuFieldDesc->Name.data(), Buffer)) 
		{
			free(ReturnBuffer);
			return NULL;
		}
		TotalBytes += NbBytes+3;
		Buffer += NbBytes+3;

		// Write field description description tag
		NbBytes = (int)(SecuFieldDesc->Description.length());
		if(!::WriteBufferTag(FILE_DFIELD_DESC, (unsigned long)NbBytes, SecuFieldDesc->Description.data(), Buffer)) 
		{
			free(ReturnBuffer);
			return NULL;
		}
		TotalBytes += NbBytes+3;
		Buffer += NbBytes+3;

		// Write field description type tag
		NbBytes = 1;
		if(!::WriteBufferTag(FILE_DFIELD_TYPE, (unsigned long)NbBytes, &(SecuFieldDesc->Type), Buffer)) 
		{
			free(ReturnBuffer);
			return NULL;
		}
		TotalBytes += NbBytes+3;
		Buffer += NbBytes+3;
	}
	
	
	// Now, write records
	for(j=0; j<(int)Records.size(); j++)
	{
		SecuRecord = (Intdbrecord*)(Records[j]);

		// Write record tag
		NbBytes = 0;
		if(!::WriteBufferTag(FILE_RECORD, (unsigned long)NbBytes, NULL, Buffer)) 
		{
			free(ReturnBuffer);
			return NULL;
		}
		TotalBytes += NbBytes+3;
		Buffer += NbBytes+3;

		
		// Now, write fields
		for(k=0; k<(int)(SecuRecord->Fields.size()); k++)
		{
			SecuField = (Intdbfield*)(SecuRecord->Fields[k]);

			// Write field tag
			NbBytes = 0;
			if(!::WriteBufferTag(FILE_FIELD, (unsigned long)NbBytes, NULL, Buffer)) 
			{
				free(ReturnBuffer);
				return NULL;
			}
			TotalBytes += NbBytes+3;
			Buffer += NbBytes+3;

		
			// Now, write fields lines (entries)
			for(l=0; l<(int)(SecuField->Entries.size()); l++)
			{
				// Write field entry tag
				NbBytes = (int)(SecuField->Entries[l].length());
				if(!::WriteBufferTag(FILE_ENTRY, (unsigned long)NbBytes, SecuField->Entries[l].data(), Buffer)) 
				{
					free(ReturnBuffer);
					return NULL;
				}
				TotalBytes += NbBytes+3;
				Buffer += NbBytes+3;
				if(NbBytes >= 32768)
				{
					TotalBytes += 4;
					Buffer += 4;
				}
			}
		}
	}

	// Write end file tag
	NbBytes = 0;
	if(!::WriteBufferTag((char)FILE_DB_END, (unsigned long)NbBytes, NULL, Buffer)) 
	{
		free(ReturnBuffer);
		return NULL;
	}
	TotalBytes += NbBytes+3;
	Buffer += NbBytes+3;

	if(TotalBytes != BufferSizeHere)
	{
		free(ReturnBuffer);
		return NULL;
	}

	*BufferSize = BufferSizeHere;
	return ReturnBuffer;
}



int Intdbtable::Import(const char* ImportFile, bool Append, const char* FieldSeparator, const char* EntrySeparator, int MaxLineLength, int Offset, int SkipLines, bool WithLf)
{

	if(ImportFile == NULL)
	{
		ClearData();
		return 0;
	}

	if(strlen(ImportFile) == 0)
	{
		ClearData();
		return 0;
	}

	if(FieldSeparator == NULL)
		return INTDB_INVALIDSEPARATOR;
	if(strlen(FieldSeparator) == 0)
		return INTDB_INVALIDSEPARATOR;

	if(EntrySeparator == NULL)
		return INTDB_INVALIDSEPARATOR;
	if(strlen(EntrySeparator) == 0)
		return INTDB_INVALIDSEPARATOR;

	if(Offset < 0)
		Offset = 0;

	if(SkipLines < 0)
		SkipLines = 0;

	FILE *stream = Nvfopen(ImportFile, "rb");
	if(stream == NULL)
		return INTDB_CANNOTOPENFILE;

	if(Offset > 0)
		fseek(stream, Offset, SEEK_SET);

	if(!Append)
		ClearData();

	string line = "";

	char *Buffer = NVNEW char[MaxLineLength];
	if(Buffer == NULL)
		return INTDB_MEMORYERROR;
	bool AppendLine = false;	// set to true when the last character of the previous line is '\'
	int LineLength = 0;
	int i;
	Intdbrecord *SecuRecord = NULL;
	Intdbfield *SecuField = NULL;
	int FieldSeparatorLength = (int)strlen(FieldSeparator);
	int EntrySeparatorLength = (int)strlen(EntrySeparator);
	int NumLines = 0;
	string PrimaryValue = "";
	string::size_type index;
	while(freadline(Buffer, MaxLineLength, stream))
	{
		NumLines++;
		if(NumLines <= SkipLines)
			continue;

		PrimaryValue = "";

//		CleanString(Buffer);
		LineLength = (int)strlen(Buffer);
		if(!AppendLine)
			line = "";
		AppendLine = false;
		if(LineLength && Buffer[LineLength-1] == '\\')
		{
			// We must count the number of antislashes
			int Numanti = 1;
			while((LineLength>Numanti) && (Buffer[LineLength-Numanti-1]) == '\\')
				Numanti++;
			if((!WithLf) || (Numanti%2))
			{
				// If we have an impair number of antislashes, the last one is for line continuing
				AppendLine = true;
				Buffer[LineLength-1] = ' ';
				line += Buffer;
				continue;
			}
		}
		line += Buffer;
		// remove carriage returns
		while((index = line.find('\r')) != string::npos)
			line.erase(index, 1);
		if(WithLf)
		{
			// Replace all strings "\\" by "@@$$NV_ANTISLASH$$@@"
			while((index = line.find("\\\\")) != string::npos)
			{
				string Provistr = line.substr(0, index);
				Provistr += "@@$$NV_ANTISLASH$$@@";
				Provistr += line.substr(index+2);
				line = Provistr;
			}
			// Replace all strings "\n" by a carriage return
			while((index = line.find("\\n")) != string::npos)
			{
				string Provistr = line.substr(0, index);
				Provistr += "\n";
				Provistr += line.substr(index+2);
				line = Provistr;
			}
			// Replace all strings "@@$$NV_ANTISLASH$$@@" by "\"
			while((index = line.find("@@$$NV_ANTISLASH$$@@")) != string::npos)
			{
				string Provistr = line.substr(0, index);
				Provistr += "\\";
				Provistr += line.substr(index+20);
				line = Provistr;
			}
		}
		else
		{
			// Replace line feeds with space
			while((index = line.find("\n")) != string::npos)
				line.replace(index, 1, " ");
		}

		SecuRecord = NVNEW Intdbrecord;
		if(SecuRecord == NULL)
		{
			delete Buffer;
			ClearData();
			return INTDB_MEMORYERROR;
		}
		vector<string> FieldsInLine;
		string ActualField;
		while(FieldsInLine.size() < FieldDescriptions.size())
		{
			ActualField = line;
			if((index = line.find(FieldSeparator)) != string::npos)
			{
				ActualField = line.substr(0, index);
				line = line.substr(index+FieldSeparatorLength);
			}
			else
				line = "";
//			CleanString(ActualField);
			FieldsInLine.push_back(ActualField);
		}

		// Now all the field contents are in FieldsInLine
		for(i=0; i<(int)FieldDescriptions.size(); i++)
		{
			SecuField = NVNEW Intdbfield;
			if(SecuField == NULL)
			{
				delete Buffer;
				ClearData();
				return INTDB_MEMORYERROR;
			}
			SecuRecord->Fields.push_back((void *)SecuField);

			// we must now write the field entries
			ActualField = FieldsInLine[i];
			string ActualFieldEntry;
			bool bPrimaryColumn = false;
			if(IsPrimaryColumn(i))
				bPrimaryColumn = true;
			while(1)
			{
				if((index = ActualField.find(EntrySeparator)) == string::npos)
				{
//					CleanString(ActualField);
					if(ActualField.length())
					{
						if(ActualField == "NV_EMPTY")
							ActualField = "";
						SecuField->Entries.push_back(ActualField);
					}
					if(bPrimaryColumn)
						PrimaryValue = ActualField;
					break;
				}

				ActualFieldEntry = ActualField.substr(0, index);
//				CleanString(ActualFieldEntry);
				ActualField = ActualField.substr(index+EntrySeparatorLength);
				if(bPrimaryColumn)
					continue;
				if(ActualFieldEntry.length())
				{
					if(ActualFieldEntry == "NV_EMPTY")
						ActualFieldEntry = "";
					SecuField->Entries.push_back(ActualFieldEntry);
				}
			}
		}
		if(PrimaryValue.length())
		{
			// If there is a primary key, we import only values that are not yet in the table
			if(GetPrimary(PrimaryValue) == -1)
			{
				Records.push_back((void *)SecuRecord);
				SetPrimary(PrimaryValue, (int)Records.size() - 1);
			}
		}
		else
			Records.push_back((void *)SecuRecord);

	}

	// We have finished with the file
	fclose(stream);	     
	delete Buffer;

	return 0;
}


int Intdbtable::Export(const char* ExportFile, const char* FieldSeparator, const char* EntrySeparator, int MaxLineLength, bool WithLf)
{
	if(ExportFile == NULL)
		return INTDB_INVALIDFILE;

	if(strlen(ExportFile) == 0)
		return INTDB_INVALIDFILE;

	if(FieldSeparator == NULL)
		return INTDB_INVALIDSEPARATOR;
	if(strlen(FieldSeparator) == 0)
		return INTDB_INVALIDSEPARATOR;

	if(EntrySeparator == NULL)
		return INTDB_INVALIDSEPARATOR;
	if(strlen(EntrySeparator) == 0)
		return INTDB_INVALIDSEPARATOR;

	FILE *stream = Nvfopen(ExportFile, "wb");
	if(stream == NULL)
		return INTDB_CANNOTOPENFILE;

	Intdbrecord *SecuRecord = NULL;
	Intdbfield *SecuField = NULL;
	int i, j, k;
	int LineLength;
	int FieldSeparatorLength = (int)strlen(FieldSeparator);
	int EntrySeparatorLength = (int)strlen(EntrySeparator);
	for(i=0; i<(int)Records.size(); i++)
	{
		LineLength = 0;
		// We first write a new line character
		if(i>0)
		{
			if(fwrite("\n", 1, 1, stream) != 1)
			{
				fclose(stream);	       
				return INTDB_CANNOTWRITEFILE;
			}
		}
		SecuRecord = (Intdbrecord *)(Records[i]);
		for(j=0; j<(int)(SecuRecord->Fields.size()); j++)
		{
			SecuField = (Intdbfield *)(SecuRecord->Fields[j]);
			// We first write a new field separator
			if(j>0)
			{
				// We check for line length and sets a new line if necesary
				if(LineLength+1 > MaxLineLength)
				{
					if(fwrite("\\\n", 1, 2, stream) != 2)
					{
						fclose(stream);	       
						return INTDB_CANNOTWRITEFILE;
					}
					LineLength = 0;
				}
				if(fwrite(FieldSeparator, 1, FieldSeparatorLength, stream) != FieldSeparatorLength)
				{
					fclose(stream);	       
					return INTDB_CANNOTWRITEFILE;
				}
				LineLength +=FieldSeparatorLength;
			}
			for(k=0; k<(int)(SecuField->Entries.size()); k++)
			{
				// We first write a new entry separator
				if(k>0)
				{
					// We check for line length and sets a new line if necesary
					if(LineLength+1 > MaxLineLength)
					{
						if(fwrite("\\\n", 1, 2, stream) != 2)
						{
							fclose(stream);	       
							return INTDB_CANNOTWRITEFILE;
						}
						LineLength = 0;
					}
					if(fwrite(EntrySeparator, 1, EntrySeparatorLength, stream) != EntrySeparatorLength)
					{
						fclose(stream);	       
						return INTDB_CANNOTWRITEFILE;
					}
					LineLength +=EntrySeparatorLength;
				}
				// now we can write the field entry
				string Entry = SecuField->Entries[k];
				string::size_type index;
				// remove carriage returns
				while((index = Entry.find('\r')) != string::npos)
					Entry.erase(index, 1);
				if(WithLf)
				{
					// Double all antislashes
					while((index = Entry.find("\\")) != string::npos)
					{
						string Provistr = Entry.substr(0, index);
						Provistr += "@@$$NV_ANTISLASH$$@@";
						Provistr += Entry.substr(index+1);
						Entry = Provistr;
					}
					// Replace all carriage return by the string "\n"
					while((index = Entry.find("\n")) != string::npos)
					{
						string Provistr = Entry.substr(0, index);
						Provistr += "\\n";
						Provistr += Entry.substr(index+1);
						Entry = Provistr;
					}
					// Replace all strings "@@$$NV_ANTISLASH$$@@" by "\\"
					while((index = Entry.find("@@$$NV_ANTISLASH$$@@")) != string::npos)
					{
						string Provistr = Entry.substr(0, index);
						Provistr += "\\\\";
						Provistr += Entry.substr(index+20);
						Entry = Provistr;
					}
				}
				else
				{
					// Replace line feeds with space
					while((index = Entry.find("\n")) != string::npos)
						Entry.replace(index, 1, " ");
				}
				if(LineLength+(int)Entry.length() > MaxLineLength)
				{
					if(fwrite("\\\n", 1, 2, stream) != 2)
					{
						fclose(stream);	       
						return INTDB_CANNOTWRITEFILE;
					}
					LineLength = 0;
				}
				if(fwrite((void *)(Entry.data()), 1, Entry.length(), stream) != Entry.length())
				{
					fclose(stream);	       
					return INTDB_CANNOTWRITEFILE;
				}
				LineLength += (int)Entry.length();
			}
		}
	}

	// We have finished with the file
	fclose(stream);	       
	return 0;
}



int Intdbtable::Modify(const char* Query, int FieldIndex, const char* FieldValue)
{
	if(Query == NULL)
		return INTDB_INVALIDQUERY;
	
	if(!strlen(Query))
		return INTDB_INVALIDQUERY;

	if(FieldIndex >= (int)FieldDescriptions.size())
		return INTDB_INVALIDFIELD;

	if(FieldIndex < 0)
		return INTDB_INVALIDFIELD;

	if(IsPrimaryColumn(FieldIndex))
		return INTDB_NOTPOSSIBLEONPRIMARY;
	
	char *QueryHere = (char *)malloc(strlen(Query) + 1);
	strcpy(QueryHere, Query);
	strupr(QueryHere);
	CleanString(QueryHere);

	string sQueryHere = QueryHere;

	FreeData(QueryHere);

	string::size_type index;

	// Replace "= " with "=" 
	while((index = sQueryHere.find("= ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace "> " with ">" 
	while((index = sQueryHere.find("> ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace "< " with "<" 
	while((index = sQueryHere.find("< ")) !=  string::npos)
		sQueryHere.erase(index+1, 1);

	// Replace " =" with "=" 
	while((index = sQueryHere.find(" =")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " >" with ">" 
	while((index = sQueryHere.find(" >")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " <" with "<" 
	while((index = sQueryHere.find(" <")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " !" with "!" 
	while((index = sQueryHere.find(" !")) !=  string::npos)
		sQueryHere.erase(index, 1);

	// Replace " AND " with "@@AN" 
	while((index = sQueryHere.find(" AND ")) !=  string::npos)
	{
		string provistr = sQueryHere.substr(0, index);
		provistr += "@@AN";
		provistr += sQueryHere.substr(index+5);
		sQueryHere = provistr;
	}

	// Replace " OR " with "@@OR" 
	while((index = sQueryHere.find(" OR ")) !=  string::npos)
	{
		string provistr = sQueryHere.substr(0, index);
		provistr += "@@OR";
		provistr += sQueryHere.substr(index+4);
		sQueryHere = provistr;
	}

	vector<string> Operand1s;
	vector<string> Operand2s;
	vector<int> Operators;
	vector<int> Combinators;


	int i, j;
	Intdbrecord *SecuRecord;
	Intdbfield *FieldToModify;

	string FieldValueHere = "";
	if(FieldValue != NULL)
		FieldValueHere = FieldValue;

	
	// Check for search all string
	if(sQueryHere == "*")
	{
		for(i=0; i<(int)Records.size(); i++)
		{
			SecuRecord = (Intdbrecord *)(Records[i]);
			if(SecuRecord == NULL)
				continue;
			FieldToModify = SecuRecord->GetField(FieldIndex);
			if(FieldToModify == NULL)
				continue;
			FieldToModify->Clear();
			if(FieldValueHere.length())
				FieldToModify->Entries.push_back(FieldValueHere);
		}
		return 0;
	}

	while(1)
	{
		int Combinator = SEARCH_COMBINATOR_NONE;
		string Expression = "";
		if(((index = sQueryHere.find("@@")) != string::npos) && (sQueryHere.length() > index+4))
		{
			Expression = sQueryHere.substr(0, index);
			if((sQueryHere[index+2] == 'A') && (sQueryHere[index+3] == 'N'))
				Combinator = SEARCH_COMBINATOR_AND;
			if((sQueryHere[index+2] == 'O') && (sQueryHere[index+3] == 'R'))
				Combinator = SEARCH_COMBINATOR_OR;
			sQueryHere  = sQueryHere.substr(index+4);
		}
		else
			Expression = sQueryHere;

		int Operator = SEARCH_OPERATOR_NONE;
		int OperatorLength = 1;
		if((index = Expression.find("!=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_NE;
			OperatorLength = 2;
		}
		else if((index = Expression.find("<=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_STE;
			OperatorLength = 2;
		}
		else if((index = Expression.find(">=")) != string::npos)
		{
			Operator = SEARCH_OPERATOR_GTE;
			OperatorLength = 2;
		}
		else if((index = Expression.find("<")) != string::npos)
			Operator = SEARCH_OPERATOR_ST;
		else if((index = Expression.find(">")) != string::npos)
			Operator = SEARCH_OPERATOR_GT;
		else if((index = Expression.find("=")) != string::npos)
			Operator = SEARCH_OPERATOR_E;
		
		if(Operator == SEARCH_OPERATOR_NONE)
			return INTDB_INVALIDQUERY;

		string Operand1 = Expression.substr(0, index);
		string Operand2 = Expression.substr(index+OperatorLength);

		CleanString(Operand1);
		CleanString(Operand2);

		if((!Operand1.length()) || (!Operand1.length()))
			return INTDB_INVALIDQUERY;

		Operand1s.push_back(Operand1);
		Operand2s.push_back(Operand2);
		Operators.push_back(Operator);
		Combinators.push_back(Combinator);

		if(Combinator == SEARCH_COMBINATOR_NONE)
			break;
	}

	
	// Now, we make the search
	Intdbfield *SecuField;
	Intdbfielddesc *SecuFieldDesc;
	for(i=0; i<(int)Records.size(); i++)
	{
		SecuRecord = (Intdbrecord *)(Records[i]);
		FieldToModify = SecuRecord->GetField(FieldIndex);
		if(FieldToModify == NULL)
			continue;

		for(j=0; j<(int)Operators.size(); j++)
		{
			int Operator = Operators[j];
			int Combinator = Combinators[j];
			string Operand1 = Operand1s[j];
			string Operand2 = Operand2s[j];
			int FieldIndexHere = GetFieldIndex(Operand1.c_str());
			if(FieldIndexHere == -1)	// The field must exist
				return INTDB_INVALIDQUERY;
			SecuField = SecuRecord->GetField(FieldIndexHere);
			SecuFieldDesc = (Intdbfielddesc *)(FieldDescriptions[FieldIndexHere]);
			bool Numeric = false;
			if(SecuFieldDesc->Type == FIELD_TYPE_NUMERIC)
				Numeric = true;

			bool Verify = SecuField->Verify(Operand2.c_str(), Operator, Numeric);
			if(!Verify)	// If the field doesn't verify the request, we go to the next expression
			{
				if(Combinators[j] == SEARCH_COMBINATOR_OR)
					continue;
				else
					break;
			}

//			if((j>0) && (Combinators[j-1] == SEARCH_COMBINATOR_OR))
			if(Combinators[j] == SEARCH_COMBINATOR_OR)
			{
				// If we have a OR operator, we modify the record and go to the next record
				// This avoid to verify other requests
				FieldToModify->Clear();
				if(FieldValueHere.length())
					FieldToModify->Entries.push_back(FieldValueHere);
				break;
			}
			
			if(j == Operators.size() - 1)
			{
				// If we are here, the record satisfies the request
				FieldToModify->Clear();
				if(FieldValueHere.length())
					FieldToModify->Entries.push_back(FieldValueHere);
				break;
			}

		}
	}

	return 0;
}



int Intdbtable::Modify(const char* Query, const char* FieldName, const char* FieldValue)
{
	int FieldIndexToSet = GetFieldIndex(FieldName);
	if(FieldIndexToSet == -1)
		return INTDB_INVALIDFIELD;

	return Modify(Query, FieldIndexToSet, FieldValue);
}




void Intdbtable::ClearCell(int row_index, int col_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return;

	if(IsPrimaryColumn(col_index))
		RemovePrimary(GetRecordPrimaryValue(row_index));

	Field->Clear();
}


void Intdbtable::SortCell(int row_index, int col_index, bool Ascending)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return;

	// If primary column, only one line is allowed so ther is no possible sort
	if(IsPrimaryColumn(col_index))
		return;

	Intdbfielddesc *FieldDesc = (Intdbfielddesc *)(FieldDescriptions[col_index]);
	if(FieldDesc == NULL)
		return;
	bool TypeNumeric = false;
	if(FieldDesc->Type == FIELD_TYPE_NUMERIC)
		TypeNumeric = true;


	Field->Sort(TypeNumeric, Ascending);
}


int Intdbtable::GetNumCellLines(int row_index, int col_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return -1;

	return Field->GetSize();

}


string Intdbtable::GetCellLine(int row_index, int col_index, int line_index, string Default)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return Default;

	string Result = Field->GetEntry(line_index, Default.c_str());

	return Result;
}


void Intdbtable::GetCellLine(int row_index, int col_index, char *Buffer, int BufferSize, int line_index, const char *Default)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	if(Buffer == NULL)
		return;
	if(!BufferSize)
		return;
	Buffer[0] = 0;

	string sDefault = "";
	if(Default != NULL)
		sDefault = Default;
	if(((int)sDefault.length() + 1) > BufferSize)
		sDefault.erase(BufferSize-1);
	strcpy(Buffer, sDefault.c_str());

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return;

	string Result = Field->GetEntry(line_index, sDefault.c_str());
	
	if(((int)Result.length() + 1) > BufferSize)
		Result.erase(BufferSize-1);
	strcpy(Buffer, Result.c_str());
}

bool Intdbtable::SetCellLine(int row_index, int col_index, string Line, int line_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;


	return SetCellLine(row_index, col_index, Line.c_str(), line_index);
}


bool Intdbtable::SetCellLine(int row_index, int col_index, const char* Line, int line_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return false;

	if(line_index >= (int)(Field->Entries.size()))
	{
		if(!InsertCellLine(row_index, col_index, ""))	// We add an empty new line if not existing
			return false;
		line_index = (int)(Field->Entries.size() - 1);
	}
	if(line_index < 0)
		return false;
	if(Line == NULL)
		return false;

	bool bSetPrimary = false;
	string OldPrimaryValue = Field->GetEntry(line_index, "NV_INTERNAL_NO_VALUE");
	if(IsPrimaryColumn(col_index))
	{
		// Only one line for the primary key
		if(line_index != 0)
			return false;

		int PrimaryIndex = GetPrimary(Line);
			if((PrimaryIndex != -1) && (PrimaryIndex != row_index))
				return false;	// cannot set the primary key because it ever exists
		bSetPrimary = true;
	}

	Field->SetEntry(line_index, Line);

	if(bSetPrimary)
	{
		if(OldPrimaryValue != "NV_INTERNAL_NO_VALUE")
			RemovePrimary(OldPrimaryValue);
		SetPrimary(Line, row_index);
	}

	return true;
}

bool Intdbtable::InsertCellLine(int row_index, int col_index, string Line, int line_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	return InsertCellLine(row_index, col_index, Line.c_str(), line_index);
}

bool Intdbtable::InsertCellLine(int row_index, int col_index, const char* Line, int line_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return false;

	if(Line == NULL)
		return false;

	bool bSetPrimary = false;
	if(IsPrimaryColumn(col_index))
	{
		// Only one line for the primary key
		if(Field->Entries.size())
			return false;

		int PrimaryIndex = GetPrimary(Line);
		if(PrimaryIndex != -1)
			return false;	// cannot set the primary key because it ever exists
		bSetPrimary = true;
	}

	Field->InsertEntry(Line, line_index);
	
	if(bSetPrimary)
		SetPrimary(Line, row_index);

	return true;
}

void Intdbtable::RemoveCellLine(int row_index, int col_index, int line_index)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return;
	if(line_index < 0)
		return;
	if(line_index >= (int)(Field->Entries.size()))
		return;

	if(line_index == -1)
		line_index = Field->GetSize() - 1;

	if(IsPrimaryColumn(col_index))
		RemovePrimary(GetRecordPrimaryValue(row_index));

	Field->RemoveEntry(line_index);
}

void Intdbtable::SwapCellLines(int row_index, int col_index, int line_index1, int line_index2)
{
	if(row_index == -1)
		row_index = GetSize() - 1;

	Intdbfield* Field = GetField(row_index, col_index);
	if(Field == NULL)
		return;

	if(IsPrimaryColumn(col_index))
		return;		// Cannot swap the primary key cell lines because there is only one cell line

	Field->SwapEntries(line_index1, line_index2);
}



/////////////////////////////////////////////////////////////////////////////
// Intdb

Intdb::Intdb()
{
	Name = "";
	Description = "";
	Modified = false;
	Mutation = false;
	FileName = "";
	Password = "";
	Version = "";
	InSearch = 0;
	InUse = 0;
	ToRemove = 0;
	Admin = false;
	MUTEXCREATE(ProtectInUse);
	MUTEXCREATE(ProtectDb);
	MUTEXCREATE(ProtectMutation);
	MUTEXCREATE(ProtectAdmin);
}

Intdb::~Intdb()
{
	if(Modified)
		Save();
	WaitForMutation();
	Clear();
	ClearMutationMode();
	MUTEXUNLOCK(ProtectMutation);
	MUTEXUNLOCK(ProtectDb);
	MUTEXUNLOCK(ProtectInUse);
	MUTEXUNLOCK(ProtectAdmin);
	MUTEXDESTROY(ProtectMutation);
	MUTEXDESTROY(ProtectDb);
	MUTEXDESTROY(ProtectInUse);
	MUTEXDESTROY(ProtectAdmin);
}

void Intdb::Clear()
{
	Name = "";
	Description = "";
	Modified = false;
	Version = "";
	InSearch = 0;

	int i;
	Intdbtable *SecuTable;
	for(i=0; i<(int)Tables.size(); i++)
	{
		SecuTable = (Intdbtable*)(Tables[i]);
		delete SecuTable;
	}
	Tables.clear();
}

void Intdb::IncInUse()
{
	MUTEXLOCK(ProtectInUse);
	if(InUse < 0)
		InUse = 0;
	InUse++;
	MUTEXUNLOCK(ProtectInUse);
}

void Intdb::DecInUse()
{
	MUTEXLOCK(ProtectInUse);
	if(InUse > 0)
		InUse--;
	else
		InUse = 0;
	MUTEXUNLOCK(ProtectInUse);
}

int Intdb::IsInUse()
{
	int Ret = 0;
	MUTEXLOCK(ProtectInUse);
	if(InUse > 0)
		Ret = 1;
	else
	MUTEXUNLOCK(ProtectInUse);

	return Ret;
}

void Intdb::CopyFrom(Intdb* Source)
{
	
	// The file name is not copied
	WaitForMutation();
	Clear();
	Name = Source->Name;
	Description = Source->Description;
	Modified = true;
	Version = Source->Version;
	Password = Source->Password;

	int i;
	Intdbtable *SecuTable;
	Intdbtable *NewSecuTable;
	for(i=0; i<(int)(Source->Tables.size()); i++)
	{
		NewSecuTable = NVNEW Intdbtable;
		if(NewSecuTable == NULL)
			return;
		SecuTable = (Intdbtable*)(Source->Tables[i]);
		NewSecuTable->CopyFrom(SecuTable);
		Tables.push_back((void *)NewSecuTable);
	}
	ClearMutationMode();
}

Intdbtable* Intdb::GetTable(int index)
{
	if(index >= (int)Tables.size())
		return NULL;
	return (Intdbtable*)(Tables[index]);
}

Intdbtable* Intdb::GetTable(const char* TableName)
{
	int i;
	Intdbtable *SecuTable;
	bool found = false;
	for(i=0; i<(int)Tables.size(); i++)
	{
		SecuTable = (Intdbtable*)(Tables[i]);
		if(strcmpnocase(SecuTable->Name.c_str(), TableName) == 0)
		{
			found = true;
			break;
		}
	}

	if(!found)
		return NULL;
	return SecuTable;
}


int Intdb::GetSize()
{

	WaitForSearch();
	int NumTables = (int)Tables.size();
	ClearSearchMode();
	return NumTables;
}


int Intdb::Load(bool JustHeader)
{

	WaitForMutation();
	Clear();
	
	if(FileName.length() == 0)
	{
		ClearMutationMode();
		return INTDB_INVALIDFILE;
	}

	FILE *stream = Nvfopen(FileName.c_str(), "rb");
	if(stream == NULL)
	{
		ClearMutationMode();
		return INTDB_CANNOTOPENFILE;
	}

	// we get the file size
	fseek(stream, 0, SEEK_END);
	unsigned long FileSize = ftell(stream);
	fseek(stream, 0, SEEK_SET);

	// The file must contain at least the identifier
	if(FileSize <= DBFILE_SIZE_IDENTIFIER)
	{
		fclose(stream);
		ClearMutationMode();
		return INTDB_INVALIDFILE;
	}

	// We check that the identifier is OK
	char Identifier[DBFILE_SIZE_IDENTIFIER];
	memset(Identifier, 0, sizeof(Identifier));
	fread(Identifier, 1, DBFILE_SIZE_IDENTIFIER, stream);
	if(strcmp(Identifier, DBFILE_IDENTIFIER) != 0)
	{
		fclose(stream);	       
		ClearMutationMode();
		return INTDB_INVALIDFILE;
	}

	// and we allocate memory for the file buffer
	char *FileBuffer = (char *)malloc(FileSize - DBFILE_SIZE_IDENTIFIER);
	if(FileBuffer == NULL)
	{
		fclose(stream);	       
		ClearMutationMode();
		return INTDB_MEMORYERROR;
	}

	// Read file
	unsigned long nbBytes = (unsigned long)fread(FileBuffer, 1, FileSize - DBFILE_SIZE_IDENTIFIER, stream);
	if(ferror(stream) || (nbBytes != (FileSize - DBFILE_SIZE_IDENTIFIER)))
	{
		fclose(stream);	       
		FreeData(FileBuffer);
		ClearMutationMode();
		return INTDB_CANNOTREADFILE;
	}

	// We have finished with the file
	fclose(stream);	       

	char *FileStream = FileBuffer;
	unsigned char TagType;
	unsigned long TagLength;
	char *TagData;

	Intdbtable *SecuTable = NULL;
	Intdbrecord *SecuRecord = NULL;
	Intdbfielddesc *SecuFieldDesc = NULL;
	Intdbfield *SecuField = NULL;
	
	string FileVersion = DBFILE_VERSION;
	int TotalBytes = FileSize - DBFILE_SIZE_IDENTIFIER;
	int BytesRead = 0;

	bool NameOk = false;
	bool DescriptionOk = false;
	bool VersionOk = false;
	int BytesReadInTag;

	while(1)
	{
		BytesReadInTag = ReadNextTag(FileStream, &TagType, &TagLength, &TagData);
		if(BytesReadInTag == -1)
		{
			Clear();
			FreeData(FileBuffer);
			ClearMutationMode();
			return INTDB_MEMORYERROR;
		}
		BytesRead += BytesReadInTag;
		FileStream += BytesReadInTag;
		if(BytesRead > TotalBytes)
		{
			Clear();
			FreeData(FileBuffer);
			ClearMutationMode();
			return INTDB_INVALIDFILE;
		}
		if(TagType == FILE_DB_END)
			break;

		if(JustHeader && NameOk && DescriptionOk && VersionOk)
			break;

		switch(TagType)
		{
			case FILE_VERSION:
				if(TagData != NULL)
				{
					FileVersion = TagData;
					CleanString(FileVersion);
				}
				break;
			case FILE_DB:
				Clear();
				SecuTable = NULL;
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_DB_NAME:
				if(TagData != NULL)
				{
					Name = TagData;
					CleanString(Name);
					NameOk = true;
				}
				break;
			case FILE_DB_DESC:
				if(TagData != NULL)
				{
					Description = TagData;
					CleanString(Description);
					DescriptionOk = true;
				}
				break;
			case FILE_DB_VERSION:
				if(TagData != NULL)
				{
					Version = TagData;
					CleanString(Version);
					VersionOk = true;
				}
				break;
			case FILE_TABLE:
				SecuTable = NVNEW Intdbtable;
				if(SecuTable == NULL)
				{
					Clear();
					FreeData(FileBuffer);
					ClearMutationMode();
					FreeData(TagData);
					return INTDB_MEMORYERROR;
				}
				Tables.push_back((void *)SecuTable);
				SecuRecord = NULL;
				SecuFieldDesc = NULL;
				SecuField = NULL;
				break;
			case FILE_TABLE_NAME:
				if(SecuTable != NULL)
				{
					if(TagData != NULL)
					{
						SecuTable->Name = TagData;
						CleanString(SecuTable->Name);
					}
				}
				break;
			case FILE_TABLE_DESC:
				if(SecuTable != NULL)
				{
					if(TagData != NULL)
					{
						SecuTable->Description = TagData;
						CleanString(SecuTable->Description);
					}
				}
				break;
			case FILE_TABLE_PRIMARY:
				if(SecuTable != NULL)
				{
					if(TagData != NULL)
					{
						SecuTable->PrimaryName = TagData;
						CleanString(SecuTable->PrimaryName);
					}
				}
				break;
			case FILE_TABLE_PRIMARY_CASE:
				if(SecuTable != NULL)
				{
					if(TagData != NULL)
					{
						string sPrimaryCaseSensitive;
						CleanString(sPrimaryCaseSensitive);
						if(sPrimaryCaseSensitive == "1")
							SecuTable->PrimaryCaseSensitive = true;
						else
							SecuTable->PrimaryCaseSensitive = false;
					}
				}
				break;
			case FILE_DFIELD:
				SecuFieldDesc = NULL;
				if(SecuTable != NULL)
				{
					SecuFieldDesc = NVNEW Intdbfielddesc;
					if(SecuFieldDesc == NULL)
					{
						Clear();
						FreeData(FileBuffer);
						ClearMutationMode();
						FreeData(TagData);
						return INTDB_MEMORYERROR;
					}
					SecuTable->FieldDescriptions.push_back((void *)SecuFieldDesc);
				}
				break;
			case FILE_DFIELD_NAME:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Name = TagData;
						CleanString(SecuFieldDesc->Name);
					}
				}
				break;
			case FILE_DFIELD_DESC:
				if(SecuFieldDesc != NULL)
				{
					if(TagData != NULL)
					{
						SecuFieldDesc->Description = TagData;
						CleanString(SecuFieldDesc->Description);
					}
				}
				break;
			case FILE_DFIELD_TYPE:
				if(SecuFieldDesc != NULL)
					if(TagData != NULL)
						SecuFieldDesc->Type = *TagData; // just one byte here
				break;
			case FILE_RECORD:
				SecuRecord = NULL;
				if(SecuTable != NULL)
				{
					SecuRecord = NVNEW Intdbrecord;
					if(SecuRecord == NULL)
					{
						Clear();
						FreeData(FileBuffer);
						ClearMutationMode();
						FreeData(TagData);
						return INTDB_MEMORYERROR;
					}
					SecuTable->Records.push_back((void *)SecuRecord);
				}
				SecuField = NULL;
				break;
			case FILE_FIELD:
				SecuField = NULL;
				if(SecuRecord != NULL)
				{
					SecuField = NVNEW Intdbfield;
					if(SecuField == NULL)
					{
						Clear();
						FreeData(FileBuffer);
						ClearMutationMode();
						FreeData(TagData);
						return INTDB_MEMORYERROR;
					}
					SecuRecord->Fields.push_back((void *)SecuField);
				}
				break;
			case FILE_ENTRY:
				if(SecuField != NULL)
				{
					if(TagData != NULL)
					{
						string Provistr = TagData;
//						CleanString(Provistr);
						if(Provistr.size())
							SecuField->Entries.push_back(TagData);
					}
				}
				break;
			default:
				Clear();
				FreeData(FileBuffer);
				ClearMutationMode();
				FreeData(TagData);
				return INTDB_INVALIDFILE;
				break;
		}
		FreeData(TagData);
	}

	FreeData(FileBuffer);

	if(!CheckValid())
	{
		Clear();	       
		ClearMutationMode();
		return INTDB_INVALIDDB;
	}

	ClearMutationMode();
	return 0;
}


// Check the consistency of a database
bool Intdb::CheckValid()
{
	int i, j;
	Intdbtable *SecuTable;
	Intdbrecord *SecuRecord = NULL;
	Intdbfielddesc *SecuFieldDesc = NULL;

	// The database must have a name
	if(!Name.size())
		return false;

	for(i=0; i<(int)Tables.size(); i++)
	{
		SecuTable = (Intdbtable*)(Tables[i]);
		int NbFields = (int)(SecuTable->FieldDescriptions.size());
		for(j=0; j<(int)(SecuTable->FieldDescriptions.size()); j++)
		{
			SecuFieldDesc = (Intdbfielddesc*)(SecuTable->FieldDescriptions[j]);

			// A field must have a name
			if(!SecuFieldDesc->Name.size())
				return false;

			// Check for doubles
			if(SecuTable->GetFieldIndex(SecuFieldDesc->Name.c_str(), j+1) != -1)
				return false;
		}
		
		// Now check that each record has the correct number of fields
		for(j=0; j<(int)(SecuTable->Records.size()); j++)
		{
			SecuRecord = (Intdbrecord*)(SecuTable->Records[j]);
			if(SecuRecord->Fields.size() != NbFields)
				return false;
		}

	}

	return true;
}

int Intdb::Save()
{
	WaitForMutation();
	if(FileName.length() == 0)
	{
		ClearMutationMode();
		return INTDB_INVALIDFILE;
	}

	FILE *stream = Nvfopen(FileName.c_str(), "wb");
	if(stream == NULL)
	{
		ClearMutationMode();
		return INTDB_CANNOTOPENFILE;
	}

	char buffer[256];
	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer, DBFILE_IDENTIFIER);
							  
	Intdbtable *SecuTable;
	Intdbrecord *SecuRecord;
	Intdbfielddesc *SecuFieldDesc;
	Intdbfield *SecuField;
	
	// Write file identifier
	if(!WriteTag(FILE_NOTAG, DBFILE_SIZE_IDENTIFIER, buffer, stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	// Write file version
	strcpy(buffer, DBFILE_VERSION);
	if(!WriteTag(FILE_VERSION, (unsigned long)strlen(buffer), buffer, stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	// Write database tag
	if(!WriteTag(FILE_DB, 0, NULL, stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	// Write database name tag
	if(!WriteTag(FILE_DB_NAME, (unsigned long)Name.length(), Name.data(), stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	// Write database description tag
	if(!WriteTag(FILE_DB_DESC, (unsigned long)Description.length(), Description.data(), stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	// Write database version tag
	if(!WriteTag(FILE_DB_VERSION, (unsigned long)Version.length(), Version.data(), stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	int i,j,k,l;
	for(i=0; i<(int)Tables.size(); i++)
	{
		SecuTable = (Intdbtable*)(Tables[i]);

		// Write table tag
		if(!WriteTag(FILE_TABLE, 0, NULL, stream)) 
		{
			ClearMutationMode();
			return INTDB_CANNOTWRITEFILE;
		}

		// Write table name tag
		if(!WriteTag(FILE_TABLE_NAME, (unsigned long)(SecuTable->Name.length()), SecuTable->Name.data(), stream)) 
		{
			ClearMutationMode();
			return INTDB_CANNOTWRITEFILE;
		}

		// Write table description tag
		if(!WriteTag(FILE_TABLE_DESC, (unsigned long)(SecuTable->Description.length()), SecuTable->Description.data(), stream)) 
		{
			ClearMutationMode();
			return INTDB_CANNOTWRITEFILE;
		}
		

		// Write table primary key tag
		if(!WriteTag(FILE_TABLE_PRIMARY, (unsigned long)(SecuTable->PrimaryName.length()), SecuTable->PrimaryName.data(), stream)) 
		{
			ClearMutationMode();
			return INTDB_CANNOTWRITEFILE;
		}

		// Write table primary key case sensitive tag
		string sValue = "0";
		if(SecuTable->PrimaryCaseSensitive)
			sValue = "1";
		if(!::WriteTag(FILE_TABLE_PRIMARY_CASE, (unsigned long)sValue.length(), sValue.data(), stream)) 
		{
			ClearMutationMode();
			return INTDB_CANNOTWRITEFILE;
		}
		
		// Now, write field descriptions
		for(j=0; j<(int)(SecuTable->FieldDescriptions.size()); j++)
		{
			SecuFieldDesc = (Intdbfielddesc*)(SecuTable->FieldDescriptions[j]);

			// Write field description tag
			if(!WriteTag(FILE_DFIELD, 0, NULL, stream)) 
			{
				ClearMutationMode();
				return INTDB_CANNOTWRITEFILE;
			}

			// Write field description name tag
			if(!WriteTag(FILE_DFIELD_NAME, (unsigned long)(SecuFieldDesc->Name.length()), SecuFieldDesc->Name.data(), stream)) 
			{
				ClearMutationMode();
				return INTDB_CANNOTWRITEFILE;
			}

			// Write field description description tag
			if(!WriteTag(FILE_DFIELD_DESC, (unsigned long)(SecuFieldDesc->Description.length()), SecuFieldDesc->Description.data(), stream)) 
			{
				ClearMutationMode();
				return INTDB_CANNOTWRITEFILE;
			}

			// Write field description type tag
			if(!WriteTag(FILE_DFIELD_TYPE, 1, &(SecuFieldDesc->Type), stream)) 
			{
				ClearMutationMode();
				return INTDB_CANNOTWRITEFILE;
			}
		}
		
		
		// Now, write records
		for(j=0; j<(int)(SecuTable->Records.size()); j++)
		{
			SecuRecord = (Intdbrecord*)(SecuTable->Records[j]);

			// Write record tag
			if(!WriteTag(FILE_RECORD, 0, NULL, stream)) 
			{
				ClearMutationMode();
				return INTDB_CANNOTWRITEFILE;
			}

			
			// Now, write fields
			for(k=0; k<(int)(SecuRecord->Fields.size()); k++)
			{
				SecuField = (Intdbfield*)(SecuRecord->Fields[k]);

				// Write field tag
				if(!WriteTag(FILE_FIELD, 0, NULL, stream)) 
				{
					ClearMutationMode();
					return INTDB_CANNOTWRITEFILE;
				}

			
				// Now, write fields lines (entries)
				for(l=0; l<(int)(SecuField->Entries.size()); l++)
				{
					// Write field entry tag
					if(!WriteTag(FILE_ENTRY, (unsigned long)(SecuField->Entries[l].length()), SecuField->Entries[l].data(), stream)) 
					{
						ClearMutationMode();
						return INTDB_CANNOTWRITEFILE;
					}
				}
			}
		}
	}

	// Write end file tag
	if(!WriteTag((char)FILE_DB_END, 0, NULL, stream)) 
	{
		ClearMutationMode();
		return INTDB_CANNOTWRITEFILE;
	}

	fclose(stream);	     
	Modified = false;
	ClearMutationMode();
	return 0;
}


int Intdb::Info(const char* TableName, Intdbtable **Result)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;
	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	
	WaitForSearch();
	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearSearchMode();
		return INTDB_INVALIDTABLE;
	}

	int iResult = SecuTable->Info(Result);
	ClearSearchMode();
	return iResult;
}




int Intdb::GetTableSize(const char* TableName, int *TableSize)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;
	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	
	WaitForSearch();
	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearSearchMode();
		return INTDB_INVALIDTABLE;
	}

	*TableSize = SecuTable->GetSize();
	ClearSearchMode();
	return 0;
}




int Intdb::Search(const char* TableName, const char* Query, Intdbtable **Result, const char* SortField)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;
	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	
	WaitForSearch();
	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearSearchMode();
		return INTDB_INVALIDTABLE;
	}

	int iResult = SecuTable->Search(Query, Result, SortField);
	ClearSearchMode();
	return iResult;
}


int Intdb::Import(const char* TableName, const char* ImportFile, bool Append, const char* FieldSeparator, const char* EntrySeparator, int MaxLineLength, int Offset, int SkipLines)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	if(Append && GetAdminMode())
		return INTDB_ADMINWARNING;


	// If the table name is CI_ALLTABLES, we import the entire database file
	// This is an undocumented feature
	if(strcmpnocase(TableName, "CI_ALLTABLES") == 0)
	{
		Intdb ProviDb;
		ProviDb.FileName = ImportFile;
		int iResult = ProviDb.Load();
		if(iResult != 0)
			return iResult;
		if(ProviDb.Name != Name)	// They must have the same name
			return INTDB_INVALIDDB;

		// We now importing the database data
		WaitForMutation();
		
		// First clear actual data
		int i;
		Intdbtable *SecuTable;
		for(i=0; i<(int)Tables.size(); i++)
		{
			SecuTable = (Intdbtable*)(Tables[i]);
			delete SecuTable;
		}
		Tables.clear();
		
		// Description and version may be changed
		Description = ProviDb.Description;
		Version = ProviDb.Version;

		// Load new tables
		Intdbtable *NewSecuTable;
		for(i=0; i<(int)ProviDb.Tables.size(); i++)
		{
			NewSecuTable = NVNEW Intdbtable;
			if(NewSecuTable == NULL)
			{
				ClearMutationMode();
				return INTDB_MEMORYERROR;
			}
			SecuTable = (Intdbtable*)(ProviDb.Tables[i]);
			NewSecuTable->CopyFrom(SecuTable);
			Tables.push_back((void *)NewSecuTable);
		}

		ClearMutationMode();

		if(GetAdminMode())
		{
			// If in admin mode, we must apply all modifications made in the 
			// database by users
			MUTEXLOCK(ProtectAdmin);
			int i;

			for(i=0; i<(int)Modifs.size(); i++)
			{
				string Value = Modifs[i];
				string TableInValue = "";
				string QueryInValue = "";
				string FieldInValue = "";
				string ValueInValue = "";
				string::size_type index;

				index = Value.find("@;@");
				if(index == string::npos)
					continue;
				TableInValue = Value.substr(0, index);
				Value = Value.substr(index+3);

				index = Value.find("@;@");
				if(index == string::npos)
					continue;
				QueryInValue = Value.substr(0, index);
				Value = Value.substr(index+3);

				index = Value.find("@;@");
				if(index == string::npos)
				{
					FieldInValue = Value;
					Modify(TableInValue.c_str(), QueryInValue.c_str(), FieldInValue.c_str(), NULL, false);
				}
				else
				{
					FieldInValue = Value.substr(0, index);
					ValueInValue = Value.substr(index+3);
					Modify(TableInValue.c_str(), QueryInValue.c_str(), FieldInValue.c_str(), ValueInValue.c_str(), false);
				}
			}
			MUTEXUNLOCK(ProtectAdmin);
		}

		// Save the database data to file
		iResult = Save();
		
		// If error, we try to reload origin data
		if(iResult != 0)
		{
			Load();
			return iResult;
		}


		return 0;		// Everything is OK now
	}

	WaitForMutation();
	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearMutationMode();
		return INTDB_INVALIDTABLE;
	}

	int iResult = SecuTable->Import(ImportFile, Append, FieldSeparator, EntrySeparator, MaxLineLength, Offset, SkipLines);
	
	// Set modification flag
	Modified = true;
	ClearMutationMode();
	return iResult;
}


int Intdb::Export(const char* TableName, const char* ExportFile, const char* FieldSeparator, const char* EntrySeparator, int MaxLineLength)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	WaitForSearch();

	// If the table name is CI_ALLTABLES, we export the entire database file
	// This is an undocumented feature
	if(strcmpnocase(TableName, "CI_ALLTABLES") == 0)
	{
		if(!CopyFileEx(FileName.c_str(), ExportFile))
		{
			ClearSearchMode();
			return INTDB_CANNOTWRITEFILE;
		}
		ClearSearchMode();
		return 0;
	}

	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearSearchMode();
		return INTDB_INVALIDTABLE;
	}

	int iResult = SecuTable->Export(ExportFile, FieldSeparator, EntrySeparator, MaxLineLength);
	ClearSearchMode();
	return iResult;
}


int Intdb::RemoveTable(const char* TableName)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	WaitForMutation();

	if(strcmp(TableName, "*") == 0)
	{
		int i;
		Intdbtable *SecuTable;
		for(i=0; i<(int)Tables.size(); i++)
		{
			SecuTable = (Intdbtable*)(Tables[i]);
			delete SecuTable;
		}
		Tables.clear();
		Modified = true;
	}
	else
	{
		int i;
		Intdbtable *SecuTable;
		bool found = false;
		for(i=0; i<(int)Tables.size(); i++)
		{
			SecuTable = (Intdbtable*)(Tables[i]);
			if(strcmpnocase(SecuTable->Name.c_str(), TableName) == 0)
			{
				found = true;
				delete SecuTable;
				for(;i<(int)(Tables.size()-1); i++)
					Tables[i] = Tables[i+1];
				Tables.pop_back();
				Modified = true;
				break;
			}
		}

		if(!found)
		{
			ClearMutationMode();
			return INTDB_INVALIDTABLE;
		}
	}

	ClearMutationMode();
	return 0;
}




int Intdb::RemoveRecords(const char* TableName, const char* Query)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	if(GetAdminMode())
		return INTDB_ADMINWARNING;

	WaitForMutation();

	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearMutationMode();
		return INTDB_INVALIDTABLE;
	}

	int iResult = SecuTable->RemoveRecords(Query);
	
	// Set modification flag
	if(iResult == 0)
		Modified = true;
	ClearMutationMode();
	return iResult;
}


int Intdb::CreateTable(const char* TableName, const char* FileIn, const char* TableDescription, char Separator, int MaxLineLength)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	WaitForMutation();

	Intdbtable *SecuTable = GetTable(TableName);

	if(SecuTable != NULL)
	{
		ClearMutationMode();
		return INTDB_EXISTINGTABLE;
	}


	if(FileIn == NULL)
	{
		ClearMutationMode();
		return INTDB_INVALIDFILE;
	}


	FILE *stream = Nvfopen(FileIn, "rb");
	if(stream == NULL)
	{
		ClearMutationMode();
		return INTDB_CANNOTOPENFILE;
	}

	string line = "";

	char *Buffer = NVNEW char[MaxLineLength];
	if(Buffer == NULL)
	{
		fclose(stream);	     
		delete Buffer;
		ClearMutationMode();
		return INTDB_MEMORYERROR;
	}
	bool AppendLine = false;	// set to true when the last character of the previous line is '\'
	int LineLength = 0;
	Intdbfielddesc *SecuFieldDesc = NULL;

	SecuTable = NVNEW Intdbtable;
	if(SecuTable == NULL)
	{
		fclose(stream);	     
		delete Buffer;
		ClearMutationMode();
		return INTDB_MEMORYERROR;
	}
	SecuTable->Name = TableName;
	if(TableDescription != NULL)
		SecuTable->Description = TableDescription;
	while(freadline(Buffer, MaxLineLength, stream))
	{
		CleanString(Buffer);
		LineLength = (int)strlen(Buffer);
		if(!AppendLine)
			line = "";
		AppendLine = false;
		if(LineLength && Buffer[LineLength-1] == '\\')
		{
			AppendLine = true;
			Buffer[LineLength-1] = ' ';
			line += Buffer;
			continue;
		}
		line += Buffer;

		string FieldName = "";
		string FieldDesc = "";
		string FieldType = "ALPHA";

		FieldName = line;
		string::size_type index;

		if((index = FieldName.find(Separator)) != string::npos)
		{
			FieldDesc = FieldName.substr(index+1);
			FieldName = FieldName.substr(0, index);
		}

		if((index = FieldDesc.find(Separator)) != string::npos)
		{
			FieldType = FieldDesc.substr(index+1);
			FieldDesc = FieldDesc.substr(0, index);
		}

		CleanString(FieldName);
		CleanString(FieldDesc);
		CleanString(FieldType);

		if(FieldName.size() == 0)
			continue;

		SecuFieldDesc = NVNEW Intdbfielddesc;
		if(SecuFieldDesc == NULL)
		{
			fclose(stream);	     
			delete Buffer;
			ClearMutationMode();
			return INTDB_MEMORYERROR;
		}
		SecuFieldDesc->Name = FieldName;
		SecuFieldDesc->Description = FieldDesc;
		if(strcmpnocase(FieldType.c_str(), "ALPHA") == 0)
			SecuFieldDesc->Type = FIELD_TYPE_ALPHA;
		if(strcmpnocase(FieldType.c_str(), "NUMERIC") == 0)
			SecuFieldDesc->Type = FIELD_TYPE_NUMERIC;

		SecuTable->FieldDescriptions.push_back((void *)SecuFieldDesc);

	}

	// We have finished with the file
	fclose(stream);	     
	delete Buffer;

	if(!SecuTable->FieldDescriptions.size())
	{
		delete SecuTable;
		ClearMutationMode();
		return INTDB_NOFIELDINFILE;
	}

	Tables.push_back((void *)SecuTable);
	Modified = true;

	ClearMutationMode();
	return 0;
}





int Intdb::Modify(const char* TableName, const char* Query, const char* FieldName, const char* FieldValue, bool SaveModif)
{

	if(TableName == NULL)
		return INTDB_INVALIDTABLE;

	if(!strlen(TableName))
		return INTDB_INVALIDTABLE;

	WaitForMutation();
	Modified = true;

	Intdbtable* SecuTable = GetTable(TableName);

	if(SecuTable == NULL)
	{
		ClearMutationMode();
		return INTDB_INVALIDTABLE;
	}


	int iResult = SecuTable->Modify(Query, FieldName, FieldValue);
	if(SaveModif)
	{
		MUTEXLOCK(ProtectAdmin);
		if(!iResult && Admin)
		{
			string ValueForTable = TableName;
			ValueForTable += "@;@";
			ValueForTable += Query;
			ValueForTable += "@;@";
			ValueForTable += FieldName;
			if(FieldValue != NULL)
			{
				ValueForTable += "@;@";
				ValueForTable += FieldValue;
			}
			Modifs.push_back(ValueForTable);
		}
		MUTEXUNLOCK(ProtectAdmin);
	}

	ClearMutationMode();
	return iResult;
}

	
bool Intdb::WriteTag(char TagType, unsigned long TagLength, const char* TagData, FILE *stream)
{
	return ::WriteTag(TagType, TagLength, TagData, stream, FileName.c_str());
}

int Intdb::ReadNextTag(char *Stream, unsigned char *TagType, unsigned long *TagLength, char** TagData)
{
	return ::ReadNextTag(Stream, TagType, TagLength, TagData);
}



void Intdb::FreeData(char* Data)
{
	::FreeData(Data);
}


void Intdb::SetAdminMode()
{
	MUTEXLOCK(ProtectAdmin);
	Admin = true;
	Modifs.clear();
	MUTEXUNLOCK(ProtectAdmin);
}


void Intdb::ClearAdminMode()
{
	MUTEXLOCK(ProtectAdmin);
	Admin = false;
	Modifs.clear();
	MUTEXUNLOCK(ProtectAdmin);
}


bool Intdb::GetAdminMode()
{
	bool RetVal = false;
	MUTEXLOCK(ProtectAdmin);
	RetVal = Admin;
	MUTEXUNLOCK(ProtectAdmin);
	return RetVal;
}


void Intdb::ClearMutationMode()
{
	MUTEXLOCK(ProtectDb);
	Mutation = false;
	MUTEXUNLOCK(ProtectDb);
	MUTEXUNLOCK(ProtectMutation);	// Only one mutation at a time
}


void Intdb::ClearSearchMode()
{
	MUTEXLOCK(ProtectDb);
	InSearch--;
	if(InSearch < 0)
		InSearch = 0;
	MUTEXUNLOCK(ProtectDb);
}


// This wait for no pending mutation
void Intdb::WaitForSearch()
{
	while(1)
	{
		MUTEXLOCK(ProtectDb);
		if(!Mutation)
		{
			InSearch++;
			MUTEXUNLOCK(ProtectDb);
			break;
		}
		MUTEXUNLOCK(ProtectDb);
		SLEEP(50);
	}
}


// This wait for no pending searches
void Intdb::WaitForMutation()
{
	MUTEXLOCK(ProtectMutation);	// Only one mutation at a time
	while(1)
	{
		MUTEXLOCK(ProtectDb);
		Mutation = true;	// This is here in order to block the search in case of mutation request
		if(!InSearch)		// The mutation request is waiting for searches to be terminated
		{
			MUTEXUNLOCK(ProtectDb);
			break;
		}
		MUTEXUNLOCK(ProtectDb);
		SLEEP(50);
	}
}





