// ***************************************************************************************
// File: intdb.h
// Description: internal database classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _INTDB_H
#define _INTDB_H

#include <vector>
#include <string>
#include <map>

// Classes declared in this file
class Intdbfield;
class Intdbfielddesc;
class Intdbrecord;
class Intdbtable;
class Intdb;

// Error codes
#define INTDB_INVALIDFILE			15001
#define INTDB_CANNOTOPENFILE		15002
#define INTDB_CANNOTREADFILE		15003
#define INTDB_CANNOTWRITEFILE		15004
#define INTDB_INVALIDCOMMAND		15005
#define INTDB_INVALIDPARAMETER		15006
#define INTDB_MEMORYERROR			15007
#define INTDB_INVALIDDB				15008
#define INTDB_INVALIDQUERY			15009
#define INTDB_INVALIDTABLE			15010
#define INTDB_INVALIDPASSWORD		15012
#define INTDB_NOFIELDINFILE			15013
#define INTDB_EXISTINGTABLE			15014
#define INTDB_INVALIDFIELD			15015
#define INTDB_ADMINWARNING			15016
#define INTDB_INVALIDSEPARATOR		15017
#define INTDB_CANNOTSETHASHTABLE	15018
#define INTDB_NOTPOSSIBLEONPRIMARY	15019


// File defines
#define DBFILE_SIZE_IDENTIFIER		20
#define DBFILE_IDENTIFIER			"NIRVA DBFILE"
#ifdef NV_SERVER
#define DBFILE_VERSION				"1.01" // added unicode entry used only on server side (for registry and storage(in fact container import))
#else
#define DBFILE_VERSION				"1.00"
#endif

// Field types
#define FIELD_TYPE_ALPHA	0x01
#define FIELD_TYPE_NUMERIC	0x02

// Database
#define DB_VERSION			"1.00"

// File tags
#define FILE_NOTAG			0x00
#define FILE_VERSION		0x01
#define FILE_DB			0x10
#define FILE_DB_NAME		0x11
#define FILE_DB_DESC		0x12
#define FILE_DB_VERSION	0x13
#define FILE_TABLE			0x20
#define FILE_TABLE_NAME		0x21
#define FILE_TABLE_DESC		0x22
#define FILE_TABLE_PRIMARY	0x23
#define FILE_TABLE_PRIMARY_CASE	0x24
#define FILE_TABLE_UNICODE	0x25
#define FILE_DFIELD			0x30
#define FILE_DFIELD_NAME	0x31
#define FILE_DFIELD_DESC	0x32
#define FILE_DFIELD_TYPE	0x33
#define FILE_RECORD			0x40
#define FILE_FIELD			0x50
#define FILE_ENTRY			0x60
#define FILE_LENTRY			0x61
#define FILE_DB_END		0xFF

// Search operators
#define SEARCH_OPERATOR_NONE	0x00
#define SEARCH_OPERATOR_E		0x01
#define SEARCH_OPERATOR_NE		0x02
#define SEARCH_OPERATOR_ST		0x03
#define SEARCH_OPERATOR_GT		0x04
#define SEARCH_OPERATOR_STE		0x05
#define SEARCH_OPERATOR_GTE		0x06
#define SEARCH_OPERATOR_ALL		0x07

// Search combinators
#define SEARCH_COMBINATOR_NONE	0x00
#define SEARCH_COMBINATOR_AND	0x01
#define SEARCH_COMBINATOR_OR	0x02

// Search function mode
#define SEARCH_MODE_SEARCH		0x00
#define SEARCH_MODE_REMOVE		0x01
#define SEARCH_MODE_CLEAR		0x02

// Encoding mode
#define ENCODING_MODE_NONE				0x00
#define ENCODING_MODE_ENCODE			0x01	// Used when sending object to client
#define ENCODING_MODE_DECODE			0x02	// Used when receiving object from client
#define ENCODING_MODE_DECODE_FROM_FILE	0x03	// Used when loading object from file (restricted in Nirva to server only and read registry from disk and container/object import)

// Add mode
#define ADD_MODE_NORMAL			0x00
#define ADD_MODE_REPLACE		0x01
#define ADD_MODE_NOREPLACE		0x02



bool WriteTag(char TagType, unsigned long TagLength, const char* TagData, FILE *stream, const char* FileName = NULL, bool RemoveFileIfError = true);
bool WriteBufferTag(char TagType, unsigned long TagLength, const char* TagData, char *WriteBuffer);
int ReadNextTag(char *Stream, unsigned char *TagType, unsigned long *TagLength, char** TagData);
void FreeData(char* Data);


/////////////////////////////////////////////////////////////////////////////
// Intdbfield class

class Intdbfield
{
public:
	std::vector<std::string> Entries;	// field lines

	
	Intdbfield();
	virtual ~Intdbfield();
	
	void Clear();
	void CopyFrom(Intdbfield *Source);
	void Sort(bool Numeric = false, bool Ascending = true);
	int SortFunc(std::string& a, std::string& b, bool Numeric, bool Ascending);
	std::string GetEntry(int index, const char *Default = NULL);
	bool SetEntry(int index, const char *Value);
	bool InsertEntry(const char *Value, int index = -1);
	bool RemoveEntry(int index);
	bool SwapEntries(int index1, int index2);
	int GetSize();
	bool Verify(const char *Value, int Operator, bool Numeric = false);
	bool IsFromField(Intdbfield *Source, bool Remove = false);
};


/////////////////////////////////////////////////////////////////////////////
// Intdbfielddesc class

class Intdbfielddesc
{
public:
	std::string Name;				// Field name
	std::string Description;		// Field description
	char Type;						// Field type

	
	Intdbfielddesc();
	virtual ~Intdbfielddesc();
	
	void Clear();
	void CopyFrom(Intdbfielddesc *Source);
};


/////////////////////////////////////////////////////////////////////////////
// Intdbrecord class

class Intdbrecord
{
public:
	std::vector<void *> Fields;		// field contents

	
	Intdbrecord();
	virtual ~Intdbrecord();
	
	void Clear();
	void ClearFieldData();
	void RemoveField(int index);
	void InsertField(int index = -1);
	void SwapFields(int index1, int index2);
	void CopyFrom(Intdbrecord *Source);
	Intdbfield* GetField(int index);
	int GetSize();
};


/////////////////////////////////////////////////////////////////////////////
// Intdbtable class

class Intdbtable
{
public:
	std::string Name;								// Table name
	std::string Description;						// Table description
	std::string PrimaryName;						// Primary key field
	std::vector<void *> FieldDescriptions;			// Field description list
	std::vector<void *> Records;					// Record list
	std::map<std::string, int> Primary;				// Primary key hash table		
	bool PrimaryCaseSensitive;						// Primary key is case sensitive or not
	bool Unicode;									// utf-8 or isolatin-1
	
	Intdbtable();
	virtual ~Intdbtable();
	
	void Clear();
	void CopyFrom(Intdbtable *Source, bool WithData = true, int MaxDocs = -1);
	bool SelectFrom(Intdbtable* Source, int source_field_index, int dest_field_index, int selected_field_index, const char* SelectedYes = NULL, const char* SelectedNo = NULL, bool CaseSensitive = true);
	bool Join(Intdbtable* Source, const char* Columns, const char* Foreign, bool Replace = false);
	bool Add(Intdbtable* Source, int Mode = ADD_MODE_NORMAL);
	bool SetPrimaryHashTable(bool bRemoveDoubleEntries = false);
	std::string GetHashValue(std::string Value);
	bool IsPrimaryColumn(int col_index);
	int GetPrimary(std::string Value);
	void RemovePrimary(std::string Value);
	void SetPrimary(std::string Value, int index);
	bool SetPrimaryField(std::string FieldName, bool CaseSensitive = false, bool bRemoveDoubleEntries = false);
	std::string GetPrimaryField();
	std::string GetRecordPrimaryValue(int index);
	int Load(const char *FileName, int iEncoding = ENCODING_MODE_NONE, bool bUnicode = false);
	int Save(const char *FileName);
	char* SaveToBuffer(unsigned long *BufferSize);
	bool LoadFromBuffer(char *Buffer, unsigned long BufferSize, int iEncoding = ENCODING_MODE_NONE, bool bUnicode = false);
	bool CheckValid();
	Intdbrecord* GetRecord(int index);
	Intdbfield* GetField(int record_index, int field_index);
	void ClearData(bool OnlyFieldContent = false);
	int GetFieldIndex(const char* FieldName, int StartAt = 0);
	std::string GetFieldName(int FieldIndex);
	std::string GetFieldDesc(int FieldIndex);
	bool SetFieldName(int FieldIndex, std::string FieldName);
	void SetFieldDesc(int FieldIndex, std::string FieldDesc);
	void SetFieldNumeric(int FieldIndex, bool Numeric);
	bool IsNumericField(int FieldIndex);
	int GetSize();
	int GetNumFields();
	void RemoveRecord(int index);
	int RemoveRecords(const char* Query);
	void RemoveField(int FieldIndex);
	int ClearDataRecords(const char* Query);
	void ClearRecord(int index);
	Intdbrecord* CreateEmptyRecord();
	void InsertRecords(int num_records = 1, int record_index = -1);
	bool SetRecordData(int index, const char* RecordData = NULL, const char* FieldSeparator = ";", const char* EntrySeparator = "|");
	bool AddRecords(const char* RecordData, const char* RecordSeparator = "\n", const char* FieldSeparator = ";", const char* EntrySeparator = "|");
	void SwapRecords(int index1, int index2);
	int GetRecords(int first, int last, Intdbtable **Result, const char* Columns);
	void SwapFields(int Fieldindex1, int Fieldindex2);
	void InsertField(const char* FieldName, const char* FieldDesc = NULL, bool numeric = false, int FieldIndex = -1);
	bool AddToField(int index, const char* FieldName, const char* Value, bool AppendMode = true);
	bool FilterFields(const char* Columns = NULL);
	std::string GetFirstLineField(int index, const char* FieldName, const char *Default = NULL);
	int Search(const char* Query, Intdbtable **Result, const char* SortField = NULL, int Mode = SEARCH_MODE_SEARCH, int *FirstIndex = NULL, int MaxDocs = -1, const char* Columns = NULL, bool RowIndexCol = false);
	int Info(Intdbtable **Result);
	int Import(const char* ImportFile, bool Append = false, const char* FieldSeparator = ";", const char* EntrySeparator = "|", int MaxLineLength = 1024, int Offset = 0, int SkipLines = 0, bool WithLf = false);
	int Export(const char* ExportFile, const char* FieldSeparator = ";", const char* EntrySeparator = "|", int MaxLineLength = 1024, bool WithLf = false);
	int Modify(const char* Query, int FieldIndex, const char* FieldValue = NULL);
	int Modify(const char* Query, const char* FieldName, const char* FieldValue = NULL);
	void Sort(const char* SortField);
	int SortFunc(void* a, void* b, int FieldIndex, bool Numeric, bool Ascending);

	void ClearCell(int row_index, int col_index);
	void SortCell(int row_index, int col_index, bool Ascending = true);
	int GetNumCellLines(int row_index, int col_index);
	std::string GetCellLine(int row_index, int col_index, int line_index = 0, std::string Default = "");
	void GetCellLine(int row_index, int col_index, char *Buffer, int BufferSize, int line_index = 0, const char *Default = NULL);
	bool SetCellLine(int row_index, int col_index, std::string Line, int line_index = 0);
	bool SetCellLine(int row_index, int col_index, const char* Line, int line_index = 0);
	bool InsertCellLine(int row_index, int col_index, std::string Line, int line_index = -1);
	bool InsertCellLine(int row_index, int col_index, const char* Line, int line_index = -1);
	void RemoveCellLine(int row_index, int col_index, int line_index = -1);
	void SwapCellLines(int row_index, int col_index, int line_index1, int line_index2);
};


/////////////////////////////////////////////////////////////////////////////
// Intdb class
// NOT USED IN NIRVA

class Intdb
{
public:
	std::string Name;				// Database name
	std::string Description;		// Database description
	std::string Version;			// Database version
	std::string FileName;			// Database associated file name (complete path)
	std::string Password;			// Password for modification (the password is managed externally)
	std::vector<void *> Tables;		// table list
	bool Modified;					// Modify flag
	bool Mutation;					// Set to one when in mutation mode
	int InSearch;					// Increased by one for each search
	int		InUse;					// Counter for the number of clients using the database
	int		ToRemove;				// This database cannot be used any more
	bool Admin;						// Set to one when in admin mode
	std::vector<std::string> Modifs;// Modification orders

	MUTEXDECLARE(ProtectDb);		// Database protection for mutations (search not allowed when a mutation is requested)
	MUTEXDECLARE(ProtectMutation);	// Database protection for mutations (only one mutation at a time)
	MUTEXDECLARE(ProtectInUse);		// protection for InUse variable
	MUTEXDECLARE(ProtectAdmin);		// protection for Admin variable

	
	Intdb();
	virtual ~Intdb();
	
	Intdbtable* GetTable(const char* TableName);
	Intdbtable* GetTable(int index);

	void Clear();
	bool WriteTag(char TagType, unsigned long TagLength, const char* TagData, FILE *stream);
	int ReadNextTag(char *Stream, unsigned char *TagType, unsigned long *TagLength, char** TagData);
	void FreeData(char* Data);
	bool CheckValid();
	void ClearMutationMode();
	void ClearSearchMode();
	void WaitForSearch();
	void WaitForMutation();
	void SetAdminMode();
	void ClearAdminMode();
	bool GetAdminMode();


	// External interface
	int Load(bool JustHeader = false);
	int Save();
	int Search(const char* TableName, const char* Query, Intdbtable **Result, const char* SortField = NULL);
	int Info(const char* TableName, Intdbtable **Result);
	int GetTableSize(const char* TableName, int *TableSize);
	int Import(const char* TableName, const char* ImportFile, bool Append = false, const char* FieldSeparator = ";", const char* EntrySeparator = "|", int MaxLineLength = 1024, int Offset = 0, int SkipLines = 0);
	int Export(const char* TableName, const char* ExportFile, const char* FieldSeparator = ";", const char* EntrySeparator = "|", int MaxLineLength = 1024);
	int RemoveTable(const char* TableName);
	int RemoveRecords(const char* TableName, const char* Query);
	int CreateTable(const char* TableName, const char* FileIn, const char* TableDescription = NULL, char Separator = ';', int MaxLineLength = 1024);
	int Modify(const char* TableName, const char* Query, const char* FieldName, const char* FieldValue = NULL, bool SaveModif = true);
	void CopyFrom(Intdb *Source);
	int GetSize();
	void IncInUse();
	void DecInUse();
	int IsInUse();

};



#endif //_INTDB_H
