// ***************************************************************************************
// File: log.h
// Description: declarations of the Nirva log classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _LOG_H
#define _LOG_H


// Classes declared in this file
class NvLog;
class NvLogList;

#include "task.h"
#include "util.h"
#include <string>


// log entry type
#define NV_LOG_ALL				-1
#define NV_LOG_INFO				0
#define NV_LOG_WARNING			1
#define NV_LOG_ERROR			2
#define NV_LOG_UNKNOWNTYPE		3

// Log file separator
#define NV_LOG_FIELD_SEPARATOR			"*@NV@*"
#define NV_LOG_MAX_LINE_LENGTH			65536

// Log file identifier
#define NV_LOG_FILE_IDENTIFIER			"NIRVA_LOG_FILE"
#define NV_LOG_FILE_VERSION				"1.00"


/////////////////////////////////////////////////////////////////////////////
// NvLog class
class NvLog : public NvObject
{
public:
		
	NvMutex ProtectInUse;						// protect locking of log for updating the in use counter
	int InUse;									// In use counter
	std::string m_sName;						// Log name
	std::string m_sDesc;						// Log description
	int m_iMaxFileTime;							// Max log file time in minutes
	bool m_bSaveIt;								// Save log file flag
	int m_iMaxSaveDays;							// Max number of save days
	std::string m_sLogDir;						// Log file directory
	int m_iLevel;								// Actual log level
	int m_iLogFileStartTime;					// Log file start time
	FILE *logstream;								// Log open stream


	NvLog();
	virtual ~NvLog();

	bool IncInuse();
	void DecInUse();
	int GetInUse();
	static void CleanLogName(std::string& Name);

	bool Write(const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1, const char *Identifier = NULL, const char *Source = NULL);
	bool Search(NvObjTable *LogView, const char* FromDate = NULL, const char* ToDate = NULL, const char* FromHour = NULL, const char* ToHour = NULL, const char* Identifier = NULL, int Type = NV_LOG_ALL, const char* Message = NULL, const char* Extra = NULL, const char* Source = NULL, int Level = 0, const char* ExportFile = NULL, int MaxFileSize = 1024);
	std::string GetLogHeader(const char *LogFile, const char *Header);
	int GetLogStartOffset(const char *LogFile);
	bool BuildLogFile(const char* FileOut, int MaxFileSize = 1024, const char* FromDate = NULL, const char* ToDate = NULL, const char* FromHour = NULL, const char* ToHour = NULL);
	bool RemoveLogData(const char* FromDate, const char* ToDate, bool Lock = true, bool ClearIt = false);
	void CheckEmptyDirs();
};


/////////////////////////////////////////////////////////////////////////////
// NvLogList class
class NvLogList : public NvObject
{
private:
	NvLog* InternalGetLog(const char *Name);

public:
	std::map<std::string,void *> m_vLogs;	// log list
	std::string m_sAppName;					// Application name for an application log list
	std::string m_sServiceName;				// Service name for a service log list
	void *Owner;							// This is a pointer to main NvsServer class or NvService or NvApplication 
	NvLog* LastLog;							// Last used log
	
	NvLogList();
	virtual ~NvLogList();

	// Removes all logs from the list
	void Clear();

	// Load the log list from registry
	bool Load();

	// Saves the log list to the registry
	bool Save();

	// Creates a new log with the given parameters
	bool Create(const char *Name, bool UseLocking = true, bool Erase = false, const char *ImportFile = NULL);

	// Set log description
	bool SetDescription(const char *Name, const char *Description);

	// Set log level
	bool SetLevel(const char *Name, int NewLevel = 0);

	// Get log level
	int GetLevel(const char *Name);

	// Test if log exist
	bool Exist(const char *Name);
	
	// Set log save flag
	bool SetSave(const char *Name, bool SaveIt = true);
	
	// Set log max time
	bool SetMaxTime(const char *Name, int MaxTime = 10);
	
	// Set log max save days
	bool SetMaxSaveDays(const char *Name, int MaxDays = 2);
	
	// Remove the given log from the list
	void RemoveLog(const char *Name, bool Erase = false, bool ProtectSystemLog = true);

	// Get log list
	bool GetInfo(const char *Name, NvObjTable *LogInfo);

	// Write log entry
	bool Write(const char *Name, const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1, const char *Identifier = NULL, const char *Source = NULL);

	// Search log
	bool Search(const char *Name, NvObjTable *LogView, const char* FromDate = NULL, const char* ToDate = NULL, const char* FromHour = NULL, const char* ToHour = NULL, const char* Identifier = NULL, int Type = NV_LOG_ALL, const char* Message = NULL, const char* Extra = NULL, const char* Source = NULL, int Level = 0, const char* ExportFile = NULL, int MaxFileSize = 1024);
	
	// Removes some log data
	bool RemoveLogData(const char *Name, const char* FromDate, const char* ToDate, bool ClearIt = false);
};




#endif //_LOG_H
