// ***************************************************************************************
// File: test.cpp
// Description: nirva test features
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;


/////////////////////////////////////////////////////////////////////////////
// NvTest class
NvTest::NvTest()
{
	m_sName = "";
	m_sDesc = "";
	m_sProc = "";
	m_sUrl = "";
	m_iType = NV_TEST_PROC;
}

NvTest::~NvTest()
{
}


/////////////////////////////////////////////////////////////////////////////
// NvTestSet

NvTestSet::NvTestSet()
{
	m_sName = "";
	m_sDesc = "";
}

NvTestSet::~NvTestSet()
{
	Clear();
}



void NvTestSet::Clear()
{
	map<string,NvTest*>::iterator pos;
	for(pos = m_mTests.begin(); pos != m_mTests.end(); ++pos)
	{
		NvTest *Test = pos->second;
		if(Test != NULL)
			delete Test;
	}
	m_mTests.clear();
	m_sName = "";
	m_sDesc = "";
}

NvTest* NvTestSet::GetTest(const char *Name)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	CleanString(NameHere);

	if(!NameHere.length())
		return NULL;

	if(m_mTests.find(NameHere) == m_mTests.end())
		return NULL;
	
	NvTest *FoundTest = m_mTests[NameHere];

	return FoundTest;
}

// Load
bool NvTestSet::Load(string& FileName)
{
	Clear();
	
	string XslFile = GetMain()->GetNirvaDir();
	XslFile += "Files";
	AddDirSuffix(XslFile);
	XslFile += "Test";
	AddDirSuffix(XslFile);
	XslFile += "testset.xsl";
	if(!FileExist(XslFile.c_str()))
		return false;

	NvXml XmlParser;
	unsigned long FileDataSize = (unsigned long)(GetFileLength(FileName.c_str()));
	if(FileDataSize <= 0)
		return false;
	char *FileData = (char *)malloc(FileDataSize);
	if(FileData == NULL)
		return false;
	FILE *File;
	File = Nvfopen(FileName.c_str(), "rb");
	if(File == NULL)
	{
		free(FileData);
		return false;
	}
	if(fread(FileData, 1, FileDataSize, File) != FileDataSize)
	{
		fclose(File);
		free(FileData);
		return false;
	}
	fclose(File);

	// We have the file data into FileData
	if(!XmlParser.XsltTransform(&FileData, &FileDataSize, XslFile.c_str()))
	{
		free(FileData);
		return false;
	}

	// Now FileData contains the input xml ready for a container
	// We create a special command for populating a container and we give it the ownership of data
	NvCommand MyCommand;
	MyCommand.Data = FileData;
	MyCommand.DataSize = FileDataSize;

	// Create a destination container
	NvContainer MyContainer;
	string NameSpace = "";
	int XmlCoding = XML_MODEL_NORMAL;
	if(!XmlParser.SetInputData(&MyContainer, &MyCommand, NameSpace, XmlCoding))
	{
		MyCommand.Clear();
		return false;
	}

	// Al is now in the container
	// We don't need any more the command data
	MyCommand.Clear();

	// Get the name
	NvObjString *MyString = (NvObjString *)(MyContainer.GetObject("NAME"));
	if(MyString == NULL)
		return false;
	m_sName = MyString->GetValue();
	CleanString(m_sName);
	if(!m_sName.length())
		return false;

	// Get description
	MyString = (NvObjString *)(MyContainer.GetObject("DESCRIPTION"));
	if(MyString == NULL)
		return false;
	m_sDesc = MyString->GetValue();
	CleanString(m_sDesc);

	// Get tests
	NvObjTable *MyTable = (NvObjTable *)(MyContainer.GetObject("TESTS"));
	if(MyTable == NULL)
		return true;	// No tests
	int NumRows = MyTable->GetNumRow();
	int i;
	string sTestName;
	string sTestDesc;
	string sTestType;
	string sTestUrl;
	string sTestProc;
	for(i=0;i<NumRows;i++)
	{
		sTestName = MyTable->GetCellLine(i, 0);
		sTestDesc = MyTable->GetCellLine(i, 1);
		sTestType = MyTable->GetCellLine(i, 2);
		sTestUrl = MyTable->GetCellLine(i, 3);
		sTestProc = MyTable->GetCellLine(i, 4);
		if(!sTestName.length())
			continue;
		CleanString(sTestName);
		if(m_mTests.find(sTestName) != m_mTests.end())
			continue;	// Entry already exist
		MakeUpper(sTestType);
		if(!sTestType.length())
		{
			if(sTestUrl.length())
				sTestType = "URL";
			if(sTestProc.length())
				sTestType = "PROC";
			if(!sTestType.length())
				continue;	// No type defined
		}
		if((sTestType != "URL") && (sTestType != "PROC"))
			continue;	// Bad type
		if(sTestType == "URL")
		{
			sTestProc = "";
			if(!sTestUrl.length())
				continue;	// No url given
		}
		if(sTestType == "PROC")
		{
			sTestUrl = "";
			if(!sTestProc.length())
				continue;	// No url given
		}
		// Everything is ok to add the test now
		NvTest *Test = new NvTest;
		Test->m_sName = sTestName;
		Test->m_sDesc = sTestDesc;
		Test->m_sProc = sTestProc;
		Test->m_sUrl = sTestUrl;
		if(sTestType == "URL")
			Test->m_iType = NV_TEST_URL;
		if(sTestType == "PROC")
			Test->m_iType = NV_TEST_PROC;
		m_mTests[sTestName] = Test;
	}
	return true;
}




/////////////////////////////////////////////////////////////////////////////
// NvTestSetList

NvTestSetList::NvTestSetList()
{
	m_sDirPath = "";
}

NvTestSetList::~NvTestSetList()
{
	NV_LOCKOBJECT(this);
	Clear();
	NV_UNLOCKOBJECT(this);
}



void NvTestSetList::Clear()
{
	map<string,NvTestSet*>::iterator pos;
	for(pos = m_mTestSets.begin(); pos != m_mTestSets.end(); ++pos)
	{
		NvTestSet *Set = pos->second;
		if(Set != NULL)
			delete Set;
	}
	m_mTestSets.clear();
}


NvTestSet* NvTestSetList::InternalGetSet(const char *Name)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	CleanString(NameHere);

	if(!NameHere.length())
		return NULL;

	if(m_mTestSets.find(NameHere) == m_mTestSets.end())
		return NULL;
	
	NvTestSet *FoundSet = m_mTestSets[NameHere];

	return FoundSet;
}

// Load
bool NvTestSetList::Load()
{
	NV_LOCKOBJECT(this);
	Clear();
	if(!m_sDirPath.length())
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(!DirExist(m_sDirPath.c_str()))
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	vector<string> list;
	FileList(m_sDirPath.c_str(), NULL, &list);

	// Process each file
	string FileName = "";
	int i;
	for(i=0; i<(int)(list.size()); i++)
	{
		FileName = m_sDirPath;
		AddDirSuffix(FileName);
		FileName += list[i];
		if(FileName.rfind(".xml") != (FileName.length() - 4))
			continue;	// Take only xml files

		NvTestSet *TestSet = new NvTestSet;
		if(!TestSet->Load(FileName))
		{
			delete TestSet;
			continue;
		}
		if(InternalGetSet(TestSet->m_sName.c_str()) != NULL)
		{
			// A set with the same name already exist
			delete TestSet;
			continue;
		}
		m_mTestSets[TestSet->m_sName] = TestSet;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}

bool NvTestSetList::GetInfo(const char *SetName, NvObjTable *SetInfo)
{
	SetInfo->Clear();

	bool AllSets = true;

	NV_LOCKOBJECT(this);

	NvTestSet *FoundSet = NULL;
	if((SetName != NULL) && strlen(SetName))
	{
		// There is a request on a specific set
		FoundSet = InternalGetSet(SetName);
		if(FoundSet == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllSets = false;
	}


	SetInfo->InsertColumn("NAME");
	SetInfo->InsertColumn("DESCRIPTION");
	SetInfo->InsertColumn("NUM_TESTS");

	if(AllSets)
	{
		int j;
		map<string,NvTestSet*>::iterator pos;
		for(pos = m_mTestSets.begin(), j=0; pos != m_mTestSets.end(); ++pos)
		{
			NvTestSet *TestSet = pos->second;
			if(TestSet == NULL)
				continue;
			SetInfo->InsertRow();
			SetInfo->InsertCellLine(j, 0, TestSet->m_sName);
			SetInfo->InsertCellLine(j, 1, TestSet->m_sDesc);
			char Buffer[20];
			sprintf(Buffer, "%d", TestSet->m_mTests.size());
			SetInfo->InsertCellLine(j, 2, Buffer);
			j++;
		}
	}
	else
	{
		SetInfo->InsertRow();
		SetInfo->InsertCellLine(0, 0, FoundSet->m_sName);
		SetInfo->InsertCellLine(0, 1, FoundSet->m_sDesc);
		char Buffer[20];
		sprintf(Buffer, "%d", FoundSet->m_mTests.size());
		SetInfo->InsertCellLine(0, 2, Buffer);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}


bool NvTestSetList::GetTestInfo(const char *SetName, const char *TestName, NvObjTable *TestInfo)
{
	TestInfo->Clear();

	bool AllTests = true;

	NV_LOCKOBJECT(this);

	NvTestSet *FoundSet = InternalGetSet(SetName);
	if(FoundSet == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	
	NvTest *FoundTest = NULL;
	if((TestName != NULL) && strlen(TestName))
	{
		// There is a request on a specific set
		FoundTest = FoundSet->GetTest(TestName);
		if(FoundTest == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllTests = false;
	}


	TestInfo->InsertColumn("NAME");
	TestInfo->InsertColumn("DESCRIPTION");
	TestInfo->InsertColumn("TYPE");
	TestInfo->InsertColumn("PROCEDURE");
	TestInfo->InsertColumn("URL");

	if(AllTests)
	{
		int j;
		map<string,NvTest*>::iterator pos;
		for(pos = FoundSet->m_mTests.begin(), j=0; pos != FoundSet->m_mTests.end(); ++pos)
		{
			NvTest *Test = pos->second;
			if(Test == NULL)
				continue;
			TestInfo->InsertRow();
			TestInfo->InsertCellLine(j, 0, Test->m_sName);
			TestInfo->InsertCellLine(j, 1, Test->m_sDesc);
			if(Test->m_iType == NV_TEST_URL)
				TestInfo->InsertCellLine(j, 2, "URL");
			else
				TestInfo->InsertCellLine(j, 2, "PROC");
			TestInfo->InsertCellLine(j, 3, Test->m_sProc);
			TestInfo->InsertCellLine(j, 4, Test->m_sUrl);
			j++;
		}
	}
	else
	{
		TestInfo->InsertRow();
		TestInfo->InsertCellLine(0, 0, FoundTest->m_sName);
		TestInfo->InsertCellLine(0, 1, FoundTest->m_sDesc);
		if(FoundTest->m_iType == NV_TEST_URL)
			TestInfo->InsertCellLine(0, 2, "URL");
		else
			TestInfo->InsertCellLine(0, 2, "PROC");
		TestInfo->InsertCellLine(0, 3, FoundTest->m_sProc);
		TestInfo->InsertCellLine(0, 4, FoundTest->m_sUrl);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}


