
package com.allaire.cfx ;


/**
*  Interface to a request made to a <code>{@link CustomTag}</code>.
*
*  This interface includes methods for retrieving attributes passed to
*  the tag (including queries) and reading global tag settings. 
*
*  @author J.J. Allaire
*/
public interface Request 
{	
	/**
	* Checks to see whether the attribute was passed to this tag.
	*
	* <P>The following example checks to see if the user passed an
	* attribute named DESTINATION to the tag and throws an exception
   * if the attribute was not passed: 
	*
	* <pre><code>   if ( ! request.attributeExists(&quot;DESTINATION&quot;) )
	*   {
	*      throw new Exception( 
	*         &quot;Missing DESTINATION parameter&quot;,
	*         &quot;You must pass a DESTINATION parameter in &quot;
	*         &quot;order for this tag to work correctly.&quot; ) ;
	*   } ; </code></pre>
	*
	* @param		name of the attribute to check (case insenstive)
	*
	* @return	Returns <code>true</code> if the attribute is available 
	*           otherwise returns <code>false</code>.
	*
	* @see		#getAttribute(String)	getAttribute
	* @see		#getAttributeList			getAttributeList
	*/
	public boolean attributeExists( String name ) ;
	

	/**
	* Retrieves the value of the passed attribute. 
	* 
	* Returns an empty string if the attribute does not exist (use
	* <code>{@link #attributeExists(String) attributeExists}</code> to test whether an 
	* attribute was passed to the tag). Use <code>{@link #getAttribute(String,String) getAttribute(String,String)}</code>
	* to return a default value rather than an empty string.
	*
	* <P>The following example retrieves an attribute named DESTINATION 
   * and writes its value back to the user: </P>
	*
	* <pre><code>   String strDestination = request.getAttribute(&quot;DESTINATION&quot;) ;
	*   response.write( &quot;The destination is: &quot; + strDestination ) ; </code></pre>
	*
	* @param		name the attribute to retrieve (case insensitive)
	*
	* @return	The value of the attribute passed to the tag. If no attribute
	*           of that name was passed to the tag then an empty string is returned.
	*
   * @see		#attributeExists(String)	attributeExists
	* @see		#getAttributeList				getAttributeList
   * @see		#getAttribute(String,String)	getAttribute(String,String)
	* @see		#getIntAttribute(String)	getIntAttribute
	*/
	public String getAttribute( String name ) ;


	/**
	* Retrieves the value of the passed attribute (returns default
	* if the attribute does not exist). 
	* 
	* <P>The following example retrieves an attribute named DESTINATION 
   * and writes its value back to the user: </P>
	*
	* <pre><code>   String strDestination = 
	*      request.getAttribute( &quot;DESTINATION&quot;, &quot;Rome&quot; ) ;
	*   response.write( &quot;The destination is: &quot; + strDestination ) ; </code></pre>
	*
	* @param		name the attribute to retrieve (case insensitive)
	* @param    def the value to return if the attribute does not exist
	*
	* @return	The value of the attribute passed to the tag. If no attribute
	*           of that name was passed to the tag then the default is returned.
	*
   * @see		#attributeExists(String)	attributeExists
	* @see		#getAttributeList				getAttributeList
   * @see		#getAttribute(String)		getAttribute(String)
	* @see		#getIntAttribute(String)	getIntAttribute
	* 
	*/
	public String getAttribute( String name, String def ) ;



	/**
	* Retrieves the value of the passed attribute as an integer. 
	* 
	* Returns -1 if the attribute does not exist. Throws a <code>NumberFormatException</code> 
	* if the attribute is not a valid number. Use <code>{@link #attributeExists(String) attributeExists}</code> to test whether an 
	* attribute was passed to the tag. Use <code>{@link #getIntAttribute(String,int) getIntAttribute(String,int)}</code>
	* to return a default value rather than throwing an exception or returning -1. 
	*
	* <P>The following example retrieves an attribute named PORT
   * and writes its value back to the user: </P>
	*
	* <pre><code>   int nPort = request.getIntAttribute(&quot;PORT&quot;) ;
	*   if ( nPort != -1 )
	*      response.write( &quot;The port is: &quot; + String.valueOf(nPort) ) ; </code></pre>
	*
	* @param		name the attribute to retrieve (case insensitive)
	*
  	* @throws <code>NumberFormatException</code> If the attribute is not a valid number.
	*
	* @return	The value of the attribute passed to the tag. If no attribute
	*           of that name was passed to the tag then -1 is returned.
	*
   * @see		#attributeExists(String)	attributeExists
	* @see		#getAttributeList				getAttributeList
   * @see		#getIntAttribute(String,int)	getIntAttribute(String,int)
	*/
	public int getIntAttribute( String name ) throws NumberFormatException ;



	/**
	* Retrieves the value of the passed attribute as an integer (returns default
	* if the attribute does not exist or is not a valid number).
	* 
	* <P>The following example retrieves an attribute named PORT
   * and writes its value back to the user: </P>
	*
	* <pre><code>   int nPort = request.getIntAttribute(&quot;PORT&quot;, 80) ;
	*   response.write( &quot;The port is: &quot; + String.valueOf(nPort) ) ; </code></pre>
	*
	* @param		name the attribute to retrieve (case insensitive)
	* @param    def the value to return if the attribute does not exist or is invalid
	*
   * @return	The value of the attribute passed to the tag. If no attribute
	*           of that name was passed to the tag or if the attribute is not 
	*           a valid number then the default is returned.
	*
   * @see		#attributeExists(String)	attributeExists
	* @see		#getAttributeList				getAttributeList
   * @see		#getIntAttribute(String)	getIntAttribute(String)
	*/
	public int getIntAttribute( String name, int def ) ;


	/**
	* Retrieves a list of all attributes passed to the tag. 
	* 
	* To retrieve the value of an individual attribute you should use the 
	* <code>{@link #getAttribute(String) getAttribute}</code> member function.
	*
	* <P>The following example retrieves the list of attributes and then
	* iterates over the list, writing each attribute and its value
	* back to the user: </P>
	*
	* <pre><code>   String[] attribs = request.getAttributeList() ;
	*   int nNumAttribs = attribs.length ;	
	*	
	*   for( int i=0; i&lt;nNumAttribs; i++ )
	*   {
	*      String strName = attribs[i] ;
	*      String strValue = request.getAttribute( strName ) ;
	*      response.write( strName + &quot;=&quot; + strValue + &quot;&lt;BR&gt;&quot; ) ;
	*   } </code></pre>
	*
	* @return  An array of strings containing the names of the  
	*          attributes passed to the tag.
	*
   * @see		#attributeExists(String)	attributeExists
	* @see		#getAttribute(String)		getAttribute
	*/
	public String[] getAttributeList() ;
		

	/**
	* Retrieves the query that was passed to this tag.
	*
	* <P>To pass a query to a custom tag you use the <code>QUERY</code>
	* attribute. This attribute should be set to the name of an existing query 
	* (e.g. created using the <code>CFQUERY</code> tag). The <code>QUERY</code>
	* attribute is optional and should only be used by tags which need to process 
	* an existing dataset. </P>
	*
	* <P>The following example retrieves the query which was passed
	* to the tag. If no query was passed then an exception is thrown:</P>
	*  
	* <pre><code>   Query query = request.getQuery() ;
	*   if ( query == null )
	*   {
	*      throw new Exception(  
	*         &quot;Missing QUERY parameter. &quot; +
	*         &quot;You must pass a QUERY parameter in &quot;
	*         &quot;order for this tag to work correctly.&quot; ) ;   
	*   } </code></pre>
	*
   * @return The <code>{@link Query Query}</code> that was passed to the
	* tag. If no query was passed to the tag then <code>null</code> is returned.
	*/
	public Query getQuery() ;


	/**
	* Retrieves the value of a global custom tag setting.
	*
	* Custom tag settings are stored within the CustomTags
	* section of the ColdFusion Registry key. 
	*
	* <P>IMPORTANT NOTE: All custom tags implemented in Java
	* share a single registry key for storing settings (the main
	* CFX_J key). This means that to avoid name conflicts
	* you should preface the names of your settings with the
	* name of your <code>{@link CustomTag CustomTag}</code> class. </P>
	*
	* <P>For example, the code below retrieves the value of a setting named 
	* 'VerifyAddress' for a <code>CustomTag</code> 
	* class named <code>MyCustomTag</code>: </P>
	*
	* <pre><code>   String strVerify = request.getSetting(&quot;MyCustomTag.VerifyAddress&quot;) ;
	*   if ( Boolean.valueOf(strVerify) )
	*   {
	*      // Do address verification...
	*   } </code></pre>
	*
	* @param name  name of the setting to retrieve (case insensitive)
	*
	* @return	The value of the custom tag setting. If no setting
	*				of that name exists then an empty string is returned.
	*/
	public String getSetting( String name ) ;


	/**
	* Checks whether the tag contains the <code>DEBUG</code> attribute. You should use
	* this method to determine whether or not you need to write debug
	* information for this request (see <code>{@link Response#writeDebug Response.writeDebug}</code> for
	* details on writing debug information).
	*
	* <P>The following example checks to see whether the <code>DEBUG</code> attribute
	* is present, and if it is then it writes a brief debug message: </P>
	*
	* <pre><code>   if ( request.debug() )
	*   {
	*      response.writeDebug( &quot;debug info&quot; ) ;
	*   } </pre></code>
	*
	* @return Returns <code>true</code> if the tag contains the <code>DEBUG</code>
	*         attribute otherwise returns <code>false</code>.
	*
	* @see Response#writeDebug Response.writeDebug 
	*/
	public boolean debug() ;
}


